unit BS_ADOClientDataSet;
{$I BS.inc}
interface

uses
  SysUtils, Classes, DB, DBClient
  , Graphics, Controls, StdCtrls, Forms , Provider
  , BS_ADODataSet, bs_db, ADODB
  {$IFDEF CompilerVersion6},Variants
  , SqlExpr, SqlConst
  //{$ELSE}
  {$ENDIF}
  ;

type
{ TInternalBS_ADODataSet }

  TBS_InternalADODataSet = class(TBS_ADODataSet);

{$IFDEF CompilerVersion6}
  TBS_CustomADOClientDataSet = class(TCustomClientDataSet)
{$ELSE}
  TBS_CustomADOClientDataSet = class(TClientDataset)
{$ENDIF}
  private
    fConnection: TADOConnection;
    fInternalConnection: TADOConnection; { Always points to internal if present }
    fInternalDataSet: TBS_InternalADODataSet;
    fProvider: TProvider;
    fStoredCursor : TCursor;
    fExcludeCopyFieldsList : TStringList;
    fTableName:String;
    procedure SetExcludeCopyFields(Value : String);
    function  GetExcludeCopyFields:String;
    function  GetWaitCursor: TCursor;
    procedure SetWaitCursor(Value: TCursor);
    procedure SetCommandText(Value:WideString);
    function  GetCommandText:WideString;
    procedure SetUniqueFields(Value:String);
    function  GetUniqueFields:String;
    procedure SetSubstituteChar(Value:Char);
    function  GetSubstituteChar:Char;
    procedure SetSqlSubstitutions(Value: TSqlSubstitutions);
    function  GetSqlSubstitutions: TSqlSubstitutions;
    procedure SetSqlSubstitutionsCheck(Value: boolean);
    function  GetSqlSubstitutionsCheck: boolean;
  protected
    procedure AllocConnection; virtual;
    procedure AllocDataSet; virtual;
    procedure AllocProvider; virtual;
    procedure CloseCursor; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure OpenCursor(InfoQuery: Boolean); override;
    procedure SetConnection(Value: TADOConnection); virtual;
    function  GetConnection: TADOConnection;
    //procedure DoGetTableName(DataSet: TDataSet; var TableName: string); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;

    function  SearchCdsRecord(const aEdtSearch:Tedit):boolean; overload;
    function  SearchCdsRecord(const aEdtSearchText:String):boolean; overload;
    function  SortData(const aIndexFields,aDescFields:String):boolean;
    procedure FreeCreatedIndexes;
    procedure CopyRecord;
    procedure Execute; override;
  published
    property CommandText: WideString read GetCommandText write SetCommandText;
    property WaitCursor: TCursor read GetWaitCursor write SetWaitCursor default crHourGlass;
    property UniqueFields: String read GetUniqueFields write SetUniqueFields;
    property SubstituteChar: Char read GetSubstituteChar write SetSubstituteChar default DEFAULT_SUBSTITUTE_CHAR;
    property SqlSubstitutions: TSqlSubstitutions read GetSqlSubstitutions write SetSqlSubstitutions;
    property SqlSubstitutionsCheck : boolean read GetSqlSubstitutionsCheck write SetSqlSubstitutionsCheck default true;
    property TableName :String read fTableName write fTableName;
    property Active;
    property Aggregates;
    property AggregatesActive;
    property AutoCalcFields;

    property Connection: TADOConnection read GetConnection write SetConnection;
    property DataSet: TBS_InternalADODataSet read FInternalDataSet;
    property Constraints;

  {$IFDEF CompilerVersion6}
    property DisableStringTrim;
  {$ELSE}
  {$ENDIF}
    property FileName;
    property Filter;
    property Filtered;
    property FilterOptions;
    property FieldDefs;
    property IndexDefs;
    property IndexFieldNames;
    property IndexName;
    property FetchOnDemand;
    property MasterFields;
    property MasterSource;
    property ObjectView;
    property PacketRecords;
    property Params;
    property Provider: TProvider read FProvider;
    property ReadOnly;
    property StoreDefs;
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property BeforeRefresh;
    property AfterRefresh;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
    property OnReconcileError;
    property BeforeApplyUpdates;
    property AfterApplyUpdates;
    property BeforeGetRecords;
    property AfterGetRecords;
    property BeforeRowRequest;
    property AfterRowRequest;
    property BeforeExecute;
    property AfterExecute;
    property BeforeGetParams;
    property AfterGetParams;
    property ExcludeCopyFields : String read GetExcludeCopyFields write SetExcludeCopyFields;
  end;

  TBS_ADOClientDataSet = class(TBS_CustomADOClientDataSet);


  {$IFDEF CompilerVersion6}
  {$ELSE}
     resourcestring
     SCircularProvider = 'Circular provider references not allowed.';
  {$ENDIF}


implementation


{ TBS_CustomADOClientDataSet }
{$D+L+}
constructor TBS_CustomADOClientDataSet.Create(AOwner: TComponent);
begin
  inherited;
  fExcludeCopyFieldsList := TStringList.Create;
  fExcludeCopyFieldsList.Sorted:=True;
  //fWaitCursor:=crHourGlass;
  AllocProvider;
  AllocDataSet;
  AllocConnection;
  Self.FInternalDataSet.CursorType:=ctOpenForwardOnly;
  Self.SqlSubstitutionsCheck:=false;
end;

destructor TBS_CustomADOClientDataSet.Destroy;
begin
  fExcludeCopyFieldsList.Free;
  fExcludeCopyFieldsList:=nil;
  inherited;
end;


procedure TBS_CustomADOClientDataSet.AllocConnection;
begin
  FConnection := TADOConnection.Create(Self);
  FInternalConnection := FConnection;
  FConnection.Name := 'InternalConnection';		{ Do not localize }

  {$IFDEF CompilerVersion6}
    FConnection.SetSubComponent(True);
  {$ELSE}
  {$ENDIF}
  FInternalDataSet.Connection := FConnection;
end;

procedure TBS_CustomADOClientDataSet.AllocDataSet;
begin
  FInternalDataSet := TBS_InternalADODataSet.Create(Self);
  FInternalDataSet.Name := 'InternalDataSet';			{ Do not localize }
  FInternalDataSet.Connection := FConnection;
  {$IFDEF CompilerVersion6}
  FInternalDataSet.SetSubComponent(True);
  {$ELSE}
  {$ENDIF}
  FProvider.DataSet := FInternalDataSet;
end;


// Append new record and copy data from current
// excluding fields from ExcludeCopyFields property
procedure TBS_CustomADOClientDataSet.CopyRecord;
var
 dataArray,fieldsArray: Variant;
 i,index: integer;
begin
  dataArray := VarArrayCreate([0, Self.Fields.Count-1], varVariant); //Data
  fieldsArray := VarArrayCreate([0, Self.Fields.Count-1], varVariant); //Fields names
  for i:=0 to Self.Fields.Count-1 do  //Copy data in var array
    begin
      dataArray[i]:=Self.Fields[i].Value;
      fieldsArray[i]:=Self.Fields[i].FieldName;
    end;
  //insert;
  Self.Append;
  for i:=0 to Self.Fields.Count-1 do
    if not Self.FExcludeCopyFieldsList.Find(fieldsArray[i],index) then
      Self.Fields[i].Value:=dataArray[i]
end;


// Free all created custom indexes
procedure TBS_CustomADOClientDataSet.FreeCreatedIndexes;
var
  i:integer;
begin
  if Self.Active then
  try
    IndexDefs.Update; //Refresh indexes list
    for i := 0 to IndexDefs.Count - 1 do
      if (IndexDefs.Items[i].Name<>'DEFAULT_ORDER') and (IndexDefs.Items[i].Name<>'CHANGEINDEX')
         and (IndexDefs.Items[i].Name<>'PRIMARY_KEY') then
           begin
             if (Self.IndexName=IndexDefs.Items[i].Name) then
               Self.IndexName:='';
             DeleteIndex(IndexDefs.Items[i].Name);
           end;
     IndexDefs.Update;
  except
    raise;
  end;
end;

// Sort data using internal indexes by
// IndexFields and DescFields lists
function TBS_CustomADOClientDataSet.SortData(const aIndexFields,
  aDescFields: String): boolean;
var
  i                :integer;
  theIndexName     :String;
  theIndexFields   :String;
  theDescFields    :String;
  oldCursor        :TCursor;
  fIndexPresent    :boolean;
  recreateFlag     :boolean;
  p: Integer;

begin
 oldCursor:=Screen.Cursor;
 Screen.Cursor:=crHourGlass;
 theIndexFields   :=aIndexFields;
 theDescFields    :=aDescFields;

 try
   p := 1;
   theIndexName := '';
   while p <= Length(theIndexFields) do
     theIndexName := theIndexName +ExtractFieldName(theIndexFields, p)+ '_';

   p := 1;
   while p <= Length(theDescFields) do
     theIndexName := theIndexName +ExtractFieldName(theDescFields, p)+ 'Desc_';

   if theIndexName <> '' then
    theIndexName := 'i'+Copy(theIndexName,1,Length(theIndexName)-1);

   if theIndexName<>copy(theIndexName,1,30) then
     recreateFlag:=True
   else
     recreateFlag:=False;

   theIndexName:=copy(theIndexName,1,30);
   fIndexPresent:=False;
   Self.IndexDefs.Update;//Refresh
   //If index exist, then set it
   for i := 0 to Self.IndexDefs.Count - 1 do
    if LowerCase(Self.IndexDefs.Items[i].Name) = LowerCase(theIndexName) then
      begin
        if Not recreateFlag then
          begin
            fIndexPresent:=True; // Sign
            Self.IndexName := Self.IndexDefs.Items[i].Name;
          end
        else
          begin
            Self.DeleteIndex(theIndexName);
            Self.IndexName := '';
          end;
        break;
      end;
   //If index not exist, then try create it
   if Not(fIndexPresent)and (theIndexName<>'')  then
     try
       Self.AddIndex(theIndexName, theIndexFields,[ixCaseInsensitive],theDescFields,'',0);
       Self.IndexName := theIndexName;
     except
       //DatabaseErrorFmt('Can not create index ', [theIndexName], GetDataSet);
     end;
   Result:=True;
  finally
    Screen.Cursor:=oldCursor;
  end;
end;


function TBS_CustomADOClientDataSet.SearchCdsRecord(
  const aEdtSearchText: String): boolean;
var
  fieldsNamesStr:String;
  i:integer;
  searchTextStr:String;
  tmpSubStr:String;
  searchValuesArray:Variant;
begin
  fieldsNamesStr:='';
  searchTextStr:=aEdtSearchText;
  Result:=False;
  if Self.Active then
  try
    searchValuesArray := VarArrayCreate([0, Self.IndexFieldCount - 1], varVariant);
    for i := 0 to Self.IndexFieldCount - 1 do
      begin
        if fieldsNamesStr<>'' then
          fieldsNamesStr:=fieldsNamesStr+';';
        fieldsNamesStr:=fieldsNamesStr + Self.IndexFields[i].FieldName;
        if Pos(';',searchTextStr) > 0 then
          begin
            tmpSubStr:=Copy(searchTextStr,1,Pos(';',searchTextStr)-1);
            searchTextStr:=Copy(searchTextStr,Pos(';',searchTextStr)+1,
                           Length(searchTextStr)-(Pos(';',searchTextStr)));
          end
        else
          begin
            tmpSubStr:=searchTextStr;
            searchTextStr:='';
          end;
        try
          case Self.IndexFields[i].DataType of
            ftSmallint,ftInteger,ftBytes,ftVarBytes,
            ftAutoInc,ftLargeint,ftTypedBinary:
              if tmpSubStr<>'' then
                searchValuesArray[i]:=StrToInt(tmpSubStr)
              else
                searchValuesArray[i]:=0;
            ftWord,ftFloat,ftBCD,ftCurrency:
              if tmpSubStr<>'' then
                searchValuesArray[i]:=StrToFloat(tmpSubStr)
              else
                searchValuesArray[i]:=0;
            ftDate, ftTime, ftDateTime:
              searchValuesArray[i]:=StrToDate(tmpSubStr);
          else
            searchValuesArray[i]:=tmpSubStr;
          end;
        except
          searchValuesArray[i]:=tmpSubStr;
        end;
      end;

    if Self.Locate(fieldsNamesStr, searchValuesArray,
       [loCaseInsensitive,loPartialKey]) then
      Result:=True
    else
      Result:=False;
  except      //Try restore last position
    Result:=False;
  end;
end;

function TBS_CustomADOClientDataSet.SearchCdsRecord(
  const aEdtSearch: Tedit): boolean;
begin
  Result:=Self.SearchCdsRecord(aEdtSearch.text);
  if Result  then
    aEdtSearch.Font.Color:=clWindowText
  else
    aEdtSearch.Font.Color:=clRed; //Can't find
end;

// overrided to show WaitCursor while open dataset
procedure TBS_CustomADOClientDataSet.OpenCursor(InfoQuery: Boolean);
begin
  Self.fStoredCursor:=Screen.Cursor;
  if (WaitCursor <> crDefault) and (WaitCursor<>Self.fStoredCursor) then
    Screen.Cursor:=WaitCursor;
  try
    if Assigned(FProvider) then
      SetProvider(FProvider);
    if FProvider.DataSet = Self then
      raise Exception.Create(SCircularProvider);
    inherited;
  finally
    if (WaitCursor <> crDefault) and (WaitCursor<>Self.fStoredCursor) then
      Screen.Cursor:=Self.fStoredCursor;
  end;

end;

//Try extract Exclude copy fields from string to list
function TBS_CustomADOClientDataSet.GetExcludeCopyFields: String;
var i:integer;
begin
  Result:='';
  for i:=0 to FExcludeCopyFieldsList.Count-1 do
    begin
      if i<>(FExcludeCopyFieldsList.Count-1) then
        Result:=Result+FExcludeCopyFieldsList.Strings[i]+';'
      else
        Result:=Result+FExcludeCopyFieldsList.Strings[i]
    end;
end;


procedure TBS_CustomADOClientDataSet.SetExcludeCopyFields(Value: String);
var
  p: Integer;
begin
  FExcludeCopyFieldsList.Clear;
  p := 1;
  while p <= Length(Value) do
    FExcludeCopyFieldsList.Add(ExtractFieldName(Value, p));
end;


procedure TBS_CustomADOClientDataSet.Execute;
begin
  TBS_InternalADODataSet(Self.DataSet).SubstituteSQL;
  inherited;
end;

procedure TBS_CustomADOClientDataSet.CloseCursor;
begin
  inherited;
end;

procedure TBS_CustomADOClientDataSet.SetWaitCursor(Value: TCursor);
begin
  Self.FInternalDataSet.WaitCursor:=Value;
end;

function TBS_CustomADOClientDataSet.GetWaitCursor: TCursor;
begin
  {if Assigned(Self.DataSet) then
  Result:=TBS_InternalADODataSet(Self.DataSet).WaitCursor
  else
    Result:=crHourGlass;}
  Result:=Self.FInternalDataSet.WaitCursor;
end;


procedure TBS_CustomADOClientDataSet.SetConnection(Value: TADOConnection);
begin
  { Assigning existing value or clearing internal connection is a NOP }
  if (Value = FConnection) or ((Value = nil) and Assigned(FInternalConnection)) then
    Exit;
  { Remove FreeNotification from existing external reference }
  if FConnection <> FInternalConnection then
    FConnection.RemoveFreeNotification(Self);
  { Reference to external connection was cleared, recreate internal }
  if (Value = nil) then
    AllocConnection
  else
  begin
    { Free the internal connection when assigning an external connection }
    if Assigned(FInternalConnection) and
       { but not if we are streaming in, then wait until loaded is called }
       not (csLoading in FInternalConnection.ComponentState) then
      FreeAndNil(FInternalConnection);
    FConnection := Value;
    FConnection.FreeNotification(Self);
    FInternalDataSet.Connection := FConnection;
  end;
end;

function TBS_CustomADOClientDataSet.GetConnection: TADOConnection;
begin
  Result:=Self.FInternalDataSet.Connection;
end;

function TBS_CustomADOClientDataSet.GetUniqueFields: String;
begin
  Result:=Self.FInternalDataSet.UniqueFields;
end;

procedure TBS_CustomADOClientDataSet.SetUniqueFields(Value: String);
begin
  Self.FInternalDataSet.UniqueFields:=Value;
end;

function TBS_CustomADOClientDataSet.GetSubstituteChar: Char;
begin
  Result:= Self.FInternalDataSet.SubstituteChar;
end;

procedure TBS_CustomADOClientDataSet.SetSubstituteChar(Value: Char);
begin
  Self.FInternalDataSet.SubstituteChar :=Value;
end;

function TBS_CustomADOClientDataSet.GetSqlSubstitutions: TSqlSubstitutions;
begin
  Result:= Self.FInternalDataSet.SqlSubstitutions;
end;

procedure TBS_CustomADOClientDataSet.SetSqlSubstitutions(
  Value: TSqlSubstitutions);
begin
  Self.FInternalDataSet.SqlSubstitutions :=Value;
end;

function TBS_CustomADOClientDataSet.GetSqlSubstitutionsCheck: boolean;
begin
  Result:=  Self.FInternalDataSet.SqlSubstitutionsCheck;
end;

procedure TBS_CustomADOClientDataSet.SetSqlSubstitutionsCheck(
  Value: boolean);
begin
  Self.FInternalDataSet.SqlSubstitutionsCheck :=Value;
end;

function TBS_CustomADOClientDataSet.GetCommandText: WideString;
begin
  if Assigned(Self.DataSet) then
    Result:=TBS_InternalADODataSet(Self.DataSet).CommandText
  else
    Result:='';
end;

procedure TBS_CustomADOClientDataSet.SetCommandText(Value: WideString);
begin
  if Assigned(Self.DataSet) then
    if Value<>TBS_InternalADODataSet(Self.DataSet).CommandText then
      TBS_InternalADODataSet(Self.DataSet).CommandText:=Value;
end;

procedure TBS_CustomADOClientDataSet.AllocProvider;
begin
  FProvider := TProvider.Create(Self);
  FProvider.DataSet := FInternalDataSet;
  FProvider.Name := 'InternalProvider';			{ Do not localize }
  {$IFDEF CompilerVersion6}
  FProvider.SetSubComponent(True);
  {$ELSE}
  {$ENDIF}
  SetProvider(FProvider);
  FProvider.UpdateMode:=upWhereKeyOnly;
  FProvider.ResolveToDataSet:=false;
  FProvider.Options:=[poAutoRefresh];
  //FProvider.OnGetTableName:=Self.DoGetTableName;
end;

procedure TBS_CustomADOClientDataSet.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if not (csDestroying in ComponentState ) and (Operation = opRemove) and
     (AComponent = FConnection) and (AComponent.Owner <> Self) then
  AllocConnection;
end;

end.
