unit BS_ADODataSet;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, bs_db,  ADODB, ADOInt,ADOConst
  ,stdctrls;

type
  TBS_ADODataSet = class(TADODataSet)
  private
    fWaitCursor   : TCursor;
    fStoredCursor : TCursor;
    fUniqueFields: string;

    FCommandStrings  : TStringList;
    fSubstituteChar: Char;
    fSqlSubstitutions: BS_DB.TSqlSubstitutions;
    fSqlSubstitutionsCheck:boolean;
    procedure SetSubstituteChar(Value: Char);
    procedure SetSqlSubstitutions(Value: TSqlSubstitutions);
    function  GetRealCommandText: String;
    procedure SetInternalCommandText(Value:WideString);
    function  GetInternalCommandText: WideString;
    procedure CommandTextChanged(Sender: TObject);

  protected
    procedure OpenCursor(InfoQuery: Boolean); override;
    function GetKeyFields: string;
    function PSGetKeyFields: string; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function  ParseSqlSubstitutions(aCommandText: string; DoCreate: Boolean): string;
    procedure SubstituteSQL;
    function  SqlSubstitutionByName(const Value: string): TSqlSubstitution;
    function  SearchRecord(const KeyFields: String; const KeyValues: Variant): boolean; overload;
    function  SearchRecord(const KeyFields:String; const aEdtSearch:Tedit):boolean; overload;
  published
    property WaitCursor:  TCursor read fWaitCursor write fWaitCursor default crHourGlass;
    property UniqueFields: string read FUniqueFields write FUniqueFields;
    property SubstituteChar: Char read fSubstituteChar write SetSubstituteChar default DEFAULT_SUBSTITUTE_CHAR;
    property SqlSubstitutions: TSqlSubstitutions read fSqlSubstitutions write SetSqlSubstitutions;
    property SqlSubstitutionsCheck : boolean read fSqlSubstitutionsCheck write fSqlSubstitutionsCheck default true;
  end;
{&D-L-}
function FieldTypeToADOType(const FieldType: TFieldType): DataTypeEnum;

implementation

function FieldTypeToADOType(const FieldType: TFieldType): DataTypeEnum;
begin
  case FieldType of
    ftUnknown: Result := adEmpty;
    ftString, ftWideString: Result := adVarChar;
    ftSmallint: Result := adSmallint;
    ftInteger, ftAutoInc: Result := adInteger;
    ftWord: Result := adUnsignedSmallInt;
    ftBoolean: Result := adBoolean;
    ftFloat: Result := adDouble;
    ftCurrency, ftBCD: Result := adCurrency;
    ftDate: Result := adDBDate;
    ftTime: Result := adDBTime;
    ftDateTime: Result := adDBTimeStamp;
    ftBytes: Result := adBinary;
    ftVarBytes: Result := adVarBinary;
    ftMemo: Result := adLongVarChar;
    ftBlob, ftGraphic..ftTypedBinary: Result := adLongVarBinary;
    ftFixedChar: Result := adChar;
    ftLargeint: Result := adBigInt;
    ftVariant: Result := adVariant;
    ftInterface: Result := adIUnknown;
    ftIDispatch: Result := adIDispatch;
    ftGuid: Result := adGUID;
  else
    DatabaseErrorFmt(SNoMatchingADOType, [FieldTypeNames[FieldType]]);
    Result := adEmpty;
  end;
end;


procedure TBS_ADODataSet.CommandTextChanged(Sender: TObject);
begin
  if not (csReading in ComponentState) then
    begin
      if SqlSubstitutionsCheck or (csDesigning in ComponentState) then
        begin
          Self.ParseSqlSubstitutions(CommandText, true);
        end;
      DataEvent(dePropertyChange, 0);
    end
end;
{&D-L-}
constructor TBS_ADODataSet.Create(AOwner: TComponent);
begin
  inherited;
  fCommandStrings := TStringList.Create;
  //TStringList(fCommandStrings).OnChange:=Self.CommandTextChanged;
  fSubstituteChar:=DEFAULT_SUBSTITUTE_CHAR;
  fSqlSubstitutions := TSqlSubstitutions.Create(Self);
  fSqlSubstitutionsCheck := true;
  fWaitCursor:=crHourGlass;
  Self.EnableBCD:=false;
end;


destructor TBS_ADODataSet.Destroy;
begin
  fCommandStrings.Free;
  fSqlSubstitutions.Free;
  fSqlSubstitutions:=nil;
  inherited;
end;

function TBS_ADODataSet.GetInternalCommandText: WideString;
begin
  Result:=Self.FCommandStrings.Text;
end;

function TBS_ADODataSet.GetKeyFields: string;
begin
  Result := Trim(FUniqueFields);
end;

function TBS_ADODataSet.GetRealCommandText: String;
begin
  SubstituteSQL;
  Result:=inherited CommandText;
end;

procedure TBS_ADODataSet.OpenCursor(InfoQuery: Boolean);
begin
  Self.fStoredCursor:=Screen.Cursor;
  if (WaitCursor <> crDefault) and (WaitCursor<>Self.fStoredCursor) then
    Screen.Cursor:=WaitCursor;
  try
    //SubstituteSQL;
    inherited OpenCursor(InfoQuery);
  finally
    if (WaitCursor <> crDefault) and (WaitCursor<>Self.fStoredCursor) then
      Screen.Cursor:=Self.fStoredCursor;
  end;

end;

function TBS_ADODataSet.ParseSqlSubstitutions(aCommandText: string;
  DoCreate: Boolean): string;
const
  Literals = ['''', '"', '`'];
var
  Value, CurPos, StartPos: PChar;
  CurChar: Char;
  Literal: Boolean;
  EmbeddedLiteral: Boolean;
  Name: string;
  substitutionValue:String;
  newCommandText:WideString;
  substitutionList:TSqlSubstitutions;
  function NameDelimiter: Boolean;
  begin
    Result := CurChar in [' ', ',', ';', ')', #13, #10];
  end;

  function IsLiteral: Boolean;
  begin
    Result := CurChar in Literals;
  end;

  function StripLiterals(Buffer: PChar): string;
  var
    Len: Word;
    TempBuf: PChar;

    procedure StripChar;
    begin
      if TempBuf^ in Literals then
        StrMove(TempBuf, TempBuf + 1, Len - 1);
      if TempBuf[StrLen(TempBuf) - 1] in Literals then
        TempBuf[StrLen(TempBuf) - 1] := #0;
    end;

  begin
    Len := StrLen(Buffer) + 1;
    TempBuf := AllocMem(Len);
    Result := '';
    try
      StrCopy(TempBuf, Buffer);
      StripChar;
      Result := StrPas(TempBuf);
    finally
      FreeMem(TempBuf, Len);
    end;
  end;

begin
  Result := aCommandText;
  newCommandText:='';
  substitutionList:=TSqlSubstitutions.Create(Self);
  try
   Value := PChar(Result);   //ResultCommandText
    //if DoCreate then  Self.SqlSubstitutions.Clear;
    CurPos := Value;
    Literal := False;
    EmbeddedLiteral := False;

    repeat
      while (CurPos^ in LeadBytes) do
        Inc(CurPos, 2);
      CurChar := CurPos^;
      if (CurChar = Self.SubstituteChar) and not Literal and ((CurPos + 1)^ <> Self.SubstituteChar) then
      begin
        StartPos := CurPos;
        while (CurChar <> #0) and (Literal or not NameDelimiter) do
          begin
            Inc(CurPos);
            while (CurPos^ in LeadBytes) do
              Inc(CurPos, 2);
            CurChar := CurPos^;
            if IsLiteral then
            begin
              Literal := Literal xor True;
              if CurPos = StartPos + 1 then
                EmbeddedLiteral := True;
            end;
          end;
          CurPos^ := #0;
          if EmbeddedLiteral then
            begin
              Name := StripLiterals(StartPos + 1);
              EmbeddedLiteral := False;
            end
          else
            Name := StrPas(StartPos + 1);
          //if DoCreate then
          //begin
            //if Self.SqlSubstitutions.FindSqlSubstitution(Name)= nil then
            //  Self.SqlSubstitutions.Add.Name:= Name;
            if substitutionList.FindSqlSubstitution(Name)= nil then
            begin
              substitutionList.Add.Name:= Name;
              //Try copy old substitutionvalue
              if Self.SqlSubstitutions.FindSqlSubstitution(Name)<> nil then
                TSqlSubstitution(substitutionList.FindSqlSubstitution(Name)).Value:=
                  TSqlSubstitution(Self.SqlSubstitutions.FindSqlSubstitution(Name)).Value;
            end;
          //end;
          CurPos^ := CurChar;
          substitutionValue := TSqlSubstitution(substitutionList.SqlSubstitutionByName(Name)).Value;
          newCommandText:=newCommandText+substitutionValue;
        end
      else
        if (CurChar = Self.SubstituteChar) and not Literal and ((CurPos + 1)^ = Self.SubstituteChar) then
          begin
            StrMove(CurPos, CurPos + 1, StrLen(CurPos) + 1);
          end
        else
          begin
            if IsLiteral then
              Literal := Literal xor True;
            newCommandText:=newCommandText+CurChar;
          end;
      Inc(CurPos);
    until CurChar = #0;
   {if DoCreate then //Copy new substitution list
   begin
     Self.SqlSubstitutions.Clear;
     Self.SqlSubstitutions.AssignValues(substitutionList);
   end;}
   Result:=newCommandText;
  finally
    substitutionList.Free;
  end;
end;


function TBS_ADODataSet.PSGetKeyFields: string;
begin
  // First try to obtain the unique fields through the ProviderFlags of the Fields
  Result := inherited PSGetKeyFields;
  // If these are not defined, use the UniqueFields or constraints of the DataSet
  if Result = '' then
    if (Trim(FUniqueFields) <> '')  then
      Result := GetKeyFields;
end;



procedure TBS_ADODataSet.SetInternalCommandText(Value: WideString);
begin
  Self.FCommandStrings.Text:=Value;
end;

procedure TBS_ADODataSet.SetSqlSubstitutions(Value: TSqlSubstitutions);
begin
  fSqlSubstitutions.Assign(Value);
end;

procedure TBS_ADODataSet.SetSubstituteChar(Value: Char);
begin
  if Value <> fSubstituteChar then
  begin
    fSubstituteChar := Value;
    //RecreateSubstitutions;
  end;
end;

function TBS_ADODataSet.SqlSubstitutionByName(
  const Value: string): TSqlSubstitution;
begin
  Result:=Self.fSqlSubstitutions.SqlSubstitutionByName(Value);
end;

procedure TBS_ADODataSet.SubstituteSQL;
begin
  inherited CommandText :=Self.ParseSqlSubstitutions(Self.CommandText,false);
end;

function TBS_ADODataSet.SearchRecord(const KeyFields: String;const KeyValues: Variant): boolean;
begin
  Self.First;
  if Self.Locate(KeyFields, KeyValues, [loCaseInsensitive,loPartialKey]) then
    result:=true
  else
    result:=false;

end;

function TBS_ADODataSet.SearchRecord(const KeyFields: String;
  const aEdtSearch: Tedit): boolean;
begin
  result:=Self.SearchRecord(KeyFields,aEdtSearch.text);
  if result  then
    aEdtSearch.Font.Color:=clWindowText
  else
    aEdtSearch.Font.Color:=clRed; {Can't find}
end;

end.
