unit BS_ADOOracleExtract;

interface

uses
  SysUtils, Classes,
  db,ADOdb;                         

type
  TBS_ExtractObjectTypes =
    (exoDatabase,
     exoTable,
     exoView,
     exoProcedure,
     exoFunction,
     exoSequence,
     exoException,
     exoIndexes,
     exoChecks,
     exoData,
     exoBLOBFilter,
     exoRole,
     exoTrigger,
     exoForeign
     );

  TBS_ExtractType =
    (extTable, extRole, extTrigger, extForeign,
     extIndex, extData, extGrant, extCheck, extAlterProc);

  TBS_ExtractTypes = Set of TBS_ExtractType;

  TBS_ADOOracleExtract = class(TComponent)
  private
    FConnection : TADOConnection;
    FMetaData: TStrings;
    FCommitInterval:integer;
    //FDatabaseInfo: TIBDatabaseInfo;
    FShowSystemObjects: Boolean;
    function GetConnection: TADOConnection;
    function GextIndexSegments ( indexname : String) : String;
    procedure SetConnection(const Value: TADOConnection);
    function PrintValidation(ToValidate : String;	flag : Boolean) : String;
    procedure ShowGrants(MetaObject: String; Terminator : String);
    procedure ShowGrantRoles(Terminator : String);
    procedure GetProcedureArgs(Proc : String);
    function GetFieldLength(sql : TADODataSet) : Integer;
    function CreateADODataSet : TADODataSet;
  protected
    function PrintSet(var Used : Boolean) : String;
    procedure ListSequences(aOwnerName:String = ''; aSequenceName : String = '');
    procedure ListIndex(aOwner: String = ''; aObjectName : String = ''; ExtractType : TBS_ExtractType = extIndex);

    procedure ListData(ObjectName : String);
    procedure ListRoles(ObjectName : String = '');
    procedure ListGrants;
    procedure ListProcs(ProcedureName : String = ''; AlterOnly : Boolean = false);
    procedure ListAllTables(const aOwner:String; const aShowSystem: Boolean);
    procedure ListTriggers(ObjectName : String = ''; ExtractType : TBS_ExtractType = extTrigger);
    procedure ListCheck(ObjectName : String = ''; ExtractType : TBS_ExtractType = extCheck);
    procedure ListCreateDb(TargetDb : String = '');
    procedure ListException(ExceptionName : String = '');
    procedure ListFilters(FilterName : String = '');
    procedure ListForeign(ObjectName : String = ''; ExtractType : TBS_ExtractType = extForeign);
    procedure ListFunctions(FunctionName : String = '');

    procedure ListViews(ViewName : String = '');
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;


    function GetArrayField(FieldName : String) : String;
    function GetFieldType(FieldType, FieldSubType, FieldScale, FieldSize,
      FieldPrec, FieldLen : Integer) : String;
    //function GetCharacterSets(CharSetId, Collation : SmallInt;	CollateOnly : Boolean) : String;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure ExtractObject(
      aObjectType : TBS_ExtractObjectTypes;
      aObjectName : String = '';
      aExtractTypes : TBS_ExtractTypes = []);

    //property DatabaseInfo : TIBDatabaseInfo read FDatabaseInfo;
    property Items : TStrings read FMetaData;
    function GetConsColumns(const aOwnerName,aTableName,aConstraintName: String): String;
    function GetTableComment(const aOwnerName,aTableName: String): String;

    function ExtractDDL(Flag : Boolean; TableName : String) : Boolean;
    function ListTable(
                aOwnerName,
                aRelationName,
                aNewName : String) : Boolean;
    procedure ExtractListView (ViewName : String);


  published
    { Published declarations }
    property Connection : TADOConnection read GetConnection write SetConnection;
    property ShowSystemObjects: Boolean read FShowSystemObjects write FShowSystemObjects;
    property CommitInterval: integer read FCommitInterval write FCommitInterval;
  end;

  TSQLType = record
    SqlType : Integer;
    TypeName : String;
  end;

  TPrivTypes = record
    PrivFlag : Integer;
    PrivString : String;
  end;

  TSQLTypes = Array[0..13] of TSQLType;

const

  priv_UNKNOWN = 1;
  priv_SELECT = 2;
  priv_INSERT = 4;
  priv_UPDATE = 8;
  priv_DELETE = 16;
  priv_EXECUTE = 32;
  priv_REFERENCES = 64;

  // In pascal, how does one deal with the below "#define"?
  // bst_word(n) ((n) % 256), ((n) / 256)
  bst_text                       =         14;
  bst_text2                      =         15;
  bst_SmallInt                      =          7;
  bst_long                       =          8;
  bst_quad                       =          9;
  bst_float                      =         10;
  bst_double                     =         27;
  bst_d_float                    =         11;
  bst_timestamp                  =         35;
  bst_varying                    =         37;
  bst_varying2                   =         38;
  bst_blob                       =        261;
  bst_cstring                    =         40;
  bst_cstring2                   =         41;
  bst_blob_id                    =         45;
  bst_sql_date                   =         12;
  bst_sql_time                   =         13;
  bst_int64                      =         16;
  bst_date                       =         bst_timestamp;


 PrivTypes : Array[0..5] of TPrivTypes = (
  (PrivFlag : priv_DELETE; PrivString : 'DELETE' ),
  (PrivFlag : priv_EXECUTE; PrivString : 'EXECUTE' ),
  (PrivFlag : priv_INSERT; PrivString : 'INSERT' ),
  (PrivFlag : priv_SELECT; PrivString : 'SELECT' ),
  (PrivFlag : priv_UPDATE; PrivString : 'UPDATE' ),
  (PrivFlag : priv_REFERENCES; PrivString : 'REFERENCES'));

  ColumnTypes : TSQLTypes = (
    (SqlType : bst_SmallInt; TypeName :	'SMALLINT'),	        // NTX: keyword
    (SqlType : bst_long; TypeName : 'INTEGER'),	        	// NTX: keyword
    (SqlType : bst_quad; TypeName : 'QUAD'),		        // NTX: keyword
    (SqlType : bst_float; TypeName : 'FLOAT'),	        	// NTX: keyword
    (SqlType : bst_text; TypeName : 'CHAR'),		        // NTX: keyword
    (SqlType : bst_double; TypeName : 'DOUBLE PRECISION'),	// NTX: keyword
    (SqlType : bst_varying; TypeName : 'VARCHAR'),		// NTX: keyword
    (SqlType : bst_cstring; TypeName : 'CSTRING'),		// NTX: keyword
    (SqlType : bst_blob_id; TypeName : 'BLOB_ID'),		// NTX: keyword
    (SqlType : bst_blob; TypeName : 'BLOB'),	        	// NTX: keyword
    (SqlType : bst_sql_time; TypeName : 'TIME'),		// NTX: keyword
    (SqlType : bst_sql_date; TypeName : 'DATE'),		// NTX: keyword
    (SqlType : bst_timestamp; TypeName : 'TIMESTAMP'),		// NTX: keyword
    (SqlType : bst_int64; TypeName : 'INT64'));

  SubTypes : Array[0..8] of String = (
    'UNKNOWN',			// NTX: keyword
    'TEXT',				// NTX: keyword
    'BLR',				// NTX: keyword
    'ACL',				// NTX: keyword
    'RANGES',			// NTX: keyword
    'SUMMARY',			// NTX: keyword
    'FORMAT',			// NTX: keyword
    'TRANSACTION_DESCRIPTION',	// NTX: keyword
    'EXTERNAL_FILE_DESCRIPTION');	// NTX: keyword

  TriggerTypes : Array[0..6] of String = (
    '',
    'BEFORE INSERT',			// NTX: keyword
    'AFTER INSERT',				// NTX: keyword
    'BEFORE UPDATE',			// NTX: keyword
    'AFTER UPDATE',				// NTX: keyword
    'BEFORE DELETE',			// NTX: keyword
    'AFTER DELETE');			// NTX: keyword

  IntegralSubtypes : Array[0..2] of String = (
    'UNKNOWN',			// Defined type, NTX: keyword
    'NUMERIC',			// NUMERIC, NTX: keyword
    'DECIMAL'); 		// DECIMAL, NTX: keyword

{  ODS_VERSION6 = 6;	// on-disk structure as of v3.0
  ODS_VERSION7 = 7;	// new on disk structure for fixing index bug
  ODS_VERSION8 =	8;	//new btree structure to support pc semantics
  ODS_VERSION9 =	9;	// btree leaf pages are always propogated up
  ODS_VERSION10 = 10; // V6.0 features. SQL delimited idetifier,
                      //                  SQLDATE, and 64-bit exact numeric
                      //                  type

  // flags for RDB$FILE_FLAGS
  FILE_shadow = 1;
  FILE_inactive = 2;
  FILE_manual = 4;
  FILE_cache = 8;
  FILE_conditional = 16;

  // flags for RDB$LOG_FILES
  LOG_serial = 1;
  LOG_default = 2;
  LOG_raw = 4;
  LOG_overflow = 8;



  MAX_INTSUBTYPES = 2;
  MAXSUBTYPES = 8;     // Top of subtypes array
}
// Object types used in RDB$DEPENDENCIES and RDB$USER_PRIVILEGES

{  obj_relation = 0;
  obj_view = 1;
  obj_trigger = 2;
  obj_computed = 3;
  obj_validation = 4;
  obj_procedure = 5;
  obj_expression_index = 6;
  obj_exception = 7;
  obj_user = 8;
  obj_field = 9;
  obj_index = 10;
  obj_count = 11;
  obj_user_group = 12;
  obj_sql_role = 13;
}
implementation

const
  NEWLINE = #13#10;
  TERM = ';';
  ProcTerm = '^';

  CollationSQL =
    'SELECT CST.RDB$CHARACTER_SET_NAME, COL.RDB$COLLATION_NAME, CST.RDB$DEFAULT_COLLATE_NAME ' +
    'FROM RDB$COLLATIONS COL JOIN RDB$CHARACTER_SETS CST ON ' +
    '  COL.RDB$CHARACTER_SET_ID = CST.RDB$CHARACTER_SET_ID ' +
    'WHERE ' +
    '  COL.RDB$COLLATION_ID = :COLLATION AND ' +
    '  CST.RDB$CHARACTER_SET_ID = :CHAR_SET_ID ' +
    'ORDER BY COL.RDB$COLLATION_NAME, CST.RDB$CHARACTER_SET_NAME';

  NonCollationSQL =
    'SELECT CST.RDB$CHARACTER_SET_NAME ' +
    'FROM RDB$CHARACTER_SETS CST ' +
    'WHERE CST.RDB$CHARACTER_SET_ID = :CHARSETID ' +
    'ORDER BY CST.RDB$CHARACTER_SET_NAME';

  PrecisionSQL =
    'SELECT * FROM RDB$FIELDS ' +
    'WHERE RDB$FIELD_NAME = :FIELDNAME';

  ArraySQL =
    'SELECT * FROM RDB$FIELD_DIMENSIONS FDIM ' +
    'WHERE ' +
    '  FDIM.RDB$FIELD_NAME = :FIELDNAME ' +
    'ORDER BY FDIM.RDB$DIMENSION';

{ TBS_ADOOracleExtract }



{	                ArrayDimensions
   Functional description
   Retrieves the dimensions of arrays and prints them.

  	Parameters:  fieldname -- the actual name of the array field }



function TBS_ADOOracleExtract.GetArrayField(FieldName: String): String;
//var
//  qryArray : TADODataSet;
begin
{  qryArray := CreateADODataSet;
  Result := '[';
  qryArray.SQL.Add(ArraySQL);
  qryArray.Params.ByName('FieldName').AsString := FieldName;
  qryArray.ExecQuery;

    //  Format is [lower:upper, lower:upper,..]

  while not qryArray.Eof do
  begin
    if (qryArray.FieldByName('RDB$DIMENSION').AsInteger > 0) then
      Result := Result + ', ';
    Result := Result + qryArray.FieldByName('RDB$LOWER_BOUND').AsString + ':' +
           qryArray.FieldByName('RDB$UPPER_BOUND').AsString;
    qryArray.Next;
  end;

  Result := Result + '] ';
  qryArray.Free;
}
end;




{	           ExtractListView
  Functional description
   	Show text of the specified view.
   	Use a SQL query to get the info and print it.
 	  Note: This should also contain check option }

procedure TBS_ADOOracleExtract.ExtractListView(ViewName: String);
const
  ViewsSQL = 'SELECT * FROM RDB$RELATIONS REL ' +
             ' WHERE ' +
             '  (REL.RDB$SYSTEM_FLAG <> 1 OR REL.RDB$SYSTEM_FLAG IS NULL) AND ' +
             '  NOT REL.RDB$VIEW_BLR IS NULL AND ' +
             '  REL.RDB$RELATION_NAME = :VIEWNAME AND ' +
             '  REL.RDB$FLAGS = 1 ' +
             'ORDER BY REL.RDB$RELATION_ID ';

  ColumnsSQL = 'SELECT * FROM RDB$RELATION_FIELDS RFR ' +
               'WHERE ' +
               '  RFR.RDB$RELATION_NAME = :RELATIONNAME ' +
               'ORDER BY RFR.RDB$FIELD_POSITION ';

//var
  //qryViews, qryColumns : TADODataSet;
  //RelationName, ColList : String;
begin
  {qryViews := CreateADODataSet;
  qryColumns := CreateADODataSet;
  try
    qryViews.SQL.Add(ViewsSQL);
    qryViews.Params.ByName('viewname').AsString := ViewName;
    qryViews.ExecQuery;
    while not qryViews.Eof do
    begin
      FMetaData.Add('');
      RelationName := QuoteIdentifier(FDatabase.SQLDialect,
          qryViews.FieldByName('RDB$RELATION_NAME').AsString);
      FMetaData.Add(Format('%s/* View: %s, Owner: %s */%s', [
        RelationName,
        Trim(qryViews.FieldByName('RDB$OWNER_NAME').AsString)]));
      FMetaData.Add('');
      FMetaData.Add(Format('CREATE VIEW %s (', [RelationName]));

      // Get Column List
      qryColumns.SQL.Add(ColumnsSQL);
      qryColumns.Params.ByName('relationname').AsString := RelationName;
      qryColumns.ExecQuery;
      while not qryColumns.Eof do
      begin
        ColList := ColList + QuoteIdentifier(FDatabase.SQLDialect,
              qryColumns.FieldByName('RDB$FIELD_NAME').AsString);
        qryColumns.Next;
        if not qryColumns.Eof then
          ColList := ColList + ', ';
      end;
      FMetaData.Add(ColList + ') AS');
      FMetaData.Add(qryViews.FieldByName('RDB$VIEW_SOURCE').AsString + Term);
      qryViews.Next;
    end;
  finally
    qryViews.Free;
    qryColumns.Free;
  end;
  }
end;


function TBS_ADOOracleExtract.GetConnection: TADOConnection;
begin
  result := FConnection;
end;

 { GextIndexSegments
   Functional description
  	returns the list of columns in an index. }

function TBS_ADOOracleExtract.GextIndexSegments(IndexName: String): String;
const
  IndexNamesSQL =
    'SELECT * FROM RDB$INDEX_SEGMENTS SEG ' +
    'WHERE SEG.RDB$INDEX_NAME = :INDEXNAME ' +
    'ORDER BY SEG.RDB$FIELD_POSITION';

//var
  //qryColNames : TADODataSet;
begin
// Query to get column names
  Result := '';
{  qryColNames := CreateADODataSet;
  try
    qryColNames.SQL.Add(IndexNamesSQL);
    qryColNames.Params.ByName('IndexName').AsString := IndexName;
    qryColNames.ExecQuery;
    while not qryColNames.Eof do
    begin
      // Place a comma and a blank between each segment column name

      Result := Result + QuoteIdentifier(FDatabase.SQLDialect,
        qryColNames.FieldByName('RDB$FIELD_NAME').AsString);
      qryColNames.Next;
      if not qryColNames.Eof then
        Result := Result + ', ';
    end;
  finally
    qryColNames.Free;
  end;
  }
end;


{	   ListAllGrants
  Functional description
 	 Print the permissions on all user tables.
 	 Get separate permissions on table/views and then procedures }

procedure TBS_ADOOracleExtract.ListGrants;
const
  SecuritySQL = 'SELECT * FROM RDB$RELATIONS ' +
                'WHERE ' +
                '  (RDB$SYSTEM_FLAG <> 1 OR RDB$SYSTEM_FLAG IS NULL) AND ' +
                '  RDB$SECURITY_CLASS STARTING WITH ''SQL$'' ' +
                'ORDER BY RDB$RELATION_NAME';

  ProcedureSQL = 'select * from RDB$PROCEDURES ' +
                 'Order BY RDB$PROCEDURE_NAME';

//var
  //qryRoles : TADODataSet;
  //RelationName : String;
begin
  {ListRoles;
  qryRoles := CreateADODataSet;
  try
  // This version of cursor gets only sql tables identified by security class
  //   and misses views, getting only null view_source

    FMetaData.Add('');
    FMetaData.Add('/* Grant permissions for this database */');
    FMetaData.Add('');

    try
      qryRoles.SQL.Text := SecuritySQL;
      qryRoles.ExecQuery;
      while not qryRoles.Eof do
      begin
        RelationName := Trim(qryRoles.FieldByName('rdb$relation_Name').AsString);
        ShowGrants(RelationName, Term);
        qryRoles.Next;
      end;
    finally
     qryRoles.Close;
    end;

    ShowGrantRoles(Term);

    qryRoles.SQL.Text := ProcedureSQL;
    qryRoles.ExecQuery;
    try
      while not qryRoles.Eof do
      begin
        ShowGrants(Trim(qryRoles.FieldByName('RDB$PROCEDURE_NAME').AsString), Term);
        qryRoles.Next;
      end;
    finally
      qryRoles.Close;
    end;
  finally
    qryRoles.Free;
  end;
  }
end;

{	  ListAllProcs
  Functional description
  	Shows text of a stored procedure given a name.
  	or lists procedures if no argument.
 	 Since procedures may reference each other, we will create all
  	dummy procedures of the correct name, then alter these to their
  	correct form.
       Add the parameter names when these procedures are created.

 	 procname -- Name of procedure to investigate }

procedure TBS_ADOOracleExtract.ListProcs(ProcedureName : String; AlterOnly : Boolean);
const
  CreateProcedureStr1 = 'CREATE PROCEDURE %s ';  //do not localize
  CreateProcedureStr2 = 'BEGIN EXIT; END %s%s';  //do not localize
  ProcedureSQL =
    'SELECT * FROM RDB$PROCEDURES ' +  //do not localize
    'ORDER BY RDB$PROCEDURE_NAME';     //do not localize

  ProcedureNameSQL =
    'SELECT * FROM RDB$PROCEDURES ' +    //do not localize
    'WHERE RDB$PROCEDURE_NAME = :ProcedureName ' + //do not localize
    'ORDER BY RDB$PROCEDURE_NAME';  //do not localize

//var
  //qryProcedures : TADODataSet;
  //ProcName : String;
  //SList : TStrings;
  //Header : Boolean;
begin
  {
  Header := true;
  qryProcedures := CreateADODataSet;
  SList := TStringList.Create;
  try
  //  First the dummy procedures
    create the procedures with their parameters
    if ProcedureName = '' then
      qryProcedures.SQL.Text := ProcedureSQL
    else
    begin
      qryProcedures.SQL.Text := ProcedureNameSQL;
      qryProcedures.Params.ByName('ProcedureName').AsString := ProcedureName;
    end;
    if not AlterOnly then
    begin
      qryProcedures.ExecQuery;
      while not qryProcedures.Eof do
      begin
        if Header then
        begin
          FMetaData.Add('COMMIT WORK;');
          FMetaData.Add('SET AUTODDL OFF;');
          FMetaData.Add(Format('SET TERM %s %s', [ProcTerm, Term]));
          FMetaData.Add(Format('%s/* Stored procedures */%s', [NEWLINE, NEWLINE]));
          Header := false;
        end;
        ProcName := Trim(qryProcedures.FieldByName('RDB$PROCEDURE_NAME').AsString);
        FMetaData.Add(Format(CreateProcedureStr1, [QuoteIdentifier(FDatabase.SQLDialect,
           ProcName)]));
        GetProcedureArgs(ProcName);
        FMetaData.Add(Format(CreateProcedureStr2, [ProcTerm, NEWLINE]));
        qryProcedures.Next;
      end;
      qryProcedures.Close;
    end;

    qryProcedures.ExecQuery;
    while not qryProcedures.Eof do
    begin
      SList.Clear;
      ProcName := Trim(qryProcedures.FieldByName('RDB$PROCEDURE_NAME').AsString);
      FMetaData.Add(Format('%sALTER PROCEDURE %s ', [NEWLINE,
         QuoteIdentifier(FDatabase.SQLDialect, ProcName)]));
      GetProcedureArgs(ProcName);

      FMetaData.AddStrings(SList);
      SList.Clear;
      if not qryProcedures.FieldByName('RDB$PROCEDURE_SOURCE').IsNull then
      begin
        SList.Text := qryProcedures.FieldByName('RDB$PROCEDURE_SOURCE').AsString;
        while (Slist.Count > 0) and (Trim(SList[0]) = '') do
          SList.Delete(0);
      end;
      if not AlterOnly then
        SList.Add(Format(' %s%s', [ProcTerm, NEWLINE]));
      FMetaData.AddStrings(SList);
      qryProcedures.Next;
    end;

// This query gets the procedure name and the source.  We then nest a query
//   to retrieve the parameters. Alter is used, because the procedures are
//   already there

    if not Header then
    begin
      FMetaData.Add(Format('SET TERM %s %s', [Term, ProcTerm]));
      FMetaData.Add('COMMIT WORK;');
      FMetaData.Add('SET AUTODDL ON;');
    end;
  finally
    qryProcedures.Free;
    SList.Free;
  end;
  }
end;

{            	  ListAllTables
  Functional description
  	Extract the names of all user tables from
 	 rdb$relations.  Filter SQL tables by
  	security class after we fetch them
  	Parameters:  flag -- 0, get all tables }


{	 ListAllTriggers
  Functional description
  	Lists triggers in general on non-system
  	tables with sql source only. }

procedure TBS_ADOOracleExtract.ListTriggers(ObjectName : String; ExtractType : TBS_ExtractType);
const
{ Query gets the trigger info for non-system triggers with
   source that are not part of an SQL constraint }

  TriggerSQL =
    'SELECT * FROM RDB$TRIGGERS TRG JOIN RDB$RELATIONS REL ON ' +
    '  TRG.RDB$RELATION_NAME = REL.RDB$RELATION_NAME ' +
    'WHERE ' +
    ' (REL.RDB$SYSTEM_FLAG <> 1 OR REL.RDB$SYSTEM_FLAG IS NULL) AND ' +
    '  NOT EXISTS (SELECT * FROM RDB$CHECK_CONSTRAINTS CHK WHERE ' +
    '     TRG.RDB$TRIGGER_NAME = CHK.RDB$TRIGGER_NAME) ' +
    'ORDER BY TRG.RDB$RELATION_NAME, TRG.RDB$TRIGGER_TYPE, ' +
    '    TRG.RDB$TRIGGER_SEQUENCE, TRG.RDB$TRIGGER_NAME';

  TriggerNameSQL =
    'SELECT * FROM RDB$TRIGGERS TRG JOIN RDB$RELATIONS REL ON ' +
    '  TRG.RDB$RELATION_NAME = REL.RDB$RELATION_NAME ' +
    'WHERE ' +
    ' REL.RDB$RELATION_NAME = :TableName AND ' +
    ' (REL.RDB$SYSTEM_FLAG <> 1 OR REL.RDB$SYSTEM_FLAG IS NULL) AND ' +
    '  NOT EXISTS (SELECT * FROM RDB$CHECK_CONSTRAINTS CHK WHERE ' +
    '     TRG.RDB$TRIGGER_NAME = CHK.RDB$TRIGGER_NAME) ' +
    'ORDER BY TRG.RDB$RELATION_NAME, TRG.RDB$TRIGGER_TYPE, ' +
    '    TRG.RDB$TRIGGER_SEQUENCE, TRG.RDB$TRIGGER_NAME';

  TriggerByNameSQL =
    'SELECT * FROM RDB$TRIGGERS TRG JOIN RDB$RELATIONS REL ON ' +
    '  TRG.RDB$RELATION_NAME = REL.RDB$RELATION_NAME ' +
    'WHERE ' +
    ' TRG.RDB$TRIGGER_NAME = :TriggerName AND ' +
    ' (REL.RDB$SYSTEM_FLAG <> 1 OR REL.RDB$SYSTEM_FLAG IS NULL) AND ' +
    '  NOT EXISTS (SELECT * FROM RDB$CHECK_CONSTRAINTS CHK WHERE ' +
    '     TRG.RDB$TRIGGER_NAME = CHK.RDB$TRIGGER_NAME) ' +
    'ORDER BY TRG.RDB$RELATION_NAME, TRG.RDB$TRIGGER_TYPE, ' +
    '    TRG.RDB$TRIGGER_SEQUENCE, TRG.RDB$TRIGGER_NAME';

{var
  Header : Boolean;
  TriggerName, RelationName, InActive: String;
  qryTriggers : TADODataSet;
  SList : TStrings;}
begin
  {Header := true;
  SList := TStringList.Create;
  qryTriggers := CreateADODataSet;
  try
    if ObjectName = '' then
      qryTriggers.SQL.Text := TriggerSQL
    else
    begin
      if ExtractType = extTable then
      begin
        qryTriggers.SQL.Text := TriggerNameSQL;
        qryTriggers.Params.ByName('TableName').AsString := ObjectName;
      end
      else
      begin
        qryTriggers.SQL.Text := TriggerByNameSQL;
        qryTriggers.Params.ByName('TriggerName').AsString := ObjectName;
      end;
    end;
    qryTriggers.ExecQuery;
    while not qryTriggers.Eof do
    begin
      SList.Clear;
      if Header then
      begin
        FMetaData.Add(Format('SET TERM %s %s%s', [Procterm, Term, NEWLINE]));
        FMetaData.Add(Format('%s/* Triggers only will work for SQL triggers */%s',
		       [NEWLINE, NEWLINE]));
        Header := false;
      end;
      TriggerName := qryTriggers.FieldByName('RDB$TRIGGER_NAME').AsString;
      RelationName := qryTriggers.FieldByName('RDB$RELATION_NAME').AsString;
      if qryTriggers.FieldByName('RDB$TRIGGER_INACTIVE').IsNull then
        InActive := 'INACTIVE'
      else
        if qryTriggers.FieldByName('RDB$TRIGGER_INACTIVE').AsInteger = 1 then
          InActive := 'INACTIVE'
        else
          InActive := 'ACTIVE';

      if qryTriggers.FieldByName('RDB$FLAGS').AsInteger <> 1 then
        SList.Add('/* ');

      SList.Add(Format('CREATE TRIGGER %s FOR %s %s%s %s POSITION %d',
	        [QuoteIdentifier(FDatabase.SQLDialect, TriggerName),
           QuoteIdentifier(FDatabase.SQLDialect, RelationName),
           NEWLINE, InActive,
           TriggerTypes[qryTriggers.FieldByName('RDB$TRIGGER_TYPE').AsInteger],
           qryTriggers.FieldByName('RDB$TRIGGER_SEQUENCE').AsInteger]));
      if not qryTriggers.FieldByName('RDB$TRIGGER_SOURCE').IsNull then
        SList.Text := SList.Text +
              qryTriggers.FieldByName('RDB$TRIGGER_SOURCE').AsString;
      SList.Add(' ' + ProcTerm + NEWLINE);
      if qryTriggers.FieldByName('RDB$FLAGS').AsInteger <> 1 then
        SList.Add(' */');
      FMetaData.AddStrings(SList);
      qryTriggers.Next;
    end;
    if not Header then
    begin
      FMetaData.Add('COMMIT WORK ' + ProcTerm);
      FMetaData.Add('SET TERM ' + Term + ProcTerm);
    end;
  finally
    qryTriggers.Free;
    SList.Free;
  end;}
end;


{	               ListCheck
  Functional description
 	  List check constraints for all objects to allow forward references }

procedure TBS_ADOOracleExtract.ListCheck(ObjectName : String; ExtractType : TBS_ExtractType);
const
{ Query gets the check clauses for triggers stored for check constraints }
  CheckSQL =
    'SELECT * FROM RDB$TRIGGERS TRG JOIN RDB$CHECK_CONSTRAINTS CHK ON ' +
    '  TRG.RDB$TRIGGER_NAME = CHK.RDB$TRIGGER_NAME ' +
    'WHERE ' +
    '  TRG.RDB$TRIGGER_TYPE = 1 AND ' +
    '  EXISTS (SELECT RDB$CONSTRAINT_NAME FROM RDB$RELATION_CONSTRAINTS RELC WHERE ' +
    '    CHK.RDB$CONSTRAINT_NAME = RELC.RDB$CONSTRAINT_NAME) ' +
    'ORDER BY CHK.RDB$CONSTRAINT_NAME';

  CheckNameSQL =
    'SELECT * FROM RDB$TRIGGERS TRG JOIN RDB$CHECK_CONSTRAINTS CHK ON ' +
    '  TRG.RDB$TRIGGER_NAME = CHK.RDB$TRIGGER_NAME ' +
    'WHERE ' +
    '  TRG.RDB$RELATION_NAME = :TableName AND ' +
    '  TRG.RDB$TRIGGER_TYPE = 1 AND ' +
    '  EXISTS (SELECT RDB$CONSTRAINT_NAME FROM RDB$RELATION_CONSTRAINTS RELC WHERE ' +
    '    CHK.RDB$CONSTRAINT_NAME = RELC.RDB$CONSTRAINT_NAME) ' +
    'ORDER BY CHK.RDB$CONSTRAINT_NAME';

  CheckByNameSQL =
    'SELECT * FROM RDB$TRIGGERS TRG JOIN RDB$CHECK_CONSTRAINTS CHK ON ' +
    '  TRG.RDB$TRIGGER_NAME = CHK.RDB$TRIGGER_NAME ' +
    'WHERE ' +
    '  TRG.RDB$TRIGGER_NAME = :TriggerName AND ' +
    '  TRG.RDB$TRIGGER_TYPE = 1 AND ' +
    '  EXISTS (SELECT RDB$CONSTRAINT_NAME FROM RDB$RELATION_CONSTRAINTS RELC WHERE ' +
    '    CHK.RDB$CONSTRAINT_NAME = RELC.RDB$CONSTRAINT_NAME) ' +
    'ORDER BY CHK.RDB$CONSTRAINT_NAME';

{var
  qryChecks : TADODataSet;
  SList : TStrings;
  RelationName : String;}
begin
  {qryChecks := CreateADODataSet;
  SList := TStringList.Create;
  try
    if ObjectName = '' then
      qryChecks.SQL.Text := CheckSQL
    else
      if ExtractType = extTable then
      begin
        qryChecks.SQL.Text := CheckNameSQL;
        qryChecks.Params.ByName('TableName').AsString := ObjectName;
      end
      else
      begin
        qryChecks.SQL.Text := CheckByNameSQL;
        qryChecks.Params.ByName('TriggerName').AsString := ObjectName;
      end;
    qryChecks.ExecQuery;
    while not qryChecks.Eof do
    begin
      SList.Clear;
      RelationName := qryChecks.FieldByName('RDB$RELATION_NAME').AsString;
      SList.Add(Format('ALTER TABLE %s ADD',
		    [QuoteIdentifier(FDatabase.SQLDialect, RelationName)]));
      if Pos('INTEG', qryChecks.FieldByName('RDB$CONSTRAINT_NAME').AsString) <> 1 then
        SList.Add(Format('%sCONSTRAINT %s ', [TAB,
          QuoteIdentifier(FDatabase.SQLDialect, qryChecks.FieldByName('RDB$CONSTRAINT_NAME').AsString)]));

      if not qryChecks.FieldByName('RDB$TRIGGER_SOURCE').IsNull then
        SList.Text := SList.Text + qryChecks.FieldByName('RDB$TRIGGER_SOURCE').AsString;

      SList.Strings[SList.Count - 1] := SList.Strings[SList.Count - 1] + (Term) + NEWLINE;
      FMetaData.AddStrings(SList);
      qryChecks.Next;
    end;
  finally
    qryChecks.Free;
    SList.Free;
  end;}
end;




{          ListException
 Functional description
   List all exceptions defined in the database
   Parameters:  none }

procedure TBS_ADOOracleExtract.ListException(ExceptionName : String = '');
const
  ExceptionSQL =
    'select * from RDB$EXCEPTIONS ' +
    'ORDER BY RDB$EXCEPTION_NAME';

  ExceptionNameSQL =
    'select * from RDB$EXCEPTIONS ' +
    'WHERE RDB$EXCEPTION_NAME = :ExceptionName ' +
    'ORDER BY RDB$EXCEPTION_NAME';

//var  First : Boolean;
//  qryException : TADODataSet;
begin
  {First := true;
  qryException := CreateADODataSet;
  try
    if ExceptionName = '' then
      qryException.SQL.Text := ExceptionSQL
    else
    begin
      qryException.SQL.Text := ExceptionNameSQL;
      qryException.Params.ByName('ExceptionName').AsString := ExceptionName;
    end;

    qryException.ExecQuery;
    while not qryException.Eof do
    begin
      if First then
      begin
        FMetaData.Add('');
        FMetaData.Add('/*  Exceptions */');
        FMetaData.Add('');
        First := false;
      end; //end_if

      FMetaData.Add(Format('CREATE EXCEPTION %s %s%s',
        [QuoteIdentifier(FDatabase.SQLDialect, qryException.FieldByName('RDB$EXCEPTION_NAME').AsString),
        QuotedStr(qryException.FieldByName('RDB$MESSAGE').AsString), Term]));
      qryException.Next;
    end;
  finally
    qryException.Free;
  end;
}
end;

{              ListFilters

 Functional description
  List all blob filters

  Parameters:  none
  Results in
  DECLARE FILTER <fname> INPUT_TYPE <blob_sub_type> OUTPUT_TYPE <blob_subtype>
      ENTRY_POINT <string> MODULE_NAME <string> }

procedure TBS_ADOOracleExtract.ListFilters(FilterName : String = '');
const
  FiltersSQL =
    'SELECT * FROM RDB$FILTERS ' +
    'ORDER BY RDB$FUNCTION_NAME';
  FilterNameSQL =
    'SELECT * FROM RDB$FILTERS ' +
    'WHERE RDB$FUNCTION_NAME = :FunctionName ' +
    'ORDER BY RDB$FUNCTION_NAME';

//var  First : Boolean;
//  qryFilters : TADODataSet;
begin
{  First := true;
  qryFilters := CreateADODataSet;
  try
    if FilterName = '' then
      qryFilters.SQL.Text := FiltersSQL
    else
    begin
      qryFilters.SQL.Text := FilterNameSQL;
      qryFilters.Params.ByName('FunctionName').AsString := FilterName;
    end;
    qryFilters.ExecQuery;
    while not qryFilters.Eof do
    begin
      if First then
      begin
        FMetaData.Add('');
        FMetaData.Add('/*  BLOB Filter declarations */');
        FMetaData.Add('');
        First := false;
      end; //end_if

      FMetaData.Add(Format('DECLARE FILTER %s INPUT_TYPE %d OUTPUT_TYPE %d',
        [qryFilters.FieldByName('RDB$FUNCTION_NAME').AsString,
         qryFilters.FieldByName('RDB$INPUT_SUB_TYPE').AsInteger,
         qryFilters.FieldByName('RDB$OUTPUT_SUB_TYPE').AsInteger]));
      FMetaData.Add(Format('%sENTRY_POINT ''%s'' MODULE_NAME ''%s''%s%',
        [TAB, qryFilters.FieldByName('RDB$ENTRYPOINT').AsString,
         qryFilters.FieldByName('RDB$MODULE_NAME').AsString, Term]));
      FMetaData.Add('');

      qryFilters.Next;
    end;

  finally
    qryFilters.Free;
  end;
}
end;




{    ListFunctions
 Functional description
   List all external functions

   Parameters:  none
  Results in
  DECLARE EXTERNAL FUNCTION function_name
                CHAR [256] , INTEGER, ....
                RETURNS INTEGER BY VALUE
                ENTRY_POINT entrypoint MODULE_NAME module; }

procedure TBS_ADOOracleExtract.ListFunctions(FunctionName : String = '');
const
  FunctionSQL =
    'SELECT * FROM RDB$FUNCTIONS ' +
    'ORDER BY RDB$FUNCTION_NAME';

  FunctionNameSQL =
    'SELECT * FROM RDB$FUNCTIONS ' +
    'WHERE RDB$FUNCTION_NAME = :FunctionName ' +
    'ORDER BY RDB$FUNCTION_NAME';

  FunctionArgsSQL =
    'SELECT * FROM RDB$FUNCTION_ARGUMENTS ' +
    'WHERE ' +
    '  :FUNCTION_NAME = RDB$FUNCTION_NAME ' +
    'ORDER BY RDB$ARGUMENT_POSITION';

  FuncArgsPosSQL =
    'SELECT * FROM RDB$FUNCTION_ARGUMENTS ' +
    'WHERE ' +
    '  RDB$FUNCTION_NAME = :RDB$FUNCTION_NAME AND ' +
    '  RDB$ARGUMENT_POSITION = :RDB$ARGUMENT_POSITION';

  CharSetSQL =
    'SELECT * FROM RDB$CHARACTER_SETS ' +
    'WHERE ' +
    '  RDB$CHARACTER_SET_ID = :CHARACTER_SET_ID';

var
//  qryFunctions, qryFuncArgs, qryCharSets, qryFuncPos : TADODataSet;
  First, FirstArg, DidCharset, PrecisionKnown : Boolean;
  ReturnBuffer, TypeBuffer, Line : String;
  i, FieldType : Integer;
begin
  First := true;
 { qryFunctions := CreateADODataSet;
  qryFuncArgs := CreateADODataSet;
  qryFuncPos := CreateADODataSet;
  qryCharSets := CreateADODataSet;
  try
    if FunctionName = '' then
      qryFunctions.SQL.Text := FunctionSQL
    else
    begin
      qryFunctions.SQL.Text := FunctionNameSQL;
      qryFunctions.Params.ByName('FunctionName').AsString := FunctionName;
    end;
    qryFuncArgs.SQL.Text := FunctionArgsSQL;
    qryFuncPos.SQL.Text := FuncArgsPosSQL;
    qryCharSets.SQL.Text := CharSetSQL;
    qryFunctions.ExecQuery;
    while not qryFunctions.Eof do
    begin
      if First then
      begin
        FMEtaData.Add(Format('%s/*  External Function declarations */%s',
          [NEWLINE, NEWLINE]));
        First := false;
      end; //end_if
      // Start new function declaration
      FMetaData.Add(Format('DECLARE EXTERNAL FUNCTION %s',
        [qryFunctions.FieldByName('RDB$FUNCTION_NAME').AsString]));
      Line := '';

      FirstArg := true;
      qryFuncArgs.Params.ByName('FUNCTION_NAME').AsString :=
         qryFunctions.FieldByName('RDB$FUNCTION_NAME').AsString;

      qryFuncArgs.ExecQuery;
      while not qryFuncArgs.Eof do
      begin
        // Find parameter type
        i := 0;
        FieldType := qryFuncArgs.FieldByName('RDB$FIELD_TYPE').AsInteger;
        while FieldType <> ColumnTypes[i].SQLType do
          Inc(i);

        // Print length where appropriate
        if FieldType in [ bst_text, bst_varying, bst_cstring] then
        begin
          DidCharset := false;

          qryCharSets.Params.ByName('CHARACTER_SET_ID').AsInteger :=
             qryFuncArgs.FieldByName('RDB$CHARACTER_SET_ID').AsInteger;
          qryCharSets.ExecQuery;
          while not qryCharSets.Eof do
          begin
            DidCharset := true;
            TypeBuffer := Format('%s(%d) CHARACTER SET %s',
              [ColumnTypes[i].TypeName,
               qryFuncArgs.FieldByName('RDB$FIELD_LENGTH').AsInteger div
               Max(1,qryCharSets.FieldByName('RDB$BYTES_PER_CHARACTER').AsInteger),
               qryCharSets.FieldByName('RDB$CHARACTER_SET_NAME').AsString]);
            qryCharSets.Next;
          end;
          qryCharSets.Close;
          if not DidCharset then
            TypeBuffer := Format('%s(%d)', [ColumnTypes[i].TypeName,
              qryFuncArgs.FieldByName('RDB$FIELD_LENGTH').AsInteger]);
        end //end_if
        else
        begin
          PrecisionKnown := false;
          if (FDatabaseInfo.ODSMajorVersion >= ODS_VERSION10) and
              (FieldType in [bst_SmallInt, bst_long, bst_int64]) then
          begin
            qryFuncPos.Params.ByName('RDB$FUNCTION_NAME').AsString :=
              qryFuncArgs.FieldByName('RDB$FUNCTION_NAME').AsString;
            qryFuncPos.Params.ByName('RDB$ARGUMENT_POSITION').AsInteger :=
              qryFuncArgs.FieldByName('RDB$ARGUMENT_POSITION').AsInteger;

            qryFuncPos.ExecQuery;
            while not qryFuncPos.Eof do
            begin
              // We are ODS >= 10 and could be any Dialect
              if not qryFuncPos.FieldByName('RDB$FIELD_PRECISION').IsNull then
              begin
                // We are Dialect >=3 since FIELD_PRECISION is non-NULL
                if (qryFuncPos.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger > 0) and
                    (qryFuncPos.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger <= MAX_INTSUBTYPES) then
                begin
                  TypeBuffer := Format('%s(%d, %d)',
                    [IntegralSubtypes[qryFuncPos.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger],
                     qryFuncPos.FieldByName('RDB$FIELD_PRECISION').AsInteger,
                     -qryFuncPos.FieldByName('RDB$FIELD_SCALE').AsInteger] );
                  PrecisionKnown := true;
                end; //end_if
              end; // if field_precision is not null
              qryFuncPos.Next;
            end;
            qryFuncPos.Close;
          end; // if major_ods >= ods_version10 &&
          if not PrecisionKnown then
          begin
            // Take a stab at numerics and decimals
            if (FieldType = bst_SmallInt) and
                (qryFuncArgs.FieldByName('RDB$FIELD_SCALE').AsInteger < 0) then
              TypeBuffer := Format('NUMERIC(4, %d)',
                [-qryFuncArgs.FieldByName('RDB$FIELD_SCALE').AsInteger])
            else
              if (FieldType = bst_long) and
                  (qryFuncArgs.FieldByName('RDB$FIELD_SCALE').AsInteger < 0) then
                TypeBuffer := Format('NUMERIC(9, %d)',
                  [-qryFuncArgs.FieldByName('RDB$FIELD_SCALE').AsInteger])
              else
                if (FieldType = bst_double) and
                    (qryFuncArgs.FieldByName('RDB$FIELD_SCALE').AsInteger < 0) then
                  TypeBuffer := Format('NUMERIC(15, %d)',
                      [-qryFuncArgs.FieldByName('RDB$FIELD_SCALE').AsInteger])
                else
                  TypeBuffer := ColumnTypes[i].TypeName;
          end; // if  not PrecisionKnown
        end; // if FCHAR or VARCHAR or CSTRING ... else

        if qryFunctions.FieldByName('RDB$RETURN_ARGUMENT').AsInteger =
               qryFuncArgs.FieldByName('RDB$ARGUMENT_POSITION').AsInteger then
        begin
          ReturnBuffer := 'RETURNS ' + TypeBuffer;
          if qryFuncArgs.FieldByName('RDB$MECHANISM').AsInteger = 0 then
            ReturnBuffer := ReturnBuffer + ' BY VALUE ';
          if qryFuncArgs.FieldByName('RDB$MECHANISM').AsInteger < 0 then
            ReturnBuffer := ReturnBuffer + ' FREE_IT';
        end
        else
        begin
          // First arg needs no comma
          if FirstArg then
          begin
            Line := Line + TypeBuffer;
            FirstArg := false;
          end
          else
            Line := Line + ', ' + TypeBuffer;
        end; //end_else
        qryFuncArgs.Next;
      end;
      qryFuncArgs.Close;

      FMetaData.Add(Line);
      FMetaData.Add(ReturnBuffer);
      FMetaData.Add(Format('ENTRY_POINT ''%s'' MODULE_NAME ''%s''%s%s%s',
        [qryFunctions.FieldByName('RDB$ENTRYPOINT').AsString,
         qryFunctions.FieldByName('RDB$MODULE_NAME').AsString,
         Term, NEWLINE, NEWLINE]));

      qryFunctions.Next;
    end;
  finally
    qryFunctions.Free;
    qryFuncArgs.Free;
    qryCharSets.Free;
    qryFuncPos.Free;
  end;
  }
end;



{    ListViews
 Functional description
   Show text of views.
   Use a SQL query to get the info and print it.
   Note: This should also contain check option }

procedure TBS_ADOOracleExtract.ListViews(ViewName : String);
const
  ViewSQL =
    'SELECT RDB$RELATION_NAME, RDB$OWNER_NAME, RDB$VIEW_SOURCE ' +
    'FROM RDB$RELATIONS ' +
    'WHERE ' +
    '  (RDB$SYSTEM_FLAG <> 1 OR RDB$SYSTEM_FLAG IS NULL) AND ' +
    '  NOT RDB$VIEW_BLR IS NULL AND ' +
    '  RDB$FLAGS = 1 ' +
    'ORDER BY RDB$RELATION_ID';

  ViewNameSQL =
    'SELECT RDB$RELATION_NAME, RDB$OWNER_NAME, RDB$VIEW_SOURCE ' +
    'FROM RDB$RELATIONS ' +
    'WHERE ' +
    '  (RDB$SYSTEM_FLAG <> 1 OR RDB$SYSTEM_FLAG IS NULL) AND ' +
    '  NOT RDB$VIEW_BLR IS NULL AND ' +
    '  RDB$FLAGS = 1 AND ' +
    '  RDB$RELATION_NAME = :ViewName ' +
    'ORDER BY RDB$RELATION_ID';

  ColumnSQL =
    'SELECT RDB$FIELD_NAME FROM RDB$RELATION_FIELDS ' +
    'WHERE ' +
    '  RDB$RELATION_NAME = :RELATION_NAME ' +
    'ORDER BY RDB$FIELD_POSITION';

var
  //qryView, qryColumns : TADODataSet;
  SList : TStrings;
begin
  {qryView := CreateADODataSet;
  qryColumns := CreateADODataSet;
  SList := TStringList.Create;
  try
    if ViewName = '' then
      qryView.SQL.Text := ViewSQL
    else
    begin
      qryView.SQL.Text := ViewNameSQL;
      qryView.Params.ByName('ViewName').AsString := ViewName;
    end;
    qryColumns.SQL.Text := ColumnSQL;
    qryView.ExecQuery;
    while not qryView.Eof do
    begin
      SList.Add(Format('%s/* View: %s, Owner: %s */%s',
         [NEWLINE, qryView.FieldByName('RDB$RELATION_NAME').AsString,
          qryView.FieldByName('RDB$OWNER_NAME').AsString, NEWLINE]));

      SList.Add(Format('CREATE VIEW %s (', [QuoteIdentifier(FDatabase.SQLDialect,
        qryView.FieldByName('RDB$RELATION_NAME').AsString)]));

      qryColumns.Params.ByName('RELATION_NAME').AsString :=
          qryView.FieldByName('RDB$RELATION_NAME').AsString;
      qryColumns.ExecQuery;
      while not qryColumns.Eof do
      begin
        SList.Add('  ' + QuoteIdentifier(FDatabase.SQLDialect,
           qryColumns.FieldByName('RDB$FIELD_NAME').AsString));
        qryColumns.Next;
        if not qryColumns.Eof then
          SList.Strings[SList.Count - 1] := SList.Strings[SList.Count - 1] + ', ';
      end;
      qryColumns.Close;
      SList.Text := SList.Text + Format(') AS%s', [NEWLINE]);
      if not qryView.FieldByName('RDB$VIEW_SOURCE').IsNull then
        SList.Text := SList.Text + qryView.FieldByName('RDB$VIEW_SOURCE').AsString;
      SList.Text := SList.Text + Format('%s%s', [Term, NEWLINE]);
      FMetaData.AddStrings(SList);
      SList.Clear;
      qryView.Next;
    end;
  finally
    qryView.Free;
    qryColumns.Free;
    SList.Free;
  end;
  }
end;

{    PrintSet
  Functional description
     print (using ISQL_printf) the word "SET"
     if the first line of the ALTER DATABASE
     settings options. Also, add trailing
     comma for end of prior line if needed.

  uses Print_buffer, a global }

function TBS_ADOOracleExtract.PrintSet(var Used: Boolean) : String;
begin
  if not Used then
  begin
    Result := '  SET ';
    Used := true;
  end
  else
    Result := Format(', %s      ', [NEWLINE]);
end;

{
  PrintValidation
  Functional description
    This does some minor syntax adjustmet for extracting
    validation blobs and computed fields.
    if it does not start with the word CHECK
    if this is a computed field blob,look for () or insert them.
    if flag = false, this is a validation clause,
    if flag = true, this is a computed field }

function TBS_ADOOracleExtract.PrintValidation(ToValidate: String;
  flag: Boolean): String;
var
  isSQL : Boolean;
begin
  isSql := false;

  Result := '';
  ToValidate := Trim(ToValidate);

  if flag then
  begin
    if ToValidate[1] = '(' then
      IsSQL := true;
  end
  else
    if (Pos(ToValidate, 'check') = 1) or (Pos(ToValidate, 'CHECK') = 1) then
      IsSQL := TRUE;

	if not IsSQL then
  begin
    if Flag then
      Result := Result + '/* ' + ToValidate + ' */'
    else
      Result := Result + '(' + ToValidate + ')';
  end
  else
    Result := ToValidate;
end;



function TBS_ADOOracleExtract.GetFieldType(FieldType, FieldSubType, FieldScale,
  FieldSize, FieldPrec, FieldLen: Integer): String;
var
  i : Integer;
  PrecisionKnown : Boolean;
begin
  Result := '';
{  for i := Low(ColumnTypes) to High(ColumnTypes) do
    if FieldType = ColumnTypes[i].SQLType then
    begin
      PrecisionKnown := FALSE;
      if FDatabaseInfo.ODSMajorVersion >= ODS_VERSION10 then
      begin
        if FieldType in [bst_SmallInt, bst_long, bst_int64] then
        begin
          // We are ODS >= 10 and could be any Dialect
          if (FDatabaseInfo.DBSQLDialect >= 3) and
             (FieldPrec <> 0) and
             (FieldSubType > 0) and
             (FieldSubType <= MAX_INTSUBTYPES) then
          begin
            Result := Result + Format('%s(%d, %d)', [
              IntegralSubtypes [FieldSubType],
              FieldPrec,
              -1 * FieldScale]);
            PrecisionKnown := true;
          end;
        end;
      end;
      if PrecisionKnown = false then
      begin
        // Take a stab at numerics and decimals
        if (FieldType = bst_SmallInt) and
            (FieldScale < 0) then
          Result := Result + Format('NUMERIC(4, %d)',
            [-FieldScale] )
        else
          if (FieldType = bst_long) and
              (FieldScale < 0) then
            Result := Result + Format('NUMERIC(9, %d)',
              [-FieldScale] )
          else
            if (FieldType = bst_double) and
                (FieldScale  < 0) then
              Result := Result + Format('NUMERIC(15, %d)',
                [-FieldScale] )
            else
              Result := Result + ColumnTypes[i].TypeName;
      end;
      break;
    end;
  if (FieldType in [bst_text, bst_varying]) and
     (FieldSize <> 0) then
    Result := Result + Format('(%d)', [FieldSize]);
}
end;

{  S H O W _ g r a n t s
 Functional description
   Show grants for given object name
   This function is also called by extract for privileges.
     It must extract granted privileges on tables/views to users,
     - these may be compound, so put them on the same line.
   Grant execute privilege on procedures to users
   Grant various privilegs to procedures.
   All privileges may have the with_grant option set. }

procedure TBS_ADOOracleExtract.ShowGrants(MetaObject, Terminator: String);
const
  // This query only finds tables, eliminating owner privileges
  OwnerPrivSQL =
    'SELECT PRV.RDB$USER, PRV.RDB$GRANT_OPTION, PRV.RDB$FIELD_NAME, ' +
    '       PRV.RDB$USER_TYPE, PRV.RDB$PRIVILEGE ' +
    'FROM RDB$USER_PRIVILEGES PRV, RDB$RELATIONS REL ' +
    'WHERE ' +
    '  PRV.RDB$RELATION_NAME = :METAOBJECT AND ' +
    '  REL.RDB$RELATION_NAME = :METAOBJECT AND ' +
    '  PRV.RDB$PRIVILEGE <> ''M'' AND ' +
    '  REL.RDB$OWNER_NAME <> PRV.RDB$USER ' +
    'ORDER BY  PRV.RDB$USER, PRV.RDB$FIELD_NAME, PRV.RDB$GRANT_OPTION';

  ProcPrivSQL =
    'SELECT PRV.RDB$USER, PRV.RDB$GRANT_OPTION, PRV.RDB$FIELD_NAME, ' +
    '       PRV.RDB$USER_TYPE, PRV.RDB$PRIVILEGE, PRV.RDB$RELATION_NAME ' +
    'FROM RDB$USER_PRIVILEGES PRV, RDB$PROCEDURES PRC ' +
    'where ' +
    '  PRV.RDB$OBJECT_TYPE = 5 AND ' +
    '  PRV.RDB$RELATION_NAME = :METAOBJECT AND ' +
    '  PRC.RDB$PROCEDURE_NAME = :METAOBJECT AND ' +
    '  PRV.RDB$PRIVILEGE = ''X'' AND ' +
    '  PRC.RDB$OWNER_NAME <> PRV.RDB$USER ' +
    'ORDER BY PRV.RDB$USER, PRV.RDB$FIELD_NAME, PRV.RDB$GRANT_OPTION';

  RolePrivSQL =
    'SELECT * FROM RDB$USER_PRIVILEGES ' +
    'WHERE ' +
    '  RDB$OBJECT_TYPE = 13 AND ' +
    '  RDB$USER_TYPE = 8  AND ' +
    '  RDB$RELATION_NAME = :METAOBJECT AND ' +
    '  RDB$PRIVILEGE = ''M'' ' +
    'ORDER BY RDB$USER';

var
  PrevUser, PrevField,  WithOption,
  PrivString, ColString, UserString,
  FieldName, User : String;
  c : Char;
  PrevOption, PrivFlags, GrantOption : Integer;
  First, PrevFieldNull : Boolean;
  qryOwnerPriv : TADODataSet;

    //  Given a bit-vector of privileges, turn it into a
    //   string list.
  function MakePrivString(cflags : Integer) : String;
  var
    i : Integer;
  begin
    for i := Low(PrivTypes) to High(PrivTypes) do
    begin
      if (cflags and PrivTypes[i].PrivFlag) <> 0 then
      begin
        if Result <> '' then
          Result := Result + ', ';
        Result := Result + PrivTypes[i].PrivString;
      end; //end_if
    end; //end_for
  end; //end_fcn MakePrivDtring

begin
  if MetaObject = '' then
    exit;

  First := true;
  PrevOption := -1;
  PrevUser := '';
  PrivString := '';
  ColString := '';
  WithOption := '';
  PrivFlags := 0;
  PrevFieldNull := false;
  PrevField := '';

  {qryOwnerPriv := CreateADODataSet;
  try
    qryOwnerPriv.SQL.Text := OwnerPrivSQL;
    qryOwnerPriv.Params.ByName('metaobject').AsString := MetaObject;
    qryOwnerPriv.ExecQuery;
    while not qryOwnerPriv.Eof do
    begin
      // Sometimes grant options are null, sometimes 0.  Both same
      if qryOwnerPriv.FieldByName('RDB$GRANT_OPTION').IsNull then
        GrantOption := 0
      else
        GrantOption := qryOwnerPriv.FieldByName('RDB$GRANT_OPTION').AsInteger;

      if qryOwnerPriv.FieldByName('RDB$FIELD_NAME').IsNull then
        FieldName := ''
      else
        FieldName := qryOwnerPriv.FieldByName('RDB$FIELD_NAME').AsString;

      User := Trim(qryOwnerPriv.FieldByName('RDB$USER').AsString);
      // Print a new grant statement for each new user or change of option

      if ((PrevUser <> '') and (PrevUser <> User)) or
          ((Not First) and
            (PrevFieldNull <> qryOwnerPriv.FieldByName('RDB$FIELD_NAME').IsNull)) or
          ((not PrevFieldNull) and (PrevField <> FieldName)) or
          ((PrevOption <> -1) and (PrevOption <> GrantOption)) then
      begin
        PrivString := MakePrivString(PrivFlags);

        First := false;
        FMetaData.Add(Format('GRANT %s%s ON %s TO %s%s%s', [PrivString,
          ColString, QuoteIdentifier(FDatabase.SQLDialect, MetaObject),
          UserString, WithOption, Terminator]));
        // re-initialize strings

        PrivString := '';
        WithOption := '';
        ColString := '';
        PrivFlags := 0;
      end; //end_if

      PrevUser := User;
      PrevOption := GrantOption;
      PrevFieldNull := qryOwnerPriv.FieldByName('RDB$FIELD_NAME').IsNull;
      PrevField := FieldName;

      case qryOwnerPriv.FieldByName('RDB$USER_TYPE').AsInteger of
        obj_relation,
        obj_view,
        obj_trigger,
        obj_procedure,
        obj_sql_role:
          UserString := QuoteIdentifier(FDatabase.SQLDialect, User);
        else
          UserString := User;
      end; //end_case

      case qryOwnerPriv.FieldByName('RDB$USER_TYPE').AsInteger of
        obj_view :
          UserString := 'VIEW ' + UserString;
        obj_trigger :
          UserString := 'TRIGGER '+ UserString;
        obj_procedure :
          UserString := 'PROCEDURE ' + UserString;
      end; //end_case

      c := qryOwnerPriv.FieldByName('RDB$PRIVILEGE').AsString[1];

      case c of
        'S' : PrivFlags := PrivFlags or priv_SELECT;
        'I' : PrivFlags := PrivFlags or priv_INSERT;
        'U' : PrivFlags := PrivFlags or priv_UPDATE;
        'D' : PrivFlags := PrivFlags or priv_DELETE;
        'R' : PrivFlags := PrivFlags or priv_REFERENCES;
        'X' : ;
          // Execute should not be here -- special handling below
        else
          PrivFlags := PrivFlags or priv_UNKNOWN;
      end; //end_switch

      // Column level privileges for update only

      if FieldName = '' then
        ColString := ''
      else
        ColString := Format(' (%s)', [QuoteIdentifier(FDatabase.SQLDialect, FieldName)]);

      if GrantOption <> 0 then
        WithOption := ' WITH GRANT OPTION';

      qryOwnerPriv.Next;
    end;
    // Print last case if there was anything to print
    if PrevOption <> -1 then
    begin
      PrivString := MakePrivString(PrivFlags);
      First := false;
      FMetaData.Add(Format('GRANT %s%s ON %s TO %s%s%s', [PrivString,
        ColString, QuoteIdentifier(FDatabase.SQLDialect, MetaObject),
        UserString, WithOption, Terminator]));
      //re-initialize strings
    end; //end_if
    qryOwnerPriv.Close;

    if First then
    begin
     // Part two is for stored procedures only
      qryOwnerPriv.SQL.Text := ProcPrivSQL;
      qryOwnerPriv.Params.ByName('metaobject').AsString := MetaObject;
      qryOwnerPriv.ExecQuery;
      while not qryOwnerPriv.Eof do
      begin
        First := false;
        User := Trim(qryOwnerPriv.FieldByName('RDB$USER').AsString);

        case qryOwnerPriv.FieldByName('RDB$USER_TYPE').AsInteger of
          obj_relation,
          obj_view,
          obj_trigger,
          obj_procedure,
          obj_sql_role:
            UserString := QuoteIdentifier(FDatabase.SQLDialect, User);
          else
            UserString := User;
        end; //end_case
        case qryOwnerPriv.FieldByName('RDB$USER_TYPE').AsInteger of
          obj_view :
            UserString := 'VIEW ' + UserString;
          obj_trigger :
            UserString := 'TRIGGER '+ UserString;
          obj_procedure :
            UserString := 'PROCEDURE ' + UserString;
        end; //end_case

        if qryOwnerPriv.FieldByName('RDB$GRANT_OPTION').AsInteger = 1 then
          WithOption := ' WITH GRANT OPTION'
        else
          WithOption := '';

        FMetaData.Add(Format('GRANT EXECUTE ON PROCEDURE %s TO %s%s%s',
          [QuoteIdentifier(FDatabase.SQLDialect, MetaObject), UserString,
           WithOption, terminator]));

        qryOwnerPriv.Next;
      end;
      qryOwnerPriv.Close;
    end;
    if First then
    begin
      qryOwnerPriv.SQL.Text := RolePrivSQL;
      qryOwnerPriv.Params.ByName('metaobject').AsString := MetaObject;
      qryOwnerPriv.ExecQuery;
      while not qryOwnerPriv.Eof do
      begin
        if qryOwnerPriv.FieldByName('RDB$GRANT_OPTION').AsInteger = 1 then
          WithOption := ' WITH ADMIN OPTION'
        else
          WithOption := '';

        FMetaData.Add(Format('GRANT %s TO %s%s%s',
          [QuoteIdentifier(FDatabase.SQLDialect, qryOwnerPriv.FieldByName('RDB$RELATION_NAME').AsString),
           qryOwnerPriv.FieldByName('RDB$USER_NAME').AsString,
           WithOption, terminator]));

        qryOwnerPriv.Next;
      end;
    end;
    qryOwnerPriv.Close;
  finally
    qryOwnerPriv.Free;
  end;
  }
end;

{	  ShowGrantRoles
  Functional description
   	Show grants for given role name
   	This function is also called by extract for privileges.
   	All membership privilege may have the with_admin option set. }

procedure TBS_ADOOracleExtract.ShowGrantRoles(Terminator: String);
const
  RoleSQL =
    'SELECT RDB$USER, RDB$GRANT_OPTION, RDB$RELATION_NAME ' +
    'FROM RDB$USER_PRIVILEGES ' +
    'WHERE ' +
    '  RDB$OBJECT_TYPE = %d AND ' +
    '  RDB$USER_TYPE = %d AND ' +
    '  RDB$PRIVILEGE = ''M'' ' +
    'ORDER BY  RDB$RELATION_NAME, RDB$USER';

var
  WithOption, UserString : String;
  //qryRole : TADODataSet;

begin
  {qryRole := CreateADODataSet;
  try
    qryRole.SQL.Text := Format(RoleSQL, [obj_sql_role, obj_user]);
    qryRole.ExecQuery;
    while not qryRole.Eof do
    begin
      UserString := Trim(qryRole.FieldByName('RDB$USER').AsString);

      if (not qryRole.FieldByName('RDB$GRANT_OPTION').IsNull) and
         (qryRole.FieldByName('RDB$GRANT_OPTION').AsInteger = 1) then
        WithOption := ' WITH ADMIN OPTION'
      else
        WithOption := '';
      FMetaData.Add(Format('GRANT %s TO %s%s%s%s',
        [ QuoteIdentifier(FDatabase.SQLDialect, qryRole.FieldByName('RDB$RELATION_NAME').AsString),
         UserString, WithOption, Terminator, NEWLINE]));

      qryRole.Next;
    end;
  finally
    qryRole.Free;
  end;}
end;

{	            GetProcedureArgs
  Functional description
 	 This function extract the procedure parameters and adds it to the
 	 extract file }

procedure TBS_ADOOracleExtract.GetProcedureArgs(Proc: String);
const
// query to retrieve the input parameters.
  ProcHeaderSQL =
    'SELECT * ' +
    ' FROM RDB$PROCEDURE_PARAMETERS PRM JOIN RDB$FIELDS FLD ON ' +
    ' PRM.RDB$FIELD_SOURCE = FLD.RDB$FIELD_NAME ' +
    'WHERE ' +
    '    PRM.RDB$PROCEDURE_NAME = :PROCNAME AND ' +
    '    PRM.RDB$PARAMETER_TYPE = :Input ' +
    'ORDER BY PRM.RDB$PARAMETER_NUMBER';

var
  FirstTime, PrecisionKnown : Boolean;
  Line : String;
  //qryHeader : TADODataSet;

  {function FormatParamStr : String;
  var
    i, CollationID, CharSetID : Integer;
  begin
    Result := Format('  %s ', [QuoteIdentifier(FDatabase.SQLDialect, qryHeader.FieldByName('RDB$PARAMETER_NAME').AsString)]);
    for i := Low(ColumnTypes) to High(ColumnTypes) do
      if qryHeader.FieldByName('RDB$FIELD_TYPE').AsInteger = ColumnTypes[i].SQLType then
      begin
        PrecisionKnown := FALSE;
        if FDatabaseInfo.ODSMajorVersion >= ODS_VERSION10 then
        begin
          if qryHeader.FieldByName('RDB$FIELD_TYPE').AsInteger in [bst_SmallInt, bst_long, bst_int64] then
          begin
            // We are ODS >= 10 and could be any Dialect
            if (FDatabaseInfo.DBSQLDialect >= 3) and
               (not qryHeader.FieldByName('RDB$FIELD_PRECISION').IsNull) and
               (qryHeader.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger > 0) and
               (qryHeader.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger <= MAX_INTSUBTYPES) then
            begin
              Result := Result + Format('%s(%d, %d)', [
                IntegralSubtypes [qryHeader.FieldByName('RDB$FIELD_SUB_TYPE').AsInteger],
                qryHeader.FieldByName('RDB$FIELD_PRECISION').AsInteger,
                -1 * qryHeader.FieldByName('RDB$FIELD_SCALE').AsInteger]);
              PrecisionKnown := true;
            end;
          end;
        end;
        if PrecisionKnown = false then
        begin
          // Take a stab at numerics and decimals
          if (qryHeader.FieldByName('RDB$FIELD_TYPE').AsInteger = bst_SmallInt) and
              (qryHeader.FieldByName('RDB$FIELD_SCALE').AsInteger < 0) then
            Result := Result + Format('NUMERIC(4, %d)',
              [-qryHeader.FieldByName('RDB$FIELD_SCALE').AsInteger] )
          else
            if (qryHeader.FieldByName('RDB$FIELD_TYPE').AsInteger = bst_long) and
                (qryHeader.FieldByName('RDB$FIELD_SCALE').AsInteger < 0) then
              Result := Result + Format('NUMERIC(9, %d)',
                [-qryHeader.FieldByName('RDB$FIELD_SCALE').AsInteger] )
            else
              if (qryHeader.FieldByName('RDB$FIELD_TYPE').AsInteger = bst_double) and
                  (qryHeader.FieldByName('RDB$FIELD_SCALE').AsInteger  < 0) then
                Result := Result + Format('NUMERIC(15, %d)',
                  [-qryHeader.FieldByName('RDB$FIELD_SCALE').AsInteger] )
              else
                Result := Result + ColumnTypes[i].TypeName;
        end;
        break;
      end;
    if qryHeader.FieldByName('RDB$FIELD_TYPE').AsInteger in [bst_text, bst_varying] then
      Result := Result + Format('(%d)', [GetFieldLength(qryHeader)]);

    // Show international character sets and collations

    if (not qryHeader.FieldByName('RDB$COLLATION_ID').IsNull) or
       (not qryHeader.FieldByName('RDB$CHARACTER_SET_ID').IsNull) then
    begin
      if qryHeader.FieldByName('RDB$COLLATION_ID').IsNull then
        CollationId := 0
      else
        CollationId := qryHeader.FieldByName('RDB$COLLATION_ID').AsInteger;

      if qryHeader.FieldByName('RDB$CHARACTER_SET_ID').IsNull then
        CharSetId := 0
      else
        CharSetId := qryHeader.FieldByName('RDB$CHARACTER_SET_ID').AsInteger;

      Result := Result + GetCharacterSets(CharSetId, CollationId, false);
    end;
  end;
}
begin
  FirstTime := true;
  {qryHeader := CreateADODataSet;
  try
    qryHeader.SQL.Text := ProcHeaderSQL;
    qryHeader.Params.ByName('procname').AsString := Proc;
    qryHeader.Params.ByName('Input').AsInteger := 0;
    qryHeader.ExecQuery;
    while not qryHeader.Eof do
    begin
      if FirstTime then
      begin
        FirstTime := false;
        FMetaData.Add('(');
      end;

      Line := FormatParamStr;

      qryHeader.Next;
      if not qryHeader.Eof then
        Line := Line + ',';
      FMetaData.Add(Line);
    end;

    // If there was at least one param, close parens
    if not FirstTime then
    begin
      FMetaData.Add( ')');
    end;

    FirstTime := true;
    qryHeader.Close;
    qryHeader.Params.ByName('Input').AsInteger := 1;
    qryHeader.ExecQuery;

    while not qryHeader.Eof do
    begin
      if FirstTime then
      begin
        FirstTime := false;
        FMetaData.Add('RETURNS' + NEWLINE + '(');
      end;

      Line := FormatParamStr;

      qryHeader.Next;
      if not qryHeader.Eof then
        Line := Line + ',';
      FMetaData.Add(Line);
    end;

    // If there was at least one param, close parens
    if not FirstTime then
    begin
      FMetaData.Add( ')');
    end;

    FMetaData.Add('AS');
  finally
    qryHeader.Free;
  end;
  }
end;

procedure TBS_ADOOracleExtract.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  {if (AComponent = FDatabase) and (Operation = opRemove) then
    FDatabase := nil;
  if (AComponent = FTransaction) and (Operation = opRemove) then
    FTransaction := nil;}
end;


procedure TBS_ADOOracleExtract.ListRoles(ObjectName: String);
const
  RolesSQL =
    'select * from RDB$ROLES ' +
    'order by RDB$ROLE_NAME';

  RolesByNameSQL =
    'select * from RDB$ROLES ' +
    'WHERE RDB$ROLE_NAME = :RoleName ' +
    'order by RDB$ROLE_NAME';

var
  //qryRoles : TADODataSet;
  PrevOwner, RoleName, OwnerName : String;
begin
  //Process GRANT roles
  {qryRoles := CreateADODataSet;
  try
    if FDatabaseInfo.ODSMajorVersion >= ODS_VERSION9 then
    begin
      PrevOwner := '';
      FMetaData.Add('');
      FMetaData.Add('/* Grant Roles for this database */');
      FMetaData.Add('');

      if ObjectName = '' then
        qryRoles.SQL.Text := RolesSQL
      else
      begin
        qryRoles.SQL.Text := RolesByNameSQL;
        qryRoles.Params.ByName('RoleName').AsString := ObjectName;
      end;
      qryRoles.ExecQuery;
      try
        while not qryRoles.Eof do
        begin
          RoleName := QuoteIdentifier(FDatabase.SQLDialect,
              qryRoles.FieldByName('rdb$Role_Name').AsString);
          OwnerName := Trim(qryRoles.FieldByName('rdb$Owner_Name').AsString);
          if PrevOwner <> OwnerName then
          begin
            FMetaData.Add('');
            FMetaData.Add(Format('/* Role: %s, Owner: %s */', [RoleName, OwnerName]));
            FMetaData.Add('');
            PrevOwner := OwnerName;
          end;
          FMetaData.Add('CREATE ROLE ' + RoleName + Term);
          qryRoles.Next;
        end;
      finally
        qryRoles.Close;
      end;
    end;
  finally
    qryRoles.Free;
  end;
  }
end;

function TBS_ADOOracleExtract.GetFieldLength(sql: TADODataSet): Integer;
begin
  {if sql.FieldByName('RDB$CHARACTER_LENGTH').IsNull then
    Result := sql.FieldByName('RDB$FIELD_LENGTH').AsInteger
  else
    Result := sql.FieldByName('RDB$CHARACTER_LENGTH').AsInteger;
  }
end;

function TBS_ADOOracleExtract.CreateADODataSet: TADODataSet;
begin
  Result := TADODataSet.Create(Self);
  Result.Connection := FConnection;
end;

procedure TBS_ADOOracleExtract.SetConnection(const Value: TADOConnection);
begin
  if FConnection <> Value then
  begin
    FConnection := Value;
    //FConnectionInfo.Database := FConnection;
  end;
end;

constructor TBS_ADOOracleExtract.Create(AOwner: TComponent);
begin
  inherited;
  FMetaData := TStringList.Create;
  FCommitInterval := 0;
  //FDatabaseInfo := TIBDatabaseInfo.Create(nil);
  //FDatabaseInfo.Database := FDatabase;
  if AOwner is TADOConnection then
    Connection := TADOConnection(AOwner);
end;

destructor TBS_ADOOracleExtract.Destroy;
begin
  FMetaData.Free;
  //FDatabasEInfo.Free;
  inherited;
end;




//**
//  ListCreateDb
//  Functional description
//   Print the create database command if requested.  At least put
//   the page size in a comment with the extracted db name }
//*
procedure TBS_ADOOracleExtract.ListCreateDb(TargetDb : String);
var
  Buffer,fTargetDb : String;
begin
  Buffer := '/* ';
  fTargetDb:=TargetDb;
  if fTargetDb = '' then
    fTargetDb := 'C:\' + Self.FConnection.DefaultDatabase;
  Buffer := Buffer + 'CREATE DATABASE ' + '"' + fTargetDb + '" */'+ NEWLINE;
  FMetaData.Add(Buffer);
end;

{ ListTable
  Functional description
  	Shows columns, types, info for a given table name
  	and text of views.
  	If a new_name is passed, substitute it for relation_name

  	relation_name -- Name of table to investigate
  	new_name -- Name of a new name for a replacement table
  	domain_flag -- extract needed domains before the table }

function TBS_ADOOracleExtract.ListTable(aOwnerName, aRelationName, aNewName: String) : Boolean;
const
  TableColumnsSQL =
    'select * from ALL_TAB_COLUMNS'+
    ' where OWNER = :OwnerName '+
    ' and TABLE_NAME = :TableName';

  PRIMARY_KEY_CONSTRAINT_SQL =
    'select * from ALL_CONSTRAINTS ' +
    ' where OWNER = :OwnerName '+
    ' and TABLE_NAME = :TableName'+
    ' and CONSTRAINT_TYPE=''P''' ;

  ConstraintSQL =
    'select * from ALL_CONSTRAINTS ' +
    ' where OWNER = :OwnerName '+
    ' and TABLE_NAME = :TableName';

  ALL_COL_COMMENTS_SQL =
    'select COLUMN_NAME, COMMENTS from ALL_COL_COMMENTS'+
    ' where OWNER = :OwnerName '+
    ' and TABLE_NAME = :TableName';

var
  tableComment, colList, column, constraint : String;
  qryTableColumns,
  qryConstraints,
  qryColComments : TADODataSet;
  validRelation: boolean;
begin
  Result := true;
  ValidRelation := false;


  qryTableColumns := CreateADODataSet;
  qryConstraints  := CreateADODataSet;
  qryColComments  := CreateADODataSet;
  try
    try
      qryTableColumns.CommandText:=TableColumnsSQL;
      qryTableColumns.Parameters.ParamByName('OwnerName').Value := aOwnerName;
      qryTableColumns.Parameters.ParamByName('TableName').Value := aRelationName;
      qryTableColumns.Prepared:=True;
      qryTableColumns.Open;
      if not qryTableColumns.Eof then
      begin
        ValidRelation := true;
        if (aOwnerName <> '') then
          FMetaData.Add(Format('%s/* Table: %s.%s */%s',
             [NEWLINE,
               aOwnerName,
               aRelationName,
               '']));
        if aNewName <> '' then
          FMetaData.Add(Format('create table %s.%s ', [aOwnerName,aNewName]))
        else
          FMetaData.Add(Format('create table %s.%s ', [aOwnerName,aRelationName]));
        FMetaData.Add('(');
      end;
      //COLUMN
      while not qryTableColumns.Eof do
      begin
        Column := '  ' + qryTableColumns.FieldByName('COLUMN_NAME').AsString + ' ';
        while Length(Column) <= 30 do
          Column := Column + ' ';
        //datatype
        if (qryTableColumns.FieldByName('DATA_TYPE').AsString='NUMBER')
          or (qryTableColumns.FieldByName('DATA_TYPE').AsString='NUMERIC') then
          begin
            Column := Column + qryTableColumns.FieldByName('DATA_TYPE').AsString;
            if qryTableColumns.FieldByName('DATA_PRECISION').Value<>0 then
            begin
              Column := Column + '('+qryTableColumns.FieldByName('DATA_PRECISION').AsString;
              if qryTableColumns.FieldByName('DATA_SCALE').Value<>0 then
                Column := Column + ','+qryTableColumns.FieldByName('DATA_SCALE').AsString;
              Column := Column + ')';
            end;
          end
        else if (qryTableColumns.FieldByName('DATA_TYPE').AsString='VARCHAR2')
             or (qryTableColumns.FieldByName('DATA_TYPE').AsString='VARCHAR')
             or (qryTableColumns.FieldByName('DATA_TYPE').AsString='CHAR') then
          begin
            Column := Column + qryTableColumns.FieldByName('DATA_TYPE').AsString;
            if qryTableColumns.FieldByName('DATA_LENGTH').Value<>0 then
              Column := Column + '('+qryTableColumns.FieldByName('DATA_LENGTH').AsString+')';
          end
        else if (qryTableColumns.FieldByName('DATA_TYPE').AsString='TIMESTAMP') or
                 (qryTableColumns.FieldByName('DATA_TYPE').AsString='DATE')then
          begin
            Column := Column + 'DATE';//qryTableColumns.FieldByName('DATA_TYPE').AsString;
          end
        else
          begin
            Column := Column + qryTableColumns.FieldByName('DATA_TYPE').AsString;
          end;
        //allign
        while Length(Column) <= 42 do
            Column := Column + ' ';
        //DEFAULT expr
        if Not qryTableColumns.FieldByName('DATA_DEFAULT').IsNull then
          Column := Column +' DEFAULT '+qryTableColumns.FieldByName('DATA_DEFAULT').Value;

        //COLUMN_COSTRAINT
        if qryTableColumns.FieldByName('NULLABLE').Value='Y' then
          Column := Column +' NULL'
        else
          Column := Column +' NOT NULL';

        qryTableColumns.Next;
        if not qryTableColumns.Eof then
          Column := Column + ',';
        FMetaData.Add(Column);
      end;
      qryTableColumns.Close;

      // Primary keys constraints only. References come later
      qryConstraints.CommandText:=PRIMARY_KEY_CONSTRAINT_SQL;
      qryConstraints.Parameters.ParamByName('OwnerName').Value := aOwnerName;
      qryConstraints.Parameters.ParamByName('TableName').Value := aRelationName;

      qryConstraints.Open;
      if qryConstraints.RecordCount > 0 then
      begin
        constraint := '  ';
        FMetaData.Strings[FMetaData.Count - 1] := FMetaData.Strings[FMetaData.Count - 1]  + ',';
        constraint := Constraint + 'CONSTRAINT ' +
            qryConstraints.FieldByName('CONSTRAINT_NAME').AsString;
        FMetaData.Add(Constraint);
        constraint:='    PRIMARY KEY ('+
          Self.GetConsColumns(aOwnerName,aRelationName,qryConstraints.FieldByName('CONSTRAINT_NAME').AsString)+')';
        FMetaData.Add(Constraint);
      end;
      qryConstraints.Close;

      if ValidRelation then
        FMetaData.Add(')' + Term);

      tableComment:=GetTableComment(aOwnerName,aRelationName);
      if tableComment<>'' then
        if aNewName <> '' then
          FMetaData.Add(Format('comment on table %s.%s is %s', [aOwnerName,aNewName,
          QuotedStr(tableComment)
          ])+ Term)
        else
          FMetaData.Add(Format('comment on table %s.%s is %s', [aOwnerName,aRelationName,
          QuotedStr(tableComment)
          ])+ Term);

      qryColComments.CommandText:=ALL_COL_COMMENTS_SQL;
      qryColComments.Parameters.ParamByName('OwnerName').Value := aOwnerName;
      qryColComments.Parameters.ParamByName('TableName').Value := aRelationName;
      qryColComments.Open;
      qryTableColumns.Open;
      while not qryTableColumns.Eof do
      begin
        if Not qryTableColumns.FieldByName('COLUMN_NAME').IsNull then
        begin
          qryColComments.Filtered:=false;
          qryColComments.Filter:='COLUMN_NAME='+QuotedStr(qryTableColumns.FieldByName('COLUMN_NAME').Value);
          qryColComments.Filtered:=true;
          qryColComments.First;
          if qryColComments.RecordCount > 0 then
          begin
            if aNewName <> '' then
              begin
                FMetaData.Add(Format('comment on column %s.%s.%s is %s', [
                  aOwnerName,
                  aNewName,
                  qryTableColumns.FieldByName('COLUMN_NAME').Value,
                  QuotedStr(qryColComments.FieldByName('COMMENTS').Value)
                ])+ Term);
              end
            else
              begin
                FMetaData.Add(Format('comment on column %s.%s.%s is %s', [
                  aOwnerName,
                  aRelationName,
                  qryTableColumns.FieldByName('COLUMN_NAME').Value,
                  QuotedStr(qryColComments.FieldByName('COMMENTS').Value)
                ])+ Term);
              end;
            end;
          end;
          qryTableColumns.Next;
        end;
        qryTableColumns.Close;
        qryColComments.Close;
    except
      on E: Exception do
        begin
          FMetaData.Add('');
          FMetaData.Add('/*****  ERROR  ******');
          FMetaData.Add(E.Message);
          FMetaData.Add('******  ERROR  ******/');
          FMetaData.Add('');
        end;
    end;
  finally
    qryTableColumns.Free;
    qryConstraints.Free;
    qryColComments.Free;
  end;
end;


function TBS_ADOOracleExtract.GetTableComment(const aOwnerName,
  aTableName: String): String;
const
  ALL_CONS_COLUMNS_SQL =
    'select COMMENTS from ALL_TAB_COMMENTS' +
    ' where OWNER = :OwnerName '+
    ' and TABLE_NAME = :TableName';

var
  tmpDataSet : TADODataSet;
  ownerName,tableName : String;
begin
  Result := '';
  tmpDataSet := CreateADODataSet;
  try
    try
      tmpDataSet.CommandText := ALL_CONS_COLUMNS_SQL;
      tmpDataSet.Parameters.ParamByName('OwnerName').Value := aOwnerName;
      tmpDataSet.Parameters.ParamByName('TableName').Value := aTableName;
      tmpDataSet.Open;
      if tmpDataSet.RecordCount >0 then
        Result := tmpDataSet.fieldByName('COMMENTS').Value;
      tmpDataSet.Close;
    except
      on E: Exception do
        begin
          FMetaData.Add('');
          FMetaData.Add('/*****  ERROR  ******');
          FMetaData.Add(E.Message);
          FMetaData.Add('******  ERROR  ******/');
          FMetaData.Add('');
        end;
    end;
  finally
    tmpDataSet.Free;
  end;
end;

//**
//Get constraint columns
//*
function TBS_ADOOracleExtract.GetConsColumns(const
        aOwnerName,
        aTableName,
        aConstraintName: String): String;
const
  ALL_CONS_COLUMNS_SQL =
    'select * from ALL_CONS_COLUMNS' +
    ' where OWNER = :OwnerName '+
    ' and TABLE_NAME = :TableName'+
    ' and CONSTRAINT_NAME = :ConstraintName'+
    ' order by POSITION';

var
  tmpDataSet : TADODataSet;
  ownerName,tableName : String;
begin
  Result := '';
  tmpDataSet := CreateADODataSet;
  try
    try
      tmpDataSet.Connection:=Self.FConnection;
      tmpDataSet.CommandText := ALL_CONS_COLUMNS_SQL;
      tmpDataSet.Parameters.ParamByName('OwnerName').Value := aOwnerName;
      tmpDataSet.Parameters.ParamByName('TableName').Value := aTableName;
      tmpDataSet.Parameters.ParamByName('ConstraintName').Value := aConstraintName;
      tmpDataSet.Open;
      while not tmpDataSet.Eof do
      begin
        if Result<>'' then
          Result:=Result+', ';
        Result := Result+tmpDataSet.fieldByName('COLUMN_NAME').Value;
        tmpDataSet.Next;
        sleep(0);
      end;
      tmpDataSet.Close;
    except
      on E: Exception do
        begin
          FMetaData.Add('');
          FMetaData.Add('/*****  ERROR  ******');
          FMetaData.Add(E.Message);
          FMetaData.Add('******  ERROR  ******/');
          FMetaData.Add('');
        end;
    end;
  finally
    tmpDataSet.Free;
  end;
end;

{  ListSequences
 Functional description
   Re create all non-system Sequences }
procedure TBS_ADOOracleExtract.ListSequences(aOwnerName:String = ''; aSequenceName : String = '');
const
  SequencesAll =
    'select * ' +
    ' from ALL_SEQUENCES ' +
    ' order by SEQUENCE_NAME';

  SequencesQL =
    'select * ' +
    ' from ALL_SEQUENCES ' +
    ' where SEQUENCE_OWNER=:OwnerName' +
    ' order by SEQUENCE_NAME';

  SequenceNameSQL =
    'select * ' +
    ' from ALL_SEQUENCES ' +
    ' where SEQUENCE_OWNER = :OwnerName' +
    ' and SEQUENCE_NAME = :SequenceName';
var
  qrySequence : TADODataSet;
  qryCurrVal  : TADODataSet;
  seqName     : String;
  minValue    : integer;
begin
  qrySequence := CreateADODataSet;
  qryCurrVal  := CreateADODataSet;
  try
    try
      if aSequenceName = '' then
        begin
          if aOwnerName = '' then
            qrySequence.CommandText := SequencesALL
          else
            begin
              qrySequence.CommandText := SequencesQL;
              qrySequence.Parameters.ParamByName('OwnerName').Value := aOwnerName;
            end;
        end
      else
      begin
        qrySequence.CommandText := SequenceNameSQL;
        qrySequence.Parameters.ParamByName('OwnerName').Value := aOwnerName;
        qrySequence.Parameters.ParamByName('SequenceName').Value := aSequenceName;
      end;
      qrySequence.Open;
      FMetaData.Add('');
      while not qrySequence.Eof do
      begin
        seqName := qrySequence.FieldByName('SEQUENCE_NAME').AsString;
        qryCurrVal.CommandText:='select '
         + qrySequence.FieldByName('SEQUENCE_NAME').AsString
         +'.currval as CurValue from dual';
        qryCurrVal.Open;

        if qryCurrVal.FieldByName('CurValue').AsInteger >= qrySequence.FieldByName('MIN_VALUE').AsInteger then
          minValue:=qrySequence.FieldByName('MIN_VALUE').AsInteger
        else
          minValue:=qrySequence.FieldByName('CurValue').AsInteger;

        FMetaData.Add(Format('create sequence %s.%s minvalue %s start with %s increment by %s%s',
          [qrySequence.FieldByName('SEQUENCE_OWNER').AsString,
           seqName,
           IntToStr(minValue),
           qryCurrVal.FieldByName('CurValue').AsString,
           qrySequence.FieldByName('INCREMENT_BY').AsString,
           Term]));
           qryCurrVal.Close;
        sleep(0);
        qrySequence.Next;
      end;
    except
      on E: Exception do
        begin
          FMetaData.Add('');
          FMetaData.Add('/*****  ERROR  ******');
          FMetaData.Add(E.Message);
          FMetaData.Add('******  ERROR  ******/');
          FMetaData.Add('');
        end;
    end;
  finally
    qryCurrVal.Free;
    qrySequence.Free;
  end;
end;



//**
//*
function TBS_ADOOracleExtract.ExtractDDL(Flag: Boolean; TableName: String) : Boolean;
var
  DidConnect : Boolean;
  DidStart : Boolean;
begin
  Result := true;
  DidConnect := false;
  DidStart := false;

  if not FConnection.Connected then
  begin
    FConnection.Connected := true;
    didConnect := true;
  end;

  if TableName <> '' then
  begin
    if not ListTable('',TableName, '') then
      Result := false;
  end
  else
  begin
    ListCreateDb;
    sleep(0);
    ListSequences;
    sleep(0);
    ListAllTables('',flag);
    sleep(0);
    ListIndex;
    sleep(0);
    ListForeign;
    sleep(0);
    ListFilters;
    sleep(0);
    ListFunctions;
    sleep(0);
    ListCheck;
    sleep(0);
    ListException;
    sleep(0);
    ListViews;
    sleep(0);
    ListProcs;
    sleep(0);
    ListTriggers;
    sleep(0);
    ListGrants;
  end;
  if DidStart and Self.FConnection.InTransaction then
    Self.FConnection.CommitTrans;

  if DidConnect then
    Self.FConnection.Connected := false;
end;


procedure TBS_ADOOracleExtract.ListData(ObjectName: String);
const
  SelectSQL = 'select * from %s';
var
  qrySelect : TADODataSet;
  tmpStr : String;
  i,j : Integer;
begin
  qrySelect := CreateADODataSet;
  try
    try
      qrySelect.CommandText := Format(SelectSQL, [ObjectName]);
      qrySelect.Open;
      j:=0;
      while not qrySelect.Eof do
      begin
        tmpStr := 'insert into ' + ObjectName + ' (';
        for i := 0 to qrySelect.FieldCount - 1 do
          if not(qrySelect.Fields[i].DataType in [ftVarBytes,ftBytes,ftBlob,ftMemo,
             ftGraphic,ftFmtMemo,ftParadoxOle,ftDBaseOle,ftTypedBinary,ftCursor,
             ftUnknown,ftBoolean,ftADT,ftArray,ftReference,ftDataSet,ftOraBlob,
             ftOraClob,ftVariant,ftInterface,ftIDispatch,ftGuid,ftFMTBcd])  then
          begin
            tmpStr := tmpStr + qrySelect.Fields[i].FieldName;
            if i <> (qrySelect.FieldCount - 1) then
              tmpStr := tmpStr + ', ';
          end;
        tmpStr := tmpStr + ') values (';

        for i := 0 to qrySelect.FieldCount - 1 do
        begin
          if qrySelect.Fields[i].IsNull and
             not(qrySelect.Fields[i].DataType in [ftVarBytes,ftBytes,ftBlob,ftMemo,
             ftGraphic,ftFmtMemo,ftParadoxOle,ftDBaseOle,ftTypedBinary,ftCursor,
             ftUnknown,ftBoolean,ftADT,ftArray,ftReference,ftDataSet,ftOraBlob,
             ftOraClob,ftVariant,ftInterface,ftIDispatch,ftGuid,ftFMTBcd]) then
          begin
            tmpStr := tmpStr + 'NULL';
            if i <> (qrySelect.FieldCount - 1) then
              tmpStr := tmpStr + ', ';
          end
          else
          case qrySelect.Fields[i].DataType of
            ftString,ftFixedChar,ftWideString,
            ftTimeStamp,ftDate,ftTime,ftDateTime:
            begin
              tmpStr := tmpStr + QuotedStr(qrySelect.Fields[i].AsString);
              if i <> (qrySelect.FieldCount - 1) then
                tmpStr := tmpStr + ', ';
            end;
            ftSmallint,ftInteger,ftAutoInc,ftLargeint,
            ftCurrency,ftBCD,ftWord,ftFloat:
            begin
              tmpStr := tmpStr + qrySelect.Fields[i].AsString;
              if i <> (qrySelect.FieldCount - 1) then
                tmpStr := tmpStr + ', ';
            end;
            ftVarBytes,ftBytes,ftBlob,ftMemo,
            ftGraphic,ftFmtMemo,ftParadoxOle,ftDBaseOle,ftTypedBinary,ftCursor,
            ftUnknown,ftBoolean,ftADT,ftArray,ftReference,ftDataSet,ftOraBlob,
            ftOraClob,ftVariant,ftInterface,ftIDispatch,ftGuid,ftFMTBcd : ;
            //else
            // Error Invalid Data Conversion
          end;
        end;
        tmpStr := tmpStr + ')' + Term;
        FMetaData.Add(tmpStr);
        j:=j+1;
        if (j = Self.CommitInterval) then
        begin
          j:=0;
          FMetaData.Add('commit'+ Term);
        end;
        qrySelect.Next;
      end;
    except
      on E: Exception do
        begin
          FMetaData.Add('');
          FMetaData.Add('/*****  ERROR  ******');
          FMetaData.Add(E.Message);
          FMetaData.Add('******  ERROR  ******/');
          FMetaData.Add('');
        end;
    end;
  finally
    qrySelect.Free;
  end;
end;

procedure TBS_ADOOracleExtract.ExtractObject(aObjectType : TBS_ExtractObjectTypes;
      aObjectName : String = ''; aExtractTypes : TBS_ExtractTypes = []);
var
  didActivate : Boolean;
  fOwner,fObjectName:String;
begin
  didActivate := false;
  if Pos('.',aObjectName)>0 then
    begin
      fOwner:=Copy(aObjectName,1,Pos('.',aObjectName)-1);
      fObjectName:=Copy(aObjectName,Pos('.',aObjectName)+1,Length(aObjectName)-Pos('.',aObjectName));
    end
  else
    begin
      fOwner:='';
      fObjectName:=aObjectName;
    end;

  if not Self.FConnection.InTransaction then
  begin
    Self.FConnection.BeginTrans;
    didActivate := true;
  end;

  FMetaData.Clear;
  case aObjectType of
    exoDatabase : ExtractDDL(true, '');
    exoTable :
    begin
      if fObjectName <> '' then
      begin
        ListTable(fOwner, fObjectName, '');
        if extIndex in aExtractTypes then
          ListIndex(fOwner,fObjectName, extTable);
        if extForeign in aExtractTypes then
          ListForeign(fObjectName, extTable);
        if extCheck in aExtractTypes then
          ListCheck(fObjectName, extTable);
        if extTrigger in aExtractTypes then
          ListTriggers(fObjectName, extTable);
        if extGrant in aExtractTypes then
          ShowGrants(fObjectName, Term);
        if extData in aExtractTypes then
          ListData(aObjectName);
      end
      else
        ListAllTables('',true);
    end;
    exoView :
    begin
      ListViews(fObjectName);
      if extTrigger in aExtractTypes then
        ListTriggers(fObjectName, extTable);
    end;

    exoData : ListData(fObjectName);

    exoProcedure : ListProcs(fObjectName, (extAlterProc in aExtractTypes));
    exoFunction : ListFunctions(fObjectName);
    exoSequence : ListSequences(fObjectName);
    exoException : ListException(fObjectName);
    exoBLOBFilter : ListFilters(fObjectName);
    exoRole : ListRoles(fObjectName);
    exoTrigger :
      if extTable in aExtractTypes then
        ListTriggers(fObjectName, extTable)
      else
        ListTriggers(fObjectName);

    exoForeign :
      if extTable in aExtractTypes then
        ListForeign(fObjectName, extTable)
      else
        ListForeign(fObjectName);
    exoIndexes :
      if extTable in aExtractTypes then
        ListIndex(fOwner,fObjectName, extTable)
      else
        ListIndex(fOwner,fObjectName);
    exoChecks :
      if extTable in aExtractTypes then
        ListCheck(fObjectName, extTable)
      else
        ListCheck(fObjectName);

  end;
  if didActivate and Self.FConnection.InTransaction then
    Self.FConnection.CommitTrans;
end;

procedure TBS_ADOOracleExtract.ListAllTables(const aOwner:String; const aShowSystem: Boolean);
const
  TableSQL =
    'select * from ALL_OBJECTS'+
    ' where OBJECT_TYPE=''TABLE''';

var
  qryTables : TADODataSet;
  ownerName,tableName : String;
begin
  qryTables := TADODataSet.Create(Self);
  qryTables.Connection:=Self.FConnection;
  try
    try
       qryTables.CommandText := TableSQL;
       if aOwner<>'' then
         qryTables.CommandText:=qryTables.CommandText +' and OWNER='+ #39 +aOwner+#39;
       qryTables.Open;
       while not qryTables.Eof do
       begin
         ownerName:=qryTables.FieldByName('OWNER').AsString;
         tableName:=qryTables.FieldByName('OBJECT_NAME').AsString;
         ListTable(ownerName,tableName,'');
         sleep(0);
         qryTables.Next;
       end;
    except
      on E: Exception do
        begin
          FMetaData.Add('');
          FMetaData.Add('/*****  ERROR  ******');
          FMetaData.Add(E.Message);
          FMetaData.Add('******  ERROR  ******/');
          FMetaData.Add('');
        end;
    end;
  finally
    qryTables.Free;
  end;
end;



{ ListForeign
  Functional description
  List all foreign key constraints and alter the tables }
procedure TBS_ADOOracleExtract.ListForeign(ObjectName : String; ExtractType : TBS_ExtractType);
const
  { Static queries for obtaining foreign constraints, where RELC1 is the
    foreign key constraints, RELC2 is the primary key lookup and REFC
    is the join table }
  ForeignSQL =
    'SELECT REFC.RDB$UPDATE_RULE REFC_UPDATE_RULE, REFC.RDB$DELETE_RULE REFC_DELETE_RULE, ' +
    '  RELC1.RDB$RELATION_NAME RELC1_RELATION_NAME, RELC2.RDB$RELATION_NAME RELC2_RELATION_NAME, ' +
    '  RELC1.RDB$INDEX_NAME RELC1_INDEX_NAME, RELC1.RDB$CONSTRAINT_NAME RELC1_CONSTRAINT_NAME, ' +
    '  RELC2.RDB$INDEX_NAME RELC2_INDEX_NAME ' +
    'FROM RDB$REF_CONSTRAINTS REFC, RDB$RELATION_CONSTRAINTS RELC1, ' +
    '     RDB$RELATION_CONSTRAINTS RELC2 ' +
    'WHERE ' +
    '  RELC1.RDB$CONSTRAINT_TYPE = ''FOREIGN KEY'' AND ' +
    '  REFC.RDB$CONST_NAME_UQ = RELC2.RDB$CONSTRAINT_NAME AND ' +
    '  REFC.RDB$CONSTRAINT_NAME = RELC1.RDB$CONSTRAINT_NAME AND ' +
    '  (RELC2.RDB$CONSTRAINT_TYPE = ''UNIQUE'' OR ' +
    '    RELC2.RDB$CONSTRAINT_TYPE = ''PRIMARY KEY'') ' +
    'ORDER BY RELC1.RDB$RELATION_NAME, RELC1.RDB$CONSTRAINT_NAME';

  ForeignNameSQL =
    'SELECT REFC.RDB$UPDATE_RULE REFC_UPDATE_RULE, REFC.RDB$DELETE_RULE REFC_DELETE_RULE, ' +
    '  RELC1.RDB$RELATION_NAME RELC1_RELATION_NAME, RELC2.RDB$RELATION_NAME RELC2_RELATION_NAME, ' +
    '  RELC1.RDB$INDEX_NAME RELC1_INDEX_NAME, RELC1.RDB$CONSTRAINT_NAME RELC1_CONSTRAINT_NAME, ' +
    '  RELC2.RDB$INDEX_NAME RELC2_INDEX_NAME ' +
    'FROM RDB$REF_CONSTRAINTS REFC, RDB$RELATION_CONSTRAINTS RELC1, ' +
    '     RDB$RELATION_CONSTRAINTS RELC2 ' +
    'WHERE ' +
    '  RELC1.RDB$RELATION_NAME = :TableName AND ' +
    '  RELC1.RDB$CONSTRAINT_TYPE = ''FOREIGN KEY'' AND ' +
    '  REFC.RDB$CONST_NAME_UQ = RELC2.RDB$CONSTRAINT_NAME AND ' +
    '  REFC.RDB$CONSTRAINT_NAME = RELC1.RDB$CONSTRAINT_NAME AND ' +
    '  (RELC2.RDB$CONSTRAINT_TYPE = ''UNIQUE'' OR ' +
    '    RELC2.RDB$CONSTRAINT_TYPE = ''PRIMARY KEY'') ' +
    'ORDER BY RELC1.RDB$RELATION_NAME, RELC1.RDB$CONSTRAINT_NAME';

  ForeignByNameSQL =
    'SELECT REFC.RDB$UPDATE_RULE REFC_UPDATE_RULE, REFC.RDB$DELETE_RULE REFC_DELETE_RULE, ' +
    '  RELC1.RDB$RELATION_NAME RELC1_RELATION_NAME, RELC2.RDB$RELATION_NAME RELC2_RELATION_NAME, ' +
    '  RELC1.RDB$INDEX_NAME RELC1_INDEX_NAME, RELC1.RDB$CONSTRAINT_NAME RELC1_CONSTRAINT_NAME, ' +
    '  RELC2.RDB$INDEX_NAME RELC2_INDEX_NAME ' +
    'FROM RDB$REF_CONSTRAINTS REFC, RDB$RELATION_CONSTRAINTS RELC1, ' +
    '     RDB$RELATION_CONSTRAINTS RELC2 ' +
    'WHERE ' +
    '  RELC1.RDB$CONSTRAINT_NAME = :ConstraintName AND ' +
    '  RELC1.RDB$CONSTRAINT_TYPE = ''FOREIGN KEY'' AND ' +
    '  REFC.RDB$CONST_NAME_UQ = RELC2.RDB$CONSTRAINT_NAME AND ' +
    '  REFC.RDB$CONSTRAINT_NAME = RELC1.RDB$CONSTRAINT_NAME AND ' +
    '  (RELC2.RDB$CONSTRAINT_TYPE = ''UNIQUE'' OR ' +
    '    RELC2.RDB$CONSTRAINT_TYPE = ''PRIMARY KEY'') ' +
    'ORDER BY RELC1.RDB$RELATION_NAME, RELC1.RDB$CONSTRAINT_NAME';

var
//  qryForeign : TADODataSet;
  Line : String;

begin
{  qryForeign := CreateADODataSet;
  try
    if ObjectName = '' then
      qryForeign.SQL.Text := ForeignSQL
    else
    begin
      if ExtractType = extTable then
      begin
        qryForeign.SQL.Text := ForeignNameSQL;
        qryForeign.Params.ByName('TableName').AsString := ObjectName;
      end
      else
      begin
        qryForeign.SQL.Text := ForeignByNameSQL;
        qryForeign.Params.ByName('ConstraintName').AsString := ObjectName;
      end;
    end;
    qryForeign.ExecQuery;
    while not qryForeign.Eof do
    begin
      Line := Format('ALTER TABLE %s ADD ', [QuoteIdentifier(FDatabase.SQLDialect,
        qryForeign.FieldByName('RELC1_RELATION_NAME').AsString)]);

      // If the name of the constraint is not INTEG..., print it.
      //   INTEG... are internally generated names.
      if (not qryForeign.FieldByName('RELC1_CONSTRAINT_NAME').IsNull) and
         ( Pos('INTEG', qryForeign.FieldByName('RELC1_CONSTRAINT_NAME').AsString) <> 1) then
        Line := Line + Format('CONSTRAINT %s ', [QuoteIdentifier(FDatabase.SQLDialect,
          Trim(qryForeign.FieldByName('RELC1_CONSTRAINT_NAME').AsString))]);

      Line := Line + Format('FOREIGN KEY (%s) REFERENCES %s ', [
        GextIndexSegments(qryForeign.FieldByName('RELC1_INDEX_NAME').AsString),
        Trim(qryForeign.FieldByName('RELC2_RELATION_NAME').AsString)]);

      Line := Line + Format('(%s)',
        [GextIndexSegments(qryForeign.FieldByName('RELC2_INDEX_NAME').AsString)]);

      // Add the referential actions, if any
      if (not qryForeign.FieldByName('REFC_UPDATE_RULE').IsNull) and
         (Trim(qryForeign.FieldByName('REFC_UPDATE_RULE').AsString) <> 'RESTRICT') then
        Line := Line + Format(' ON UPDATE %s',
           [Trim(qryForeign.FieldByName('REFC_UPDATE_RULE').AsString)]);

      if (not qryForeign.FieldByName('REFC_DELETE_RULE').IsNull) and
         (Trim(qryForeign.FieldByName('REFC_DELETE_RULE').AsString) <> 'RESTRICT') then
        Line := Line + Format(' ON DELETE %s',
           [Trim(qryForeign.FieldByName('REFC_DELETE_RULE').AsString)]);

      Line := Line + Term;
      FMetaData.Add(Line);
      qryForeign.Next;
    end;
  finally
    qryForeign.Free;
  end;
  }
end;



//**
// ListIndex
// Functional description
//   Define all non-constraint indices
//*
procedure TBS_ADOOracleExtract.ListIndex(aOwner: String = '';aObjectName : String = ''; ExtractType : TBS_ExtractType = extIndex);
const
  TABLE_INDEXES_SQL =
   'select * from ALL_INDEXES'+
   ' where TABLE_OWNER =:TableOwner'+
   ' and TABLE_NAME =:TableName'+
   ' and UNIQUENESS=''NONUNIQUE''';

  ALL_INDEXES_SQL =
   'select * from ALL_INDEXES'+
   ' where UNIQUENESS=''NONUNIQUE''';

  ALL_IND_COLUMNS =
   'select * from ALL_IND_COLUMNS order by COLUMN_POSITION';

var
  qryIndex : TADODataSet;
  qryIndexColumns : TADODataSet;
  First : Boolean;
  Unique, IdxType, Line : String;
begin
  First := true;
  qryIndex := CreateADODataSet;
  try
    try
      if aObjectName = '' then
        qryIndex.CommandText := ALL_INDEXES_SQL
      else
      begin
        if ExtractType = extTable then
        begin
          qryIndex.CommandText := TABLE_INDEXES_SQL;
          qryIndex.Parameters.ParamByName('OwnerName').Value := aOwner;
          qryIndex.Parameters.ParamByName('TableName').Value := aObjectName;
        end
      end;
      qryIndex.Open;
      while not qryIndex.Eof do
      begin
        if First then
        begin
          if aObjectName = '' then
            FMetaData.Add(NEWLINE + '/*  Index definitions for all user tables */' + NEWLINE)
          else
            FMetaData.Add(NEWLINE + '/*  Index definitions for ' + aObjectName + ' */' + NEWLINE);
          First := false;
        end; //end_if

        FMetaData.Add(Line);
        qryIndex.Next;
      end;
    except
      on E: Exception do
        begin
          FMetaData.Add('');
          FMetaData.Add('/*****  ERROR  ******');
          FMetaData.Add(E.Message);
          FMetaData.Add('******  ERROR  ******/');
          FMetaData.Add('');
        end;
    end;
  finally
    qryIndex.Free;
  end;
end;

end.


