{$I BS.inc}
unit BS_ClientDataSet;
// Modified by Siarhej Berdachuk from CDSDataset

{This component provides a way to link a TDBDataset to a TClientDataset at}
{design-time. This originally appeared in Delphi Informant, in an article}
{written by Dan Miser - "Using ClientDataset in 2-tier applications"}
{Copyright 1997-2000, Dan Miser}

{Permission is given to use this unit, as long as the comments above remain}
{in tact and unchanged.}

interface

uses
  Windows, Messages, SysUtils, Classes,
  Graphics, Controls, Forms, Dialogs,
  Db, DBClient, DBTables, Provider, TypInfo,
  stdctrls, Midas
  , BS_db //Common objects
  {$IFDEF CompilerVersion6},Variants
  //{$ELSE}
  {$ENDIF}
;

type
  {$IFDEF CompilerVersion6}
    TBS_ClientDataSet = class(TCustomClientDataset)
  {$ELSE}
    TBS_ClientDataSet = class(TClientDataset)
  {$ENDIF}
  private
    fSubstituteChar: Char;
    fSqlSubstitutions: BS_db.TSqlSubstitutions;
    fSqlSubstitutionsCheck:boolean;
    fDataProvider : TComponent;
    fWaitCursor   : TCursor;
    fStoredCursor : TCursor;
    FExcludeCopyFieldsList : TStringList;
    //FCommandText  : WideString;
    FCommandStrings  : TStringList;
    procedure SetExcludeCopyFields(Value : String);
    function  GetExcludeCopyFields:String;
    procedure InitAppServer;
    procedure SetDataProvider(Value : TComponent);
    procedure SetSubstituteChar(Value: Char);
    procedure SetSqlSubstitutions(Value: TSqlSubstitutions);
    function  GetRealCommandText: String;
    procedure SetInternalCommandText(Value:WideString);
    function  GetInternalCommandText: WideString;
    procedure CommandTextChanged(Sender: TObject);
  protected
    procedure CloseCursor; override;
    procedure Notification(AComponent : TComponent; Operation : TOperation); override;
    procedure OpenCursor(InfoQuery: Boolean); override;
    function DoGetRecords(Count: Integer; out RecsOut: Integer; Options: Integer;
       const CommandText: WideString; Params: OleVariant): OleVariant; override;
    procedure DoExecute(Params: OleVariant); override;
  published
    property  DataProvider : TComponent read FDataProvider write SetDataProvider;
    property  ExcludeCopyFields : String read GetExcludeCopyFields write SetExcludeCopyFields;
    property  WaitCursor:  TCursor read fWaitCursor write fWaitCursor default crHourGlass;
    property SubstituteChar: Char read fSubstituteChar write SetSubstituteChar default DEFAULT_SUBSTITUTE_CHAR;
    property SqlSubstitutions: TSqlSubstitutions read fSqlSubstitutions write SetSqlSubstitutions;
    //property CommandText: WideString read FCommandText write FCommandText;
    property CommandText: WideString read GetInternalCommandText write SetInternalCommandText;
    property RealCommandText: string read GetRealCommandText;
    property SqlSubstitutionsCheck : boolean read fSqlSubstitutionsCheck write fSqlSubstitutionsCheck default true;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    function  GetFieldValueAsOraSqlStr(AField:TField):String;
    procedure FreeCreatedIndexes;
    function  SearchCdsRecord(const aEdtSearch:Tedit):boolean; overload;
    function  SearchCdsRecord(const aEdtSearchText:String):boolean; overload;
    procedure CopyRecord;
    procedure SubstituteSQL;
    procedure Execute; override;
    function SortData(const aIndexFields,aDescFields:String):boolean;
    function ParseSqlSubstitutions(aCommandText: string; DoCreate: Boolean): string;
    function SqlSubstitutionByName(const Value: string): TSqlSubstitution;
  end;


implementation

/////////////// TBS_ClientDataSet ///////////////
{$D-L-}
constructor TBS_ClientDataSet.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fWaitCursor := crHourGlass;
  fExcludeCopyFieldsList := TStringList.Create;
  fCommandStrings := TStringList.Create;
  TStringList(fCommandStrings).OnChange:=Self.CommandTextChanged;
  fExcludeCopyFieldsList.Sorted:=True;
  fSubstituteChar:=DEFAULT_SUBSTITUTE_CHAR;
  fSqlSubstitutions := TSqlSubstitutions.Create(Self);
  fSqlSubstitutionsCheck := false;
end;


destructor TBS_ClientDataSet.Destroy();
begin
  fCommandStrings.Free;
  fExcludeCopyFieldsList.Free;
  fExcludeCopyFieldsList:=nil;
  fSqlSubstitutions.Free;
  fSqlSubstitutions:=nil;
  inherited Destroy();
end;


procedure TBS_ClientDataSet.FreeCreatedIndexes;
var
  i:integer;
begin
  if Self.Active then
  try
    IndexDefs.Update; //Refresh indexes list
    for i := 0 to IndexDefs.Count - 1 do
      if (IndexDefs.Items[i].Name<>'DEFAULT_ORDER') and (IndexDefs.Items[i].Name<>'CHANGEINDEX')
         and (IndexDefs.Items[i].Name<>'PRIMARY_KEY') then
           begin
             if (Self.IndexName=IndexDefs.Items[i].Name) then
               Self.IndexName:='';
             DeleteIndex(IndexDefs.Items[i].Name);
           end;
     IndexDefs.Update;
  except
    raise;
  end;
end;


procedure TBS_ClientDataSet.SetExcludeCopyFields(Value : String);
var
  p: Integer;
begin
  fExcludeCopyFieldsList.Clear;
  p := 1;
  while p <= Length(Value) do
    fExcludeCopyFieldsList.Add(ExtractFieldName(Value, p));
end;


function TBS_ClientDataSet.GetExcludeCopyFields:String;
var i:integer;
begin
  Result:='';
  for i:=0 to FExcludeCopyFieldsList.Count-1 do
    begin
      if i<>(FExcludeCopyFieldsList.Count-1) then
        Result:=Result+FExcludeCopyFieldsList.Strings[i]+';'
      else
        Result:=Result+FExcludeCopyFieldsList.Strings[i]
    end;
end;


// Insert new record with data from previously selected record
// Exclude fields in ExcludeCopyFieldsList property
procedure TBS_ClientDataSet.CopyRecord;
var
 dataArray,fieldsArray: Variant;
 i,index: integer;
begin
  dataArray := VarArrayCreate([0, Self.Fields.Count-1], varVariant); //Data
  fieldsArray := VarArrayCreate([0, Self.Fields.Count-1], varVariant); //Fields names
  for i:=0 to Self.Fields.Count-1 do  //Copy data in var array
    begin
      dataArray[i]:=Self.Fields[i].Value;
      fieldsArray[i]:=Self.Fields[i].FieldName;
    end;
  //insert;
  Self.Append;
  for i:=0 to Self.Fields.Count-1 do
    if not Self.FExcludeCopyFieldsList.Find(fieldsArray[i],index) then
      Self.Fields[i].Value:=dataArray[i]
end;


//private
procedure TBS_ClientDataSet.InitAppServer;
begin
  if FDataProvider is TCustomProvider then
    AppServer := TLocalAppServer.Create(TCustomProvider(FDataProvider)) else
    AppServer := TLocalAppServer.Create(TDataset(FDataProvider));
end;


// D3 and D4 used the Provider property, which was constant as you opened and
// closed the CDS. D5 frees the AppServer property in the CloseCursor method.
procedure TBS_ClientDataSet.SetDataProvider(Value : TComponent);
begin
  if Value <> FDataProvider then
  begin
    FDataProvider:=Value;
    {Calling FreeNotification ensures that this component will receive an}
    {opRemove when Value is either removed from its owner or when it is destroyed.}
    if FDataProvider<>nil then
     begin
      FDataProvider.FreeNotification(Self);
      {Assign the provider from the host provider component to the ClientDataset (ourself)}
      InitAppServer;
     end
    else
     AppServer:=nil;
  end;
end;


procedure TBS_ClientDataSet.Notification(AComponent : TComponent; Operation : TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FDataProvider) then begin
    AppServer:=nil;
  end;
end;


function TBS_ClientDataSet.GetFieldValueAsOraSqlStr(AField:TField):String;
var tmpStr:string;
begin
  tmpStr:='';
  if VarIsNull(AField.Value) and Not(AField.DataType in [ftString, ftFixedChar, ftWideString, ftGuid]) then
    tmpStr:='Null'
  else
    begin
      case AField.DataType of
       ftBlob,ftMemo, ftGraphic,
       ftFmtMemo, ftParadoxOle,
       ftDBaseOle, ftOraBlob,
       ftOraClob:
         begin
           tmpStr := 'Null';
         end;
       ftWord,ftFloat,ftBCD,
       ftSmallint,ftInteger,
       ftCurrency, ftBytes,
       ftVarBytes, ftAutoInc,ftLargeint,
       ftTypedBinary:
         begin
           tmpStr := AField.AsString;
         end;
       ftString, ftFixedChar, ftWideString, ftGuid:
         begin
           tmpStr := CHR(39)+AField.AsString+CHR(39);
         end;
       ftDate, ftTime, ftDateTime:
         begin
           if (AField.Value=0) then
             tmpStr:='Null'
           else
             begin
               tmpStr:='TO_DATE('+CHR(39)
               + FormatDateTime('dd',AField.Value)+'/'
               + FormatDateTime('mm',AField.Value)+'/'
               + FormatDateTime('yyyy',AField.Value)
               + CHR(39) + ','+CHR(39)+'DD/MM/YYYY'+CHR(39)+')';
             end;
         end;
       ftBoolean:
         begin
           if (AField.Value=Null)or(AField.Value=False) then
             tmpStr:='0'
           else
             tmpStr:='1'
         end;
      end; //case
    end;
    Result:=tmpStr;
end;



// Prepare the AppServer property again after closing. This allows maximum
// flexibility for other calls where the AppServer property was being lazily
// instantiated, as opposed to simply creating it in the OpenCursor method.
procedure TBS_ClientDataSet.CloseCursor;
begin
  Self.FreeCreatedIndexes;
  inherited CloseCursor;
  if Assigned(DataProvider) then // Only reset AppServer if we're using a local connection
    InitAppServer;
end;

{$D+L+}
function  TBS_ClientDataSet.SearchCdsRecord(const aEdtSearchText:String):boolean;
var
  fieldsNamesStr    :String;
  i:integer;
  searchTextStr     :String;
  tmpSubStr         :String;
  searchValuesArray :Variant;
begin
  fieldsNamesStr:='';
  searchTextStr:=aEdtSearchText;
  Result:=False;
  if Self.Active then
  try
    searchValuesArray := VarArrayCreate([0, Self.IndexFieldCount - 1], varVariant);
    for i := 0 to Self.IndexFieldCount - 1 do
      begin
        if fieldsNamesStr<>'' then
          fieldsNamesStr:=fieldsNamesStr+';';
        fieldsNamesStr:=fieldsNamesStr + Self.IndexFields[i].FieldName;
        if Pos(';',searchTextStr) > 0 then
          begin
            tmpSubStr:=Copy(searchTextStr,1,Pos(';',searchTextStr)-1);
            searchTextStr:=Copy(searchTextStr,Pos(';',searchTextStr)+1,
                           Length(searchTextStr)-(Pos(';',searchTextStr)));
          end
        else
          begin
            tmpSubStr:=searchTextStr;
            searchTextStr:='';
          end;
        try
          case Self.IndexFields[i].DataType of
            ftSmallint,ftInteger,ftBytes,ftVarBytes,
            ftAutoInc,ftLargeint,ftTypedBinary:
              if tmpSubStr<>'' then
                searchValuesArray[i]:=StrToInt(tmpSubStr)
              else
                searchValuesArray[i]:=0;
            ftWord,ftFloat,ftBCD,ftCurrency:
              if tmpSubStr<>'' then
                searchValuesArray[i]:=StrToFloat(tmpSubStr)
              else
                searchValuesArray[i]:=0;
            ftDate, ftTime, ftDateTime:
              searchValuesArray[i]:=StrToDate(tmpSubStr);
          else
            searchValuesArray[i]:=tmpSubStr;
          end;
        except
          searchValuesArray[i]:=tmpSubStr;
        end;
      end;

    if Self.Locate(fieldsNamesStr, searchValuesArray,
       [loCaseInsensitive,loPartialKey]) then
      Result:=True
    else
      Result:=false;
  except      //Try restore last position
    Result:=False;
  end;
end;

function TBS_ClientDataSet.SearchCdsRecord(const aEdtSearch:Tedit):boolean;
begin
  Result:=Self.SearchCdsRecord(aEdtSearch.text);
  if Result  then
    aEdtSearch.Font.Color:=clWindowText
  else
    aEdtSearch.Font.Color:=clRed; {Can't find}
end;


function TBS_ClientDataSet.SortData(const aIndexFields,aDescFields: String): boolean;
var
  i                :integer;
  theIndexName     :String;
  theIndexFields   :String;
  theDescFields    :String;
  oldCursor        :TCursor;
  fIndexPresent    :boolean;
  recreateFlag     :boolean;
  p: Integer;

begin
 oldCursor:=Screen.Cursor;
 Screen.Cursor:=crHourGlass;
 theIndexFields   :=aIndexFields;
 theDescFields    :=aDescFields;

 try
   p := 1;
   theIndexName := '';
   while p <= Length(theIndexFields) do
     theIndexName := theIndexName +ExtractFieldName(theIndexFields, p)+ '_';

   p := 1;
   while p <= Length(theDescFields) do
     theIndexName := theIndexName +ExtractFieldName(theDescFields, p)+ 'Desc_';

   if theIndexName <> '' then
    theIndexName := 'i'+Copy(theIndexName,1,Length(theIndexName)-1);

   if theIndexName<>copy(theIndexName,1,30) then
     recreateFlag:=True
   else
     recreateFlag:=False;

   theIndexName:=copy(theIndexName,1,30);
   fIndexPresent:=False;
   Self.IndexDefs.Update;//Refresh
   //If index exist, then set it
   for i := 0 to Self.IndexDefs.Count - 1 do
    if LowerCase(Self.IndexDefs.Items[i].Name) = LowerCase(theIndexName) then
      begin
        if Not recreateFlag then
          begin
            fIndexPresent:=True; // Sign
            Self.IndexName := Self.IndexDefs.Items[i].Name;
          end
        else
          begin
            Self.DeleteIndex(theIndexName);
            Self.IndexName := '';
          end;
        break;
      end;
   //If index not exist, then try create it
   if Not(fIndexPresent)and (theIndexName<>'')  then
     try
       Self.AddIndex(theIndexName, theIndexFields,[ixCaseInsensitive],theDescFields,'',0);
       Self.IndexName := theIndexName;
     except
       //DatabaseErrorFmt('Can not create index ', [theIndexName], GetDataSet);
     end;
   Result:=True;
  finally
    Screen.Cursor:=oldCursor;
  end;
end;


procedure TBS_ClientDataSet.SetSubstituteChar(Value: Char);
begin
  if Value <> fSubstituteChar then
  begin
    fSubstituteChar := Value;
    //RecreateSubstitutions;
  end;
end;


procedure TBS_ClientDataSet.SetSqlSubstitutions(Value: TSqlSubstitutions);
begin
  fSqlSubstitutions.Assign(Value);
end;

//to do: Need revision, eat blank space after substitution
function TBS_ClientDataSet.ParseSqlSubstitutions(aCommandText: string;
  DoCreate: Boolean): string;
const
  Literals = ['''', '"', '`'];
var
  Value, CurPos, StartPos: PChar;
  CurChar: Char;
  Literal: Boolean;
  EmbeddedLiteral: Boolean;
  Name: string;
  substitutionValue:String;
  newCommandText:WideString;
  substitutionList:TSqlSubstitutions;
  function NameDelimiter: Boolean;
  begin
    Result := CurChar in [' ', ',', ';', ')', #13, #10];
  end;

  function IsLiteral: Boolean;
  begin
    Result := CurChar in Literals;
  end;

  function StripLiterals(Buffer: PChar): string;
  var
    Len: Word;
    TempBuf: PChar;

    procedure StripChar;
    begin
      if TempBuf^ in Literals then
        StrMove(TempBuf, TempBuf + 1, Len - 1);
      if TempBuf[StrLen(TempBuf) - 1] in Literals then
        TempBuf[StrLen(TempBuf) - 1] := #0;
    end;

  begin
    Len := StrLen(Buffer) + 1;
    TempBuf := AllocMem(Len);
    Result := '';
    try
      StrCopy(TempBuf, Buffer);
      StripChar;
      Result := StrPas(TempBuf);
    finally
      FreeMem(TempBuf, Len);
    end;
  end;

begin
  Result := aCommandText;
  newCommandText:='';
  substitutionList:=TSqlSubstitutions.Create(Self);
  try
   Value := PChar(Result);   //ResultCommandText
    //if DoCreate then  Self.SqlSubstitutions.Clear;
    CurPos := Value;
    Literal := False;
    EmbeddedLiteral := False;

    repeat
      while (CurPos^ in LeadBytes) do
        Inc(CurPos, 2);
      CurChar := CurPos^;
      if (CurChar = Self.SubstituteChar) and not Literal and ((CurPos + 1)^ <> Self.SubstituteChar) then
      begin
        StartPos := CurPos;
        while (CurChar <> #0) and (Literal or not NameDelimiter) do
          begin
            Inc(CurPos);
            while (CurPos^ in LeadBytes) do
              Inc(CurPos, 2);
            CurChar := CurPos^;
            if IsLiteral then
            begin
              Literal := Literal xor True;
              if CurPos = StartPos + 1 then
                EmbeddedLiteral := True;
            end;
          end;
          CurPos^ := #0;
          if EmbeddedLiteral then
            begin
              Name := StripLiterals(StartPos + 1);
              EmbeddedLiteral := False;
            end
          else
            Name := StrPas(StartPos + 1);
          //if DoCreate then
          //begin
            //if Self.SqlSubstitutions.FindSqlSubstitution(Name)= nil then
            //  Self.SqlSubstitutions.Add.Name:= Name;
            if substitutionList.FindSqlSubstitution(Name)= nil then
            begin
              substitutionList.Add.Name:= Name;
              //Try copy old substitutionvalue
              if Self.SqlSubstitutions.FindSqlSubstitution(Name)<> nil then
                TSqlSubstitution(substitutionList.FindSqlSubstitution(Name)).Value:=
                  TSqlSubstitution(Self.SqlSubstitutions.FindSqlSubstitution(Name)).Value;
            end;
          //end;
          CurPos^ := CurChar;
          substitutionValue := TSqlSubstitution(substitutionList.SqlSubstitutionByName(Name)).Value;
          newCommandText:=newCommandText+substitutionValue;
        end
      else
        if (CurChar = Self.SubstituteChar) and not Literal and ((CurPos + 1)^ = Self.SubstituteChar) then
          begin
            StrMove(CurPos, CurPos + 1, StrLen(CurPos) + 1);
          end
        else
          begin
            if IsLiteral then
              Literal := Literal xor True;
            newCommandText:=newCommandText+CurChar;
          end;
      Inc(CurPos);
    until CurChar = #0;
   if DoCreate then //Copy new substitution list
   begin
     Self.SqlSubstitutions.Clear;
     Self.SqlSubstitutions.AssignValues(substitutionList);
   end;
   Result:=newCommandText;
  finally
    substitutionList.Free;
  end;
end;

function TBS_ClientDataSet.SqlSubstitutionByName(
  const Value: string): TSqlSubstitution;
begin
  Result:=Self.fSqlSubstitutions.SqlSubstitutionByName(Value);
end;

procedure TBS_ClientDataSet.SubstituteSQL;
var
  substitutedSQL:String;
begin
  substitutedSQL:=Self.ParseSqlSubstitutions(Self.CommandText,false);
  inherited CommandText := substitutedSQL;
end;

procedure TBS_ClientDataSet.OpenCursor(InfoQuery: Boolean);
begin
  Self.fStoredCursor:=Screen.Cursor;
  if (WaitCursor <> crDefault) and (WaitCursor<>Self.fStoredCursor) then
    Screen.Cursor:=WaitCursor;
  try
    //SubstituteSQL;
    inherited OpenCursor(InfoQuery);
  finally
    if (WaitCursor <> crDefault) and (WaitCursor<>Self.fStoredCursor) then
      Screen.Cursor:=Self.fStoredCursor;
  end;
end;

procedure TBS_ClientDataSet.Execute;
begin
  SubstituteSQL;
  inherited Execute;
end;

function TBS_ClientDataSet.GetRealCommandText: String;
begin
  SubstituteSQL;
  Result:=inherited CommandText;
end;

function TBS_ClientDataSet.DoGetRecords(Count: Integer;
  out RecsOut: Integer; Options: Integer; const CommandText: WideString;
  Params: OleVariant): OleVariant;
var
  rcmdText: WideString;
begin
  rcmdText:=RealCommandText;
  if rcmdText = #0 then
    rcmdText:='';
  Result:= inherited DoGetRecords(Count, RecsOut, Options, rcmdText, Params);
end;

procedure TBS_ClientDataSet.DoExecute(Params: OleVariant);
begin
  //SubstituteSQL;
  inherited DoExecute(Params);
end;

procedure TBS_ClientDataSet.CommandTextChanged(Sender: TObject);
begin
  if not (csReading in ComponentState) then
    begin
      if SqlSubstitutionsCheck or (csDesigning in ComponentState) then
        begin
          Self.ParseSqlSubstitutions(CommandText, false);//may be true
        end;
      DataEvent(dePropertyChange, 0);
    end
end;

function TBS_ClientDataSet.GetInternalCommandText: WideString;
begin
  Result:=Self.FCommandStrings.Text;
end;

procedure TBS_ClientDataSet.SetInternalCommandText(Value: WideString);
begin
  Self.FCommandStrings.Text:=Value;
end;

end.
