{
Copyright  1999 by Delphi 5 Developer's Guide - Xavier Pacheco and Steve Teixeira
}

{$IFDEF VER110}
{$OBJEXPORTALL ON}
{$ENDIF}
{$D-L-}
unit BS_Clock;

interface

uses
  Windows, Messages, Controls, Forms, SysUtils, Classes, ExtCtrls;

type

  { Declare an event type which takes the sender of the event, and
    a TDateTime variable as parameters }
  TTimeEvent = procedure(Sender: TObject; DDGTime: TDateTime) of object;

  TBS_Clock = class(TPanel)
  private
    { Data fields }
    FHour,
    FMinute,
    FSecond: Word;
    FDateTime: TDateTime;
    FOldMinute,
    FOldSecond: Word;
    FTimer: TTimer;
    { Event handlers }
    FOnHour: TTimeEvent;       // Occurs on the hour
    FOnHalfPast: TTimeEvent;   // Occurs every half-hour
    FOnMinute: TTimeEvent;     // Occurs on the minute
    FOnSecond: TTimeEvent;     // Occurs every second
    FOnHalfMinute: TTimeEvent; // Occurs every 30 seconds
    { Define OnTimer event handler for internal TTimer, FTimer }
    procedure TimerProc(Sender: TObject);
  protected
    { Override the Paint methods }
    procedure Paint; override;

    { Define the various event dispatching methods }
    procedure DoHour(Tm: TDateTime); dynamic;
    procedure DoHalfPast(Tm: TDateTime); dynamic;
    procedure DoMinute(Tm: TDateTime); dynamic;
    procedure DoHalfMinute(Tm: TDateTime); dynamic;
    procedure DoSecond(Tm: TDateTime); dynamic;

  public
    { Override the Create constructor and Destroy destructor }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Define event properties }
    property OnHour: TTimeEvent read FOnHour write FOnHour;
    property OnHalfPast: TTimeEvent read FOnHalfPast write FOnHalfPast;
    property OnMinute: TTimeEvent read FOnMinute write FOnMinute;
    property OnHalfMinute: TTimeEvent read FOnHalfMinute
             write FOnHalfMinute;
    property OnSecond: TTimeEvent read FOnSecond write FOnSecond;
  end;

implementation

constructor TBS_Clock.Create(AOwner: TComponent);
begin
  inherited Create(AOwner); // Call the inherited constructor
  Height := 25; // Set default width and height properties
  Width := 120;
  BevelInner := bvLowered; // Set Default bevel properties
  BevelOuter := bvLowered;
  { Set the inherited Caption property to an empty string }
  inherited Caption := '';
  { Create the TTimer instance and set both its Interval property and
    OnTime event handler. }
  FTimer:= TTimer.Create(self);
  FTimer.interval:= 200;
  FTimer.OnTimer:= TimerProc;
end;


destructor TBS_Clock.Destroy;
begin
  FTimer.Free; // Free the TTimer instance.
  inherited Destroy; // Call inherited Destroy method
end;

procedure TBS_Clock.Paint;
begin
  inherited Paint; // Call the inherited Paint method
  { Now set the inherited Caption property to current time. }
  inherited Caption := TimeToStr(FDateTime);
end;

procedure TBS_Clock.TimerProc(Sender: TObject);
var
  HSec: Word;
begin
  { Save the old minute and second for later use }
  FOldMinute := FMinute;
  FOldSecond := FSecond;
  FDateTime := Now; // Get the current time.
  { Extract the individual time elements }
  DecodeTime(FDateTime, FHour, FMinute, FSecond, Hsec);

  refresh; // Redraw the component so that the new time is displayed.

  { Now call the event handlers depending on the time }
  if FMinute = 0 then
    DoHour(FDateTime);
  if FMinute = 30 then
    DoHalfPast(FDateTime);
  if (FMinute <> FOldMinute) then
    DoMinute(FDateTime);
  if FSecond <> FOldSecond then
    if ((FSecond = 30) or (FSecond = 0)) then
      DoHalfMinute(FDateTime)
    else
       DoSecond(FDateTime);
end;

{ The event dispatching methods below determine if component user has
  attached event handlers to the various clock events and calls them
  if they exist }
procedure TBS_Clock.DoHour(Tm: TDateTime);
begin
  if Assigned(FOnHour) then
   TTimeEvent(FOnHour)(Self, Tm);
end;

procedure TBS_Clock.DoHalfPast(Tm: TDateTime);
begin
  if Assigned(FOnHalfPast) then
   TTimeEvent(FOnHalfPast)(Self, Tm);
end;

procedure TBS_Clock.DoMinute(Tm: TDateTime);
begin
  if Assigned(FOnMinute) then
   TTimeEvent(FOnMinute)(Self, Tm);
end;

procedure TBS_Clock.DoHalfMinute(Tm: TDateTime);
begin
  if Assigned(FOnHalfMinute) then
   TTimeEvent(FOnHalfMinute)(Self, Tm);
end;

procedure TBS_Clock.DoSecond(Tm: TDateTime);
begin
  if Assigned(FOnSecond) then
   TTimeEvent(FOnSecond)(Self, Tm);
end;

end.
