// Copyright 2000 Siarhej Berdachuk
// bs@robot.belpak.brest.by

{$I BS.inc}
{$R BS_DBNAVIGATOR.RES}

unit BS_DBNavigator;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DB, WinTypes, WinProcs, ExtCtrls, Buttons
  {$IFDEF CompilerVersion6},VDBConsts
  //{$ELSE}
  {$ENDIF}
  ;

const
  InitRepeatPause = 400;  // pause before repeat timer (ms)
  RepeatPause     = 100;  // pause before hint window displays (ms)
  SpaceSize       = 5;    // size of space between special NavControls

ResourceString
  SRecordProperty  = 'Property';


type
  TBS_NavControl = class;
  TBS_NavDataLink = class;

  TBS_DBNavKind = (dbnHorizontal, dbnVertical);

  TBSNavGlyph = (ngEnabled, ngDisabled);
  TBSNavigateControl = (nbPost, nbCancel, nbInsert, nbDelete,
     nbEdit, nbProperty, nbRefresh, nbFirst, nbPrior, nbNext, nbLast);


  TBSNavControlSet = set of TBSNavigateControl;
  TBSNavControlStyle = set of (nsAllowTimer, nsFocusRect);

  EBSNavClick = procedure (Sender: TObject; Button: TBSNavigateControl) of object;
  EBSNavBtnClick = procedure (Sender: TObject) of object;

  TBS_DBNavigator = class (TCustomPanel)
  private
    FAsChild:boolean;
    FDataLink: TBS_NavDataLink;
    FVisibleControls: TBSNavControlSet;
    FHints: TStrings;
    DBNavControlWidth: Integer;
    DBNavControlHeight: Integer;
    MinBSNavControlSize: TPoint;
    FOnNavClick: EBSNavClick;
    FOnPropertyBtnClick: EBSNavBtnClick;    
    FOnPostBtnClick: EBSNavBtnClick;
    FOnCancelBtnClick: EBSNavBtnClick;
    FOnInsertBtnClick: EBSNavBtnClick;
    FOnDeleteBtnClick: EBSNavBtnClick;
    FOnEditBtnClick: EBSNavBtnClick;
    FOnRefreshBtnClick: EBSNavBtnClick;
    FOnFirstBtnClick: EBSNavBtnClick;
    FOnPriorBtnClick: EBSNavBtnClick;
    FOnNextBtnClick: EBSNavBtnClick;
    FOnLastBtnClick: EBSNavBtnClick;
    FBeforeAction: EBSNavClick;
    FocusedNavControl: TBSNavigateControl;
    FConfirmDelete: Boolean;
    FFlat: Boolean;
    Silent: Boolean;
    FKind: TBS_DBNavKind;
    property AsChild:boolean read FAsChild write FAsChild default False;
    procedure SetKind(Value: TBS_DBNavKind);
    procedure ClickHandler(Sender: TObject);
    function GetDataSource: TDataSource;
    procedure SetDataSource(Value: TDataSource);
    procedure InitNavControls;
    procedure InitHints;
    procedure BtnMouseDown (Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SetVisible(Value: TBSNavControlSet);
    procedure SetSize (var W: Integer; var H: Integer);
    procedure HintsChanged(Sender: TObject);
    procedure SetHints(Value: TStrings);
    procedure SetFlat(Value: Boolean);
    procedure WMSize(var Message: TWMSize);  message WM_SIZE;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
  protected
    procedure DataChanged;
    procedure EditingChanged;
    procedure ActiveChanged;
    procedure Loaded; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
    function  GetBSDataSet: TDataSet;
  public
    NavControls: array[TBSNavigateControl] of TBS_NavControl;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure BtnClick(Index: TBSNavigateControl);
    constructor Create(AOwner:TComponent);overload;override;
    constructor Create(AOwner:TComponent; AParent:TwinControl);reintroduce;overload;
  published
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property VisibleControls: TBSNavControlSet read FVisibleControls write SetVisible;

    property Align;
    property Kind: TBS_DBNavKind read FKind write SetKind default dbnHorizontal;
    {$IFDEF CompilerVersion4}
      property Anchors;
      property Constraints;
      property DragKind;
      property OnEndDock;
      property OnStartDock;
    {$ENDIF}
    property DragCursor;
    property DragMode;
    property Enabled;
    property Flat: Boolean read FFlat write SetFlat default False;
    property Ctl3D;
    property Hints: TStrings read FHints write SetHints;
    property ParentCtl3D;
    property ParentShowHint;
    property PopupMenu;
    property ConfirmDelete: Boolean read FConfirmDelete write FConfirmDelete default True;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property BeforeAction: EBSNavClick read FBeforeAction write FBeforeAction;
    property OnClick: EBSNavClick read FOnNavClick write FOnNavClick;
    property OnPropertyBtnClick: EBSNavBtnClick read FOnPropertyBtnClick write FOnPropertyBtnClick;
    property OnPostBtnClick: EBSNavBtnClick read FOnPostBtnClick write FOnPostBtnClick;
    property OnCancelBtnClick: EBSNavBtnClick read FOnCancelBtnClick write FOnCancelBtnClick;
    property OnInsertBtnClick: EBSNavBtnClick read FOnInsertBtnClick write FOnInsertBtnClick;
    property OnDeleteBtnClick: EBSNavBtnClick read FOnDeleteBtnClick write FOnDeleteBtnClick;
    property OnEditBtnClick: EBSNavBtnClick read FOnEditBtnClick write FOnEditBtnClick;
    property OnRefreshBtnClick: EBSNavBtnClick read FOnRefreshBtnClick write FOnRefreshBtnClick;
    property OnFirstBtnClick: EBSNavBtnClick read FOnFirstBtnClick write FOnFirstBtnClick;
    property OnPriorBtnClick: EBSNavBtnClick read FOnPriorBtnClick write FOnPriorBtnClick;
    property OnNextBtnClick: EBSNavBtnClick read FOnNextBtnClick write FOnNextBtnClick;
    property OnLastBtnClick: EBSNavBtnClick read FOnLastBtnClick write FOnLastBtnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnResize;
    property OnStartDrag;
  end;

  TBS_NavControl = class(TSpeedButton)
  private
    FIndex: TBSNavigateControl;
    FNavStyle: TBSNavControlStyle;
    FRepeatTimer: TTimer;
    procedure TimerExpired(Sender: TObject);
  protected
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
    destructor Destroy; override;
    property NavStyle: TBSNavControlStyle read FNavStyle write FNavStyle;
    property Index : TBSNavigateControl read FIndex write FIndex;
  end;

  TBS_NavDataLink = class(TDataLink)
  private
    FNavigator: TBS_DBNavigator;
  protected
    procedure EditingChanged; override;
    procedure DataSetChanged; override;
    procedure ActiveChanged; override;
  public
    constructor Create(ANav: TBS_DBNavigator);
    destructor Destroy; override;
  end;


implementation

uses DBConsts;

// TBS_DBNavigator

var
  NavControlTypeName: array[TBSNavigateControl] of PChar = (
    'POST',
    'CANCEL',
    'INSERT',
    'DELETE',
    'EDIT',
    'PROPERTY',
    'REFRESH',
    'FIRST',
    'PRIOR',
    'NEXT',
    'LAST');
{$D-L-}
constructor TBS_DBNavigator.Create(AOwner: TComponent);
begin
  AsChild:=False;
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csAcceptsControls, csSetCaption] + [csOpaque];
  if not NewStyleControls then ControlStyle := ControlStyle + [csFramed];
  FDataLink := TBS_NavDataLink.Create(Self);
  //FKind := dbnVertical;
  FVisibleControls := [nbInsert, nbDelete, nbPost, nbCancel, nbRefresh, nbEdit];
  FHints := TStringList.Create;
  TStringList(FHints).OnChange := HintsChanged;

  BevelOuter := bvNone;
  BevelInner := bvNone;

  Width := 6 * 24 + 1;
  Height := 25;
  FFlat := True;
  InitNavControls;
  DBNavControlWidth := 0;
  FocusedNavControl := nbFirst;
  FConfirmDelete := True;
  FullRepaint := False;
  Silent := False;
  Self.Cursor:=crHandPoint;
end;

constructor TBS_DBNavigator.Create(AOwner:TComponent; AParent:TwinControl);
begin
  AsChild:=True;
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csAcceptsControls, csSetCaption] + [csOpaque];
  if not NewStyleControls then ControlStyle := ControlStyle + [csFramed];
  FDataLink := TBS_NavDataLink.Create(Self);
  //FKind := dbnVertical;
  FVisibleControls := [nbInsert, nbDelete, nbPost, nbCancel, nbRefresh, nbEdit];
    FHints := TStringList.Create;
  TStringList(FHints).OnChange := HintsChanged;

  BevelOuter := bvNone;
  BevelInner := bvNone;

  Width := 6 * 24 + 1;
  Height := 25;
  Self.Cursor:=crHandPoint;
  FFlat := True;
  InitNavControls;
  DBNavControlWidth := 0;
  FocusedNavControl := nbFirst;
  FConfirmDelete := True;
  FullRepaint := False;
  Silent := False;
  Parent:=AParent;

  Self.Cursor:=crHandPoint;
  //align:=alClient;
end;


destructor TBS_DBNavigator.Destroy;
begin
  FDataLink.Free;
  FHints.Free;
  FDataLink := nil;
  inherited Destroy;
end;

procedure TBS_DBNavigator.SetKind(Value: TBS_DBNavKind);
begin
  if FKind <> Value then
  begin
    FKind := Value;
    if not (csLoading in ComponentState) then
      SetBounds(Left, Top, Height, Width);
    RecreateWnd;
  end;
end;


procedure TBS_DBNavigator.InitNavControls;
var I: TBSNavigateControl;
    Btn: TBS_NavControl;
    X: Integer;
    Y: Integer;
    //ResName: string;
begin
  MinBSNavControlSize := Point(20, 18);
  X := 0;
  Y := 0;
  for I := Low(NavControls) to High(NavControls) do
  begin
    Btn := TBS_NavControl.Create (Self);
    Btn.Flat := Flat;
    Btn.Index := I;
    Btn.Visible := I in FVisibleControls;
    Btn.Enabled := True;
    Btn.SetBounds (X, Y, MinBSNavControlSize.X, MinBSNavControlSize.Y);

    Btn.NumGlyphs := 2;
    Btn.Enabled := False;
    Btn.Enabled := True;
    Btn.OnClick := ClickHandler;
    Btn.OnMouseDown := BtnMouseDown;
    Btn.Parent := Self;
    NavControls[I] := Btn;
    case Kind of
      dbnHorizontal:
        X := X + MinBSNavControlSize.X;
      dbnVertical:
        Y := Y + MinBSNavControlSize.Y;
    end;
  end;
  InitHints;
  NavControls[nbPrior].NavStyle := NavControls[nbPrior].NavStyle + [nsAllowTimer];
  NavControls[nbNext].NavStyle  := NavControls[nbNext].NavStyle + [nsAllowTimer];
end;

procedure TBS_DBNavigator.InitHints;
var i: Integer;
    j: TBSNavigateControl;
begin
  NavControls[nbProperty].Hint   := SRecordProperty;
  NavControls[nbFirst].Hint      := SFirstRecord;
  NavControls[nbPrior].Hint      := SPriorRecord;
  NavControls[nbNext].Hint       := SNextRecord;
  NavControls[nbLast].Hint       := SLastRecord;
  NavControls[nbInsert].Hint     := SInsertRecord;
  NavControls[nbDelete].Hint     := SDeleteRecord;
  NavControls[nbEdit].Hint       := SEditRecord;
  NavControls[nbPost].Hint       := SPostEdit;
  NavControls[nbCancel].Hint     := SCancelEdit;
  NavControls[nbRefresh].Hint    := SRefreshRecord;
  j := Low(NavControls);
  for i := 0 to (FHints.Count - 1) do
  begin
    if FHints.Strings[i] <> '' then NavControls[j].Hint := FHints.Strings[i];
    if j = High(NavControls) then Exit;
    Inc(j);
  end;
end;

procedure TBS_DBNavigator.HintsChanged(Sender: TObject);
begin
  InitHints;
end;

procedure TBS_DBNavigator.SetFlat(Value: Boolean);
var I: TBSNavigateControl;
begin
  if FFlat <> Value then
  begin
    FFlat := Value;
    for I := Low(NavControls) to High(NavControls) do NavControls[I].Flat := Value;
  end;
end;

procedure TBS_DBNavigator.SetHints(Value: TStrings);
begin
  FHints.Assign(Value);
end;

procedure TBS_DBNavigator.GetChildren(Proc: TGetChildProc; Root: TComponent);
begin
//
end;

procedure TBS_DBNavigator.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink <> nil) and
    (AComponent = DataSource) then DataSource := nil;
end;

procedure TBS_DBNavigator.SetVisible(Value: TBSNavControlSet);
var I: TBSNavigateControl;
    W, H: Integer;
begin
  W := Width;
  H := Height;
  FVisibleControls := Value;
  for I := Low(NavControls) to High(NavControls) do  NavControls[I].Visible := I in FVisibleControls;
  SetSize (W, H);
  if (W <> Width) or (H <> Height) then inherited SetBounds (Left, Top, W, H);
  Invalidate;
end;

procedure TBS_DBNavigator.SetSize (var W: Integer; var H: Integer);
var Count: Integer;
    MinW: Integer;
    MinH: Integer;
    I: TBSNavigateControl;
    Space, Temp, Remain: Integer;
    X,Y: Integer;
    ResName: string;
begin
  if (csLoading in ComponentState) then Exit;
  if NavControls[nbFirst] = nil then Exit;
  Count := 0;
  for I := Low(NavControls) to High(NavControls) do
    if NavControls[I].Visible then Inc(Count);
  if Count = 0 then Inc(Count);

  case Kind of
    dbnHorizontal:
      begin
        MinW := Count * MinBSNavControlSize.X;
        if W < MinW then
          W := MinW;
        DBNavControlWidth := W div Count;
        Temp := Count * DBNavControlWidth;

        if Align = alNone then W := Temp;
        X := 0;
        Remain := W - Temp;
        Temp := Count div 2;
        for I := Low(NavControls) to High(NavControls) do
        begin
          if NavControls[I].Visible then
          begin
            FmtStr(ResName, 'DBNH_%s', [NavControlTypeName[I]]);
            NavControls[I].Glyph.LoadFromResourceName(HInstance, ResName);
            Space := 0;
            if Remain <> 0 then
            begin
              Dec(Temp, Remain);
              if Temp < 0 then
              begin
                Inc(Temp, Count);
                Space := 1;
              end;
           end;
           NavControls[I].SetBounds(X, 0, DBNavControlWidth + Space, Height);
           Inc(X, DBNavControlWidth + Space);
         end else
           NavControls[I].SetBounds (Width + 1, 0, DBNavControlWidth, Height);
        end;
      end;
    dbnVertical:
      begin
        MinH := Count * MinBSNavControlSize.Y;
        if H < MinH then
          H := MinH;
        DBNavControlHeight := H div Count;
        Temp := Count * DBNavControlHeight;

        if Align = alNone then
          H := Temp;
        Y := 0;
        Remain := H - Temp;
        Temp := Count div 2;
        for I := Low(NavControls) to High(NavControls) do
        begin
          FmtStr(ResName, 'DBNV_%s', [NavControlTypeName[I]]);
          NavControls[I].Glyph.LoadFromResourceName(HInstance, ResName);

          if NavControls[I].Visible then
          begin
            Space := 0;
            if Remain <> 0 then
            begin
              Dec(Temp, Remain);
              if Temp < 0 then
              begin
                Inc(Temp, Count);
                Space := 1;
              end;
           end;
           NavControls[I].SetBounds(0, Y, Width, DBNavControlHeight + Space);
           Inc(Y, DBNavControlHeight + Space);
         end else
           NavControls[I].SetBounds (0, Height+1, Width, DBNavControlHeight);
        end;
      end;
    end;
end;

procedure TBS_DBNavigator.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var W, H: Integer;
begin
  W := AWidth;
  H := AHeight;
  if not HandleAllocated then
    SetSize (W, H);
  inherited SetBounds (ALeft, ATop, W, H);
end;

procedure TBS_DBNavigator.WMSize(var Message: TWMSize);
var W, H: Integer;
begin
  inherited;
  // check for minimum size
  W := Width;
  H := Height;
  SetSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds(Left, Top, W, H);
  Message.Result := 0;
end;


procedure TBS_DBNavigator.BtnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var OldFocus: TBSNavigateControl;
begin
  OldFocus := FocusedNavControl;
  FocusedNavControl := TBS_NavControl (Sender).Index;
  if TabStop and (GetFocus <> Handle) and CanFocus then
  begin
    SetFocus;
    if (GetFocus <> Handle) then  Exit;
  end else begin
    if TabStop and (GetFocus = Handle) and (OldFocus <> FocusedNavControl) then
    begin
      NavControls[OldFocus].Invalidate;
      NavControls[FocusedNavControl].Invalidate;
    end;
  end;
end;

function TBS_DBNavigator.GetBSDataSet: TDataSet;
begin
  Result := nil;
  if (DataSource <> nil) and (DataSource.DataSet is TDataSet) then
    Result := TDataSet(DataSource.DataSet);
end;


procedure TBS_DBNavigator.WMSetFocus(var Message: TWMSetFocus);
begin
  NavControls[FocusedNavControl].Invalidate;
end;

procedure TBS_DBNavigator.WMKillFocus(var Message: TWMKillFocus);
begin
  NavControls[FocusedNavControl].Invalidate;
end;

procedure TBS_DBNavigator.KeyDown(var Key: Word; Shift: TShiftState);
var NewFocus: TBSNavigateControl;
    OldFocus: TBSNavigateControl;
begin
  OldFocus := FocusedNavControl;
  case Key of
    VK_RIGHT:
      begin
        NewFocus := FocusedNavControl;
        repeat
          if NewFocus < High(NavControls) then
            NewFocus := Succ(NewFocus);
        until (NewFocus = High(NavControls)) or (NavControls[NewFocus].Visible);
        if NewFocus <> FocusedNavControl then
        begin
          FocusedNavControl := NewFocus;
          NavControls[OldFocus].Invalidate;
          NavControls[FocusedNavControl].Invalidate;
        end;
      end;
    VK_LEFT:
      begin
        NewFocus := FocusedNavControl;
        repeat
          if NewFocus > Low(NavControls) then
            NewFocus := Pred(NewFocus);
        until (NewFocus = Low(NavControls)) or (NavControls[NewFocus].Visible);
        if NewFocus <> FocusedNavControl then
        begin
          FocusedNavControl := NewFocus;
          NavControls[OldFocus].Invalidate;
          NavControls[FocusedNavControl].Invalidate;
        end;
      end;
    VK_SPACE:
      begin
        if NavControls[FocusedNavControl].Enabled then
          NavControls[FocusedNavControl].Click;
      end;
  end;
end;

procedure TBS_DBNavigator.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS;
end;

procedure TBS_DBNavigator.DataChanged;
var UpEnable, DnEnable: Boolean;
    BSDataSet: TDataSet;
begin
  BSDataSet := GetBSDataSet;
  if (BSDataSet <> nil) then
    begin
      UpEnable := Enabled and FDataLink.Active and not FDataLink.DataSet.BOF;
      DnEnable := Enabled and FDataLink.Active and not FDataLink.DataSet.EOF;
    end
  else
    begin
      UpEnable := False;
      DnEnable := False;
    end;

  NavControls[nbFirst].Enabled := UpEnable;
  NavControls[nbPrior].Enabled := UpEnable;
  NavControls[nbNext].Enabled := DnEnable;
  NavControls[nbLast].Enabled := DnEnable;
  NavControls[nbDelete].Enabled := Enabled and FDataLink.Active and FDataLink.DataSet.CanModify
                and  not (FDataLink.DataSet.BOF and FDataLink.DataSet.EOF);
  if BSDataSet <> nil then
  begin
    EditingChanged;
  end;
end;

procedure TBS_DBNavigator.EditingChanged;
var CanModify: Boolean;
    BSDataSet: TDataSet;
begin
  Silent := True;
  BSDataSet := GetBSDataSet;

  if (BSDataSet <> nil)then
  begin
    CanModify := Enabled and FDataLink.Active and FDataLink.DataSet.CanModify;
    NavControls[nbInsert].Enabled := CanModify;
    NavControls[nbEdit].Enabled := CanModify and not FDataLink.Editing;
    NavControls[nbPost].Enabled := CanModify and FDataLink.Editing;
    NavControls[nbCancel].Enabled := CanModify and FDataLink.Editing;
    NavControls[nbRefresh].Enabled := True;
    NavControls[nbProperty].Enabled := (FDataLink.DataSet.RecordCount>0);
  end;
  Silent := False;
end;

procedure TBS_DBNavigator.ActiveChanged;
var I: TBSNavigateControl;
begin
  if not (Enabled and FDataLink.Active) then
    for I := Low(NavControls) to High(NavControls) do NavControls[I].Enabled := False
  else begin
    DataChanged;
    EditingChanged;
  end;
end;

procedure TBS_DBNavigator.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if not (csLoading in ComponentState) then ActiveChanged;
end;

procedure TBS_DBNavigator.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
  if not (csLoading in ComponentState) then ActiveChanged;
  if Value <> nil then Value.FreeNotification(Self);
end;

function TBS_DBNavigator.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TBS_DBNavigator.Loaded;
var W, H: Integer;
begin
  inherited Loaded;
  W := Width;
  H := Height;
  SetSize (W, H);
  if (W <> Width) or (H <> Height) then inherited SetBounds (Left, Top, W, H);
  InitHints;
  ActiveChanged;
end;

procedure TBS_DBNavigator.ClickHandler(Sender: TObject);
begin
  BtnClick(TBS_NavControl (Sender).Index);
end;


procedure TBS_DBNavigator.BtnClick(Index: TBSNavigateControl);
//var BSDataSet: TDataSet;
begin
  if Silent then Exit;
  if (DataSource <> nil) and (DataSource.State <> dsInactive) then
  begin
    //BSDataSet := GetBSDataSet;
    if not (csDesigning in ComponentState) and Assigned(FBeforeAction) then
      FBeforeAction(Self, Index);
    with DataSource.DataSet do
    begin
      case Index of
       nbProperty:
                if Assigned(FOnPropertyBtnClick) then
                  FOnPropertyBtnClick(Self);
       nbPrior: if Assigned(FOnPriorBtnClick) then
                  FOnPriorBtnClick(Self)
                else
                  Prior;
        nbNext: if Assigned(FOnNextBtnClick) then
                  FOnNextBtnClick(Self)
                else
                  Next;
       nbFirst: if Assigned(FOnFirstBtnClick) then
                  FOnFirstBtnClick(Self)
                else
                  First;
        nbLast: if Assigned(FOnLastBtnClick) then
                  FOnLastBtnClick(Self)
                else
                  Last;
      nbInsert: if Assigned(FOnInsertBtnClick) then
                  FOnInsertBtnClick(Self)
                else
                  Insert;
        nbEdit: if Assigned(FOnEditBtnClick) then
                  FOnEditBtnClick(Self)
                else
                  Edit;
      nbCancel: if Assigned(FOnCancelBtnClick) then
                  FOnCancelBtnClick(Self)
                else
                  Cancel;
        nbPost: if Assigned(FOnPostBtnClick) then
                  FOnPostBtnClick(Self)
                else
                  Post;
      nbRefresh: if Assigned(FOnRefreshBtnClick) then
                  FOnRefreshBtnClick(Self)
                else
                  Refresh;
      nbDelete: if not FConfirmDelete or
                  (MessageDlg(SDeleteRecordQuestion, mtConfirmation,
                    mbOKCancel, 0) <> idCancel) then
                begin
                  if Assigned(FOnDeleteBtnClick) then
                    FOnDeleteBtnClick(Self)
                  else
                    Delete;
                end;
      end;
    end;
  end;
  if not (csDesigning in ComponentState) and Assigned(FOnNavClick) then
    FOnNavClick(Self, Index);
end;



// TBS_NavControl

destructor TBS_NavControl.Destroy;
begin
  if FRepeatTimer <> nil then FRepeatTimer.Free;
  inherited Destroy;
end;

procedure TBS_NavControl.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if nsAllowTimer in FNavStyle then
  begin
    if FRepeatTimer = nil then FRepeatTimer := TTimer.Create(Self);
    FRepeatTimer.OnTimer  := TimerExpired;
    FRepeatTimer.Interval := InitRepeatPause;
    FRepeatTimer.Enabled  := True;
  end;
end;

procedure TBS_NavControl.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseUp (Button, Shift, X, Y);
  if FRepeatTimer <> nil then FRepeatTimer.Enabled  := False;
end;

procedure TBS_NavControl.TimerExpired(Sender: TObject);
begin
  FRepeatTimer.Interval := RepeatPause;
  if (FState = bsDown) and MouseCapture then
  begin
    try
      Click;
    except
      FRepeatTimer.Enabled := False;
      raise;
    end;
  end;
end;

procedure TBS_NavControl.Paint;
var R: TRect;
begin
  inherited Paint;
  if (GetFocus = Parent.Handle) and
     (FIndex = TBS_DBNavigator (Parent).FocusedNavControl) then
  begin
    R := Bounds(0, 0, Width, Height);
    InflateRect(R, -3, -3);
    if FState = bsDown then OffsetRect(R, 1, 1);
    DrawFocusRect(Canvas.Handle, R);
  end;
end;



// TBS_NavDataLink

constructor TBS_NavDataLink.Create(ANav: TBS_DBNavigator);
begin
  inherited Create;
  FNavigator := ANav;
end;

destructor TBS_NavDataLink.Destroy;
begin
  FNavigator := nil;
  inherited Destroy;
end;

procedure TBS_NavDataLink.EditingChanged;
begin
  if FNavigator <> nil then
    FNavigator.EditingChanged;
end;

procedure TBS_NavDataLink.DataSetChanged;
begin
  if FNavigator <> nil then
    FNavigator.DataChanged;
end;

procedure TBS_NavDataLink.ActiveChanged;
begin
  if FNavigator <> nil then
    FNavigator.ActiveChanged;
end;

end.
