unit BS_GridProEditor;

{$I BS.Inc}
interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
{$IFDEF CompilerVersion4} ImgList,{$ENDIF}
  StdCtrls, ComCtrls, ToolWin, Menus,
  BS_DBGrid,
  DBGrids,
  DsgnIntf,
  DsgnWnds,
  LibIntf {$IFNDEF CompilerVersion3} ,ColnEdit{$ENDIF} {, dbugintf};

type
  TBS_DBGridColumnsEditor = class(TDesignWindow)
    ToolBar1: TToolBar;
    tbAddNew: TToolButton;
    tbDeleteSelected: TToolButton;
    ToolButton3: TToolButton;
    tbAddAllFields: TToolButton;
    tbRestoreDefaults: TToolButton;
    ImageList1: TImageList;
    PopupMenu1: TPopupMenu;
    miAddNew: TMenuItem;
    miDeleteSelected: TMenuItem;
    miAddAllFields: TMenuItem;
    miRestoreDefaults: TMenuItem;
    miSelectAll: TMenuItem;
    N1: TMenuItem;
    ListView1: TListView;
    tbMoveUp: TToolButton;
    tbMoveDown: TToolButton;
    ToolButton4: TToolButton;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure tbAddNewClick(Sender: TObject);
    procedure tbDeleteSelectedClick(Sender: TObject);
    procedure tbAddAllFieldsClick(Sender: TObject);
    procedure tbRestoreDefaultsClick(Sender: TObject);
    procedure ListView1Change(Sender: TObject; Item: TListItem;
      Change: TItemChange);
    procedure FormResize(Sender: TObject);
    procedure miAddNewClick(Sender: TObject);
    procedure miDeleteSelectedClick(Sender: TObject);
    procedure miSelectAllClick(Sender: TObject);
    procedure miAddAllFieldsClick(Sender: TObject);
    procedure miRestoreDefaultsClick(Sender: TObject);
    procedure tbMoveUpClick(Sender: TObject);
    procedure tbMoveDownClick(Sender: TObject);
  private
    FBS_DBGrid:TBS_DBGrid;
    function CheckCollection: Boolean;
    procedure SetBS_DBGrid(const Value: TBS_DBGrid);
    procedure UpdateData;
    procedure Copy;
    procedure Cut;
    procedure Paste;
  protected
    function UniqueName(Component: TComponent): string; override;
    procedure Activated; override;
    { Private declarations }
  public
    procedure EditAction(Action: TEditAction); override;
    procedure FormModified; override;
    procedure FormClosed(Form: TCustomForm); override;
    function GetEditState: TEditState; override;
{$IFDEF CompilerVersion4} {Borland Delphi 4.0 or C++ Builder 4.0}
    procedure ComponentDeleted(Component: IPersistent); override;
{$ELSE}
    procedure ComponentDeleted(Component: TComponent); override;
{$ENDIF}

    property BS_DBGrid: TBS_DBGrid read FBS_DBGrid
         write SetBS_DBGrid;
    { Public declarations }
  end;


{ TBS_DBGridColumnsProperty }

  TBS_DBGridColumnsProperty = class(TPropertyEditor)
    function GetAttributes: TPropertyAttributes; override;
    function GetValue: string; override;
    procedure Edit; override;
  end;

{ TBS_DBGridEditor }

  TBS_DBGridEditor = class(TComponentEditor)
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

{ TBS_DBGridFieldProperty }

  TBS_DBGridFieldProperty = class(TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValueList(List: TStrings);
    procedure GetValues(Proc: TGetStrProc); override;
  end;

{$IFNDEF CompilerVersion3}
  TBS_ColumnsFootersProperty = class(TCollectionProperty)
  public
    procedure Edit; override;
  end;

  TBS_ColumnsFootersEditor = class(TCollectionEditor)
  public
    procedure ComponentDeleted(Component: TComponent); override;
  end;
{$ENDIF}

var
  BS_DBGridColumnsEditor: TBS_DBGridColumnsEditor;

//procedure ShowItemsEditor({Designer: TDesigner;}
//  ABS_DBGrid: TBS_DBGrid);

implementation

{$R *.DFM}

{$IFDEF CompilerVersion4} {Borland Delphi 4.0 or C++ Builder 4.0}
type
  TDesigner = IDesigner;
  TFormDesigner = IFormDesigner;
{$ENDIF}

{$IFNDEF CompilerVersion5}
type
  TDesignerSelectionList = TComponentList;
{$ENDIF}


{$IFNDEF CompilerVersion3}
type
  TPersistentCracker = class(TPersistent);

procedure TBS_ColumnsFootersProperty.Edit;
var
  Obj: TPersistent;
begin
  Obj := GetComponent(0);
  while (Obj <> nil) and not (Obj is TComponent) do
    Obj := TPersistentCracker(Obj).GetOwner;
  ShowCollectionEditorClass(Designer,  GetEditorClass,
    TComponent(Obj), TCollection(GetOrdValue), GetName);
end;

procedure TBS_ColumnsFootersEditor.ComponentDeleted(Component: TComponent);
begin
  inherited;
  if Self.Component = Component then
  begin
    Collection := nil;  // Component is already in its destructor; collection is gone
    Self.Component := nil;
    Close;
  end;
end;
{$ENDIF}

function FindEditor(ABS_DBGrid: TBS_DBGrid): TBS_DBGridColumnsEditor;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to Screen.FormCount - 1 do begin
    if Screen.Forms[I] is TBS_DBGridColumnsEditor then begin
      if TBS_DBGridColumnsEditor(Screen.Forms[I]).BS_DBGrid = ABS_DBGrid then
      begin
        Result := TBS_DBGridColumnsEditor(Screen.Forms[I]);
        Break;
      end;
    end;
  end;
end;

procedure ShowItemsEditor(Designer: TDesigner;
  ABS_DBGrid: TBS_DBGrid);
var
  Editor: TBS_DBGridColumnsEditor;
begin
  if ABS_DBGrid = nil then Exit;
  Editor := FindEditor(ABS_DBGrid);
  if Editor = nil then begin
    Editor := TBS_DBGridColumnsEditor.Create(Application);
    try
      Editor.Designer := TFormDesigner(Designer);
      Editor.BS_DBGrid := ABS_DBGrid;
      Editor.Show;
    except
      Editor.Free;
      raise;
    end;
  end
  else begin
    Editor.Show;
    if Editor.WindowState = wsMinimized then
      Editor.WindowState := wsNormal;
  end;
end;


{ TDBGridColumnsEditor }

function TBS_DBGridColumnsEditor.CheckCollection: Boolean;
begin
  Result := (BS_DBGrid <> nil) and (BS_DBGrid.Columns <> nil)
    and (Designer.Form <> nil);
end;

procedure TBS_DBGridColumnsEditor.SetBS_DBGrid(const Value: TBS_DBGrid);
begin
  if FBS_DBGrid <> Value then begin
    FBS_DBGrid := Value;
    UpdateData;
  end;
end;

procedure TBS_DBGridColumnsEditor.UpdateData;
var
  Empty,Modified: Boolean;
  i:Integer;
  ListItem: TListItem;
begin
  if CheckCollection then begin
    Caption := Format('Editing %s.Columns', [BS_DBGrid.Name]);
    Empty := BS_DBGrid.Columns.Count = 0;
  end
  else Empty := True;
  if Empty then begin
    ListView1.Items.Clear;
  end
  else begin
    Modified := False;
    if (BS_DBGrid.Columns.Count <> ListView1.Items.Count) then Modified := True;
    if (Modified = False) then
      for i := 0 to BS_DBGrid.Columns.Count - 1 do
        if (ListView1.Items[i].Data <> BS_DBGrid.Columns[i]) then begin
          Modified := True;
          Break;
        end;
    if (Modified = True) then begin
      ListView1.Items.BeginUpdate;
      ListView1.Items.Clear;
      try
        if (BS_DBGrid.Columns.State = csCustomized) then begin
           for i := 0 to BS_DBGrid.Columns.Count - 1 do begin
             ListItem := ListView1.Items.Add;
             ListItem.Caption := IntToStr(i) + ' - ' + BS_DBGrid.Columns[i].DisplayName;
             ListItem.Data := BS_DBGrid.Columns[i];
           end;
        end;
      finally
        ListView1.Items.EndUpdate;
      end;
    end else //      FieldName
      for i := 0 to BS_DBGrid.Columns.Count - 1 do ListView1.Items[i].Caption := IntToStr(i) + ' - ' + BS_DBGrid.Columns[i].DisplayName;
  end;
  ListView1Change(nil,nil,ctState);
end;

procedure TBS_DBGridColumnsEditor.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
//  Action := caFree;
  Free;
end;

procedure TBS_DBGridColumnsEditor.FormCreate(Sender: TObject);
begin
  BS_DBGrid := nil;
  if NewStyleControls then Font.Style := [];
end;

procedure TBS_DBGridColumnsEditor.tbAddNewClick(Sender: TObject);
begin
  BS_DBGrid.Columns.Add;
  UpdateData;
  Designer.Modified;
  ListView1.Items[ListView1.Items.Count-1].Selected := True;
  ListView1Change(nil,nil,ctState);
end;

procedure TBS_DBGridColumnsEditor.tbDeleteSelectedClick(Sender: TObject);
var i,sel:Integer;
//  ListItem: TListItem;
  FComponents: TDesignerSelectionList;
begin
  sel := -1;
  if (ListView1.SelCount > 0) then begin
    FComponents := TDesignerSelectionList.Create;
    try
      BS_DBGrid.Columns.BeginUpdate;
      ListView1.Items.BeginUpdate;
      sel := ListView1.Items.IndexOf(ListView1.Selected);
      for i := ListView1.Items.Count - 1 downto 0 do
       if (ListView1.Items[i].Selected = True) and Assigned(ListView1.Items[i].Data) then begin
         FComponents.Add(TBS_Column(ListView1.Items[i].Data));
       end;
      ListView1.Items.Clear;

      for i := 0 to FComponents.Count - 1 do FComponents[i].Free;

    finally
      ListView1.Items.EndUpdate;
      BS_DBGrid.Columns.EndUpdate;
      FComponents.Free;
      Designer.Modified;
      if (ListView1.Items.Count > 0) then
        if (sel > ListView1.Items.Count - 1) then
           ListView1.Items[ListView1.Items.Count-1].Selected := True
        else
           ListView1.Items[sel].Selected := True;
      ListView1Change(nil,nil,ctState);
    end;
   end;
end;

procedure TBS_DBGridColumnsEditor.tbAddAllFieldsClick(Sender: TObject);
var msgValue:Word;
    i:Integer;
    Col:TBS_Column;
begin
  if (BS_DBGrid.Columns.State = csDefault) then
    BS_DBGrid.Columns.State := csCustomized
  else begin
   if (BS_DBGrid.Columns.Count > 0) then begin
     msgValue := MessageDlg('Delete existing columns?',
                      mtConfirmation, [mbYes, mbNo, mbCancel], 0);
     case msgValue of
       mrYes: BS_DBGrid.Columns.Clear;
       mrCancel: Exit;
     end;
   end;
   for i := 0 to BS_DBGrid.DataSource.DataSet.FieldCount - 1 do begin
     Col := BS_DBGrid.Columns.Add;
     Col.FieldName := BS_DBGrid.DataSource.DataSet.Fields[i].FieldName;
   end;
  end;
  UpdateData;
  Designer.Modified;
end;

procedure TBS_DBGridColumnsEditor.tbRestoreDefaultsClick(Sender: TObject);
var i:Integer;
  ListItem: TListItem;
begin
  if (ListView1.SelCount > 0) then begin
    ListItem := ListView1.Selected;
    for i := 0 to ListView1.SelCount - 1 do begin
      TBS_Column(ListItem.Data).RestoreDefaults;
      ListItem := ListView1.GetNextItem(ListItem,sdBelow,[isSelected]);
    end;
    ListView1Change(nil,nil,ctState);
    Designer.Modified;
  end;
end;

procedure TBS_DBGridColumnsEditor.ListView1Change(Sender: TObject;
  Item: TListItem; Change: TItemChange);
var
  FComponents: TDesignerSelectionList;
  i: Integer;
//  ListItem: TListItem;
begin
  if (Change <> ctState)  then Exit;
  tbDeleteSelected.Enabled := ListView1.SelCount > 0;
  tbRestoreDefaults.Enabled := ListView1.Items.Count > 0;

  miDeleteSelected.Enabled := ListView1.SelCount > 0;
  miRestoreDefaults.Enabled := ListView1.Items.Count > 0;
  miSelectAll.Enabled := ListView1.Items.Count > 0;
  tbMoveDown.Enabled := ListView1.Items.Count > 0;
  tbMoveUp.Enabled := ListView1.Items.Count > 0;
  tbAddAllFields.Enabled := Assigned(BS_DBGrid) and Assigned(BS_DBGrid.DataSource) and
    Assigned(BS_DBGrid.Datasource.Dataset) and (BS_DBGrid.Datasource.Dataset.FieldCount > 0);

  if CheckCollection and Active then
  begin
    if (ListView1.SelCount > 0) then
    begin
      FComponents := TDesignerSelectionList.Create;
      try
        for i := ListView1.Items.Count - 1 downto 0 do
          if (ListView1.Items[i].Selected = True) and Assigned(ListView1.Items[i].Data) then
            FComponents.Add(TBS_Column(ListView1.Items[i].Data));
        Designer.SetSelections(FComponents);
      finally
        FComponents.Free;
      end;
    end
    else
      Designer.SelectComponent(BS_DBGrid.Columns);
  end;
end;

procedure TBS_DBGridColumnsEditor.Activated;
begin
  ListView1Change(nil,nil,ctState);
end;

function TBS_DBGridColumnsEditor.UniqueName(Component: TComponent): string;
var
  Temp: string;
begin
  if (Component <> nil) then Temp := Component.ClassName
  else Temp := TBS_Column.ClassName;
  if (UpCase(Temp[1]) = 'T') and (Length(Temp) > 1) then
    System.Delete(Temp, 1, 1);
  Result := Designer.UniqueName(Temp);
end;

procedure TBS_DBGridColumnsEditor.EditAction(Action: TEditAction);
begin
  case Action of
    eaCut: Cut;
    eaCopy: Copy;
    eaPaste: Paste;
    eaDelete: tbDeleteSelectedClick(Self);
  end;
end;

procedure TBS_DBGridColumnsEditor.Copy;
var
  CompList: TDesignerSelectionList;
  ListItem: TListItem;
  i:Integer;
begin
  CompList := TDesignerSelectionList.Create;
  try
    if (ListView1.SelCount > 0) then begin
      ListItem := ListView1.Selected;
      for i := 0 to ListView1.SelCount - 1 do begin
        CompList.Add(TBS_Column(ListItem.Data));
        ListItem := ListView1.GetNextItem(ListItem,sdBelow,[isSelected]);
      end;
    end;
//    CopyComponents(BS_DBGrid.Columns, CompList);
  finally
    CompList.Free;
  end;
end;

procedure TBS_DBGridColumnsEditor.Cut;
begin
  Copy;
  tbDeleteSelectedClick(Self);
end;

procedure TBS_DBGridColumnsEditor.Paste;
var
  CompList: TDesignerSelectionList;
begin
  if CheckCollection then begin
    CompList := TDesignerSelectionList.Create;
    try
//      PasteComponents(BS_DBGrid.Columns, BS_DBGrid.Columns, CompList);
      UpdateData;
    finally
      CompList.Free;
    end;
  end;
end;

procedure TBS_DBGridColumnsEditor.FormModified;
begin
  if not (csDestroying in ComponentState) then UpdateData;
end;

procedure TBS_DBGridColumnsEditor.FormClosed(Form: TCustomForm);
begin
  if Form = Designer.Form then Close;
end;

{$IFDEF CompilerVersion4} {Borland Delphi 4.0 or C++ Builder 4.0}
procedure TBS_DBGridColumnsEditor.ComponentDeleted(Component: IPersistent);
begin
  if ExtractPersistent(Component) = BS_DBGrid then begin
{$ELSE}
procedure TBS_DBGridColumnsEditor.ComponentDeleted(Component: TComponent);
begin
  if Component = BS_DBGrid then begin
{$ENDIF}
    BS_DBGrid := nil;
    Close;
  end;
end;

function TBS_DBGridColumnsEditor.GetEditState: TEditState;
begin
  Result := [];
  if tbDeleteSelected.Enabled then Result := [esCanDelete];
//  if tbDeleteSelected.Enabled then Result := [esCanDelete, esCanCut, esCanCopy];
//  if ClipboardComponents then Include(Result, esCanPaste);
end;

{ TBS_DBGridColumnsProperty }

procedure TBS_DBGridColumnsProperty.Edit;
begin
  ShowItemsEditor(Designer, TBS_DBGrid(GetComponent(0)));
end;

function TBS_DBGridColumnsProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paReadOnly];
end;

function TBS_DBGridColumnsProperty.GetValue: string;
begin
  FmtStr(Result, '(%s)', [GetPropType^.Name]);
end;


{ TBS_DBGridEditor }

procedure TBS_DBGridEditor.ExecuteVerb(Index: Integer);
begin
  case Index of
    0: ShowItemsEditor(Designer, TBS_DBGrid(Component));
  end;
end;

function TBS_DBGridEditor.GetVerb(Index: Integer): string;
begin
  case Index of
    0: Result := 'Columns Editor ...';
  end;
end;

function TBS_DBGridEditor.GetVerbCount: Integer;
begin
  Result := 1;
end;

procedure TBS_DBGridColumnsEditor.FormResize(Sender: TObject);
begin
  ListView1.Columns[0].Width := ListView1.ClientWidth;
end;

procedure TBS_DBGridColumnsEditor.miAddNewClick(Sender: TObject);
begin
  tbAddNewClick(Sender);
end;

procedure TBS_DBGridColumnsEditor.miDeleteSelectedClick(Sender: TObject);
begin
  tbDeleteSelectedClick(Sender);
end;

procedure TBS_DBGridColumnsEditor.miSelectAllClick(Sender: TObject);
var i:Integer;
begin
  for i := 0 to ListView1.Items.Count - 1 do ListView1.Items[i].Selected := True;
end;

procedure TBS_DBGridColumnsEditor.miAddAllFieldsClick(Sender: TObject);
begin
  tbAddAllFieldsClick(Sender);
end;

procedure TBS_DBGridColumnsEditor.miRestoreDefaultsClick(Sender: TObject);
begin
  tbRestoreDefaultsClick(Sender);
end;

procedure TBS_DBGridColumnsEditor.tbMoveUpClick(Sender: TObject);
var i:Integer;
    idxs: TList;
begin
  idxs := TList.Create;
  try
    BS_DBGrid.Columns.BeginUpdate;
    try
      for i := 1 to ListView1.Items.Count - 1 do
        if (ListView1.Items[i].Selected = True) then
          idxs.Add(Pointer(i));
      for i := 1 to ListView1.Items.Count - 1 do
        if (ListView1.Items[i].Selected = True) then
          BS_DBGrid.Columns[i].Index := i-1;
    finally
      BS_DBGrid.Columns.EndUpdate;
    end;

    ListView1.Items.BeginUpdate;
    try
      UpdateData;
      for i := 0 to idxs.Count - 1 do
        ListView1.Items[Integer(idxs.Items[i])-1].Selected := True;
     finally
       ListView1.Items.EndUpdate;
     end;
  finally
    idxs.Free;
  end;
end;

procedure TBS_DBGridColumnsEditor.tbMoveDownClick(Sender: TObject);
var i:Integer;
    idxs: TList;
begin
  idxs := TList.Create;
  try
    BS_DBGrid.Columns.BeginUpdate;
    try
      for i := 0 to ListView1.Items.Count - 2 do
        if (ListView1.Items[i].Selected = True) then
          idxs.Add(Pointer(i));
      for i := ListView1.Items.Count - 2 downto 0 do
        if (ListView1.Items[i].Selected = True) then
          BS_DBGrid.Columns[i].Index := i+1;
    finally
      BS_DBGrid.Columns.EndUpdate;
    end;

    ListView1.Items.BeginUpdate;
    try
      UpdateData;
      for i := 0 to idxs.Count - 1 do
        ListView1.Items[Integer(idxs.Items[i])+1].Selected := True;
     finally
       ListView1.Items.EndUpdate;
     end;
  finally
    idxs.Free;
  end;
end;


{ TBS_DBGridFieldProperty }

function TBS_DBGridFieldProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paSortList, paMultiSelect];
end;

procedure TBS_DBGridFieldProperty.GetValueList(List: TStrings);
var
  Ehg : TBS_CustomDBGrid;
begin
  if (GetComponent(0) = nil) then Exit;
  if (GetComponent(0) is TBS_Column) then
    Ehg := (GetComponent(0) as TBS_Column).Grid
  else if (GetComponent(0) is TBS_ColumnFooter) then
    Ehg := (GetComponent(0) as TBS_ColumnFooter).Column.Grid
  else Exit;

  if (Ehg <> nil) and (TBS_DBGrid(Ehg).DataSource <> nil) and (TBS_DBGrid(Ehg).DataSource.DataSet <> nil) then
       TBS_DBGrid(Ehg).DataSource.DataSet.GetFieldNames(List);
end;

procedure TBS_DBGridFieldProperty.GetValues(Proc: TGetStrProc);
var
  i: Integer;
  Values: TStringList;
begin
  Values := TStringList.Create;
  try
    GetValueList(Values);
    for i := 0 to Values.Count - 1 do Proc(Values[i]);
  finally
    Values.Free;
  end;
end;

end.
