unit BS_db;

interface

uses
  Classes, DB, SysUtils;


const
  DEFAULT_SUBSTITUTE_CHAR = '&';
  SSqlSubstitutionNotFound = 'SqlSubstitution Not Found';

type
  TSqlSubstitutions = class;
  TSqlSubstitution = class(TCollectionItem)
  private
    FData: String;
    FName: string;
  published
    property Name: string read FName write FName;
    property Value: String read FData write FData;
  end;

  TSqlSubstitutions = class(TCollection)
    FOwner: TPersistent;
    function GetItem(Index: Integer): TSqlSubstitution;
    procedure SetItem(Index: Integer; Value: TSqlSubstitution);
  protected
    function GetDataSet: TDataSet;
    function GetOwner: TPersistent; override;
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(Owner: TPersistent); overload;
    destructor  Destroy;override;
    procedure AssignValues(Value: TSqlSubstitutions);
    function Add: TSqlSubstitution;
    function FindSqlSubstitution(const Value: string): TSqlSubstitution;
    function SqlSubstitutionByName(const Value: string): TSqlSubstitution;
    property Items[Index: Integer]: TSqlSubstitution read GetItem write SetItem; default;
  end;

implementation

//// TSqlSubstitutions ////

function TSqlSubstitutions.Add: TSqlSubstitution;
begin
  Result := TSqlSubstitution(inherited Add);
end;

procedure TSqlSubstitutions.AssignValues(Value: TSqlSubstitutions);
var
  i: Integer;
  s: TSqlSubstitution;
begin
  for i := 0 to Value.Count - 1 do
  begin
    s := FindSqlSubstitution(Value[I].Name);
    if s <> nil then
      s.Assign(Value[i]);
  end;

end;

constructor TSqlSubstitutions.Create(Owner: TPersistent);
begin
  FOwner:= Owner;
  inherited Create(TSqlSubstitution);
end;

destructor TSqlSubstitutions.Destroy;
begin
  inherited;
end;

function TSqlSubstitutions.FindSqlSubstitution(
  const Value: string): TSqlSubstitution;
var
  i: Integer;
begin
  for i := 0 to Count - 1 do
  begin
    Result := TSqlSubstitution(inherited Items[i]);
    if AnsiCompareText(Result.Name, Value) = 0 then
      Exit;
  end;
  Result := nil;
end;

function TSqlSubstitutions.GetDataSet: TDataSet;
begin
  if FOwner is TDataSet then
    Result := TDataSet(FOwner) else
    Result := nil;
end;

function TSqlSubstitutions.GetItem(Index: Integer): TSqlSubstitution;
begin
  Result := TSqlSubstitution(inherited GetItem(Index));
end;

function TSqlSubstitutions.GetOwner: TPersistent;
begin
  Result := FOwner;
end;


procedure TSqlSubstitutions.SetItem(Index: Integer; Value: TSqlSubstitution);
begin
  inherited SetItem(Index, Value);
end;

function TSqlSubstitutions.SqlSubstitutionByName(
  const Value: string): TSqlSubstitution;
begin
  Result := FindSqlSubstitution(Value);
  if Result = nil then
     DatabaseErrorFmt('%s '+SSqlSubstitutionNotFound, [Value], GetDataSet);
end;

procedure TSqlSubstitutions.Update(Item: TCollectionItem);
begin
  inherited;
end;


end.
