unit BS_dbf;
//Autor: Pascal GANAYE <p.ganaye@bigfoot.com>
//Modified by Siarhej Berdachuk

interface
//------------------------------------------------------------------------------
{$ifdef VER100}      // Delphi 3
   {$define NO_LARGE_INT}
{$endif}

{$ifdef VER110}      // CBuilder 3
   {$define NO_LARGE_INT}
{$endif}

uses
  Classes, Db, Windows, Forms, Controls, SysUtils, Dialogs;

type
  EBinaryDataSetError = class (Exception);
  EFieldToLongError = class (Exception);
  xBaseVersion = (xBaseIII,xBaseIV,xBaseV);

  TPagedFile = class(TObject)
  protected
    FDbf:TDataSet;
    FStreamFile : TFileStream;
    FFieldInfoList: TList;
    FHeaderSize : Integer;
    FRecordSize : Integer;
  public
    constructor Create(Parent:TDataset; const FileName: string; Mode: Word);
    destructor Destroy; override;
    property RecordSize : Integer read FRecordSize;
    function CalcRecordCount:Integer;
    procedure Seek(page:Integer);
    procedure ReadRecord(IntRecNum:Integer;Buffer:Pointer);
    procedure WriteRecord(IntRecNum:Integer;Buffer:Pointer);
  end;

  rDbfHdr = record
    VerDBF      : byte;   // 0
    Year        : byte;   // 1
    Month       : byte;   // 2
    Day         : byte;   // 3
    RecordCount : Integer;  // 4-7
    FullHdrSize : word;   // 8-9
    RecordSize  : word;   // 10-11
    Dummy1      : Word;   // 12-13
    IncTrans    : byte;   // 14
    Encrypt     : byte;   // 15
    Dummy2      : Integer; // 16-19
    Dummy3      : array[20..27] of byte; // 20-27
    MDXFlag     : char; // 28
    Language    : char; // 29
    dummy4      : word; // 30-31
  end;

  TdbfFieldInfo = class
  public
    Size:Integer;
    Prec:Integer;
    Offset:Integer;
    FieldName: string;
    charFieldType: Char;
    destructor destroy;  override;
  end;


  TBS_DbfFile = class(TPagedFile)
  protected
    FDataHeader : rDbfHdr;
    FVersion : xBaseVersion;
  public
    property Version : xBaseVersion read FVersion;
    property DataHeader : rDbfHdr read FDataHeader;

    constructor Create(Parent:TDataset; const FileName: string; Mode: Word);
    destructor Destroy; override;
    function  RecordCount:Integer;
    procedure WriteHeader;
    procedure AppendRecord(Buffer:Pointer);
    procedure CreateFieldDefs(FieldDefs:TFieldDefs);
    procedure CreateTable(FieldDefs:TFieldDefs);
    procedure CleardbfFieldInfos;
    function  GetFieldData(Column:Integer;DataType:TFieldType; Src,Dst: Pointer): Boolean;
    procedure SetFieldData(Column:integer;DataType:TFieldType; Src,Dst: Pointer);
    function  GetFieldInfo(AFieldName:String; var dbfFieldInfo:TdbfFieldInfo): Boolean;

    function  GetOraLoaderStr(const aFieldName:string; const aOraFieldName:String; aUsePosition:Boolean): string;overload;
    function  GetOraLoaderStr(const aFieldName:string; const aOraFieldName:String): string;overload;
    //Creation data file, to aTargetFile withour header
    function  CreateHeaderExtactedDataFile(aTargetFileName:String):boolean;
  end;

  rDbtHdr = record
    NextBlock:Longint;
    Dummy : array [4..7] of byte;
    DBFFile : array [0..7] of Byte; //8..15
    bVer : Byte; //16
    Dummy2 : array [17..19] of byte;
    BlockLen:  Word;
  end;

  TDbtFile = class(TPagedFile)
  protected
    FVersion:xBaseVersion;
    _MemoHdr:rDbtHdr;
  public
    constructor Create(Parent:TDataSet; const FileName: string; Mode: Word; Ver:xBaseVersion);
    procedure ReadMemo(recno:Integer;Dst:TStream);
    procedure WriteMemo(var MemoRecno:Integer;ReadSize:Integer;Src:TStream);
  end;

  rMdxHdr = record
    MdxHdr   : byte;       // 0
    Year        : byte;    // 1
    Month       : byte;    // 2
    Day         : byte;    // 3
    FileName    : array[0..15] of char; // 4..19 of byte
    BlockSize    : word;   // 20 21
    BlockAdder  : word;    // 22 23
    IndexFlag   : byte;    // 24
    NoTag       : byte;    // 25
    TagSize     : byte;    // 26
    Dummy1      : byte;    // 27
    TagUsed     : word;    // 28..29
    Dummy2      : word;    // 30..31
    NbPage      : Integer; // 32..35
    FreePage    : Integer; // 36..39
    BlockFree   : Integer; // 40..43
    UpdYear     : byte;    // 44
    UpdMonth    : byte;    // 45
    UpdDay      : byte;    // 46
  end;

  TIndexFile = class(TPagedFile)
  protected
    FVersion : xBaseVersion;
    _MdxHdr : rMdxHdr;
    FStreamFileName : string;
  public
    constructor Create(Parent:TDataset; const FileName: string; Mode: Word);
    destructor Destroy; override;
  end;

  rIndexHdr = record
    startpage : Integer; // 0..3
    nbPage : Integer; // 4..7
    keyformat: Char; //8
    keytype : char; //9
    dummy : Word; // 10..11
    keylen : Word; // 12..13
    nbkey : Word; // 14..15
    skeytype : Word; // 16..17
    keyreclen : Word; // 18..19
    dummy2 : Word; // 20..21
    dummy3 : Byte; // 22
    Unique : Byte; // 23
    KeyDesc : array[0..255] of char; // 24...
  end;

  rMdxTag = record
    pageno      : Integer; // 0..3
    tagname      : array [0..11] of char; // 4..14
    keyformat    : byte; // 15
    forwardTag1  : char; // 16
    forwardTag2 : byte; // 17
    backwardTag : byte; // 18
    dummy       : byte; // 19
    keytype     : byte; // 20
  end;

  NdxKeyType = (N,C);
  PNdxPage  = ^rNdxPage;
  rNdxPage  = record
    NbEntries : longint;  //  0..3 lower page
    Entries   : ARRAY [0..507] OF char;
  end;

  PNdxentry  = ^rNdxentry;
  rNdxentry  = record
    _LowerPage : longint;  //  0..3 lower page
    RecNo     : Longint;  //  4..7 recno
    case NdxKeyType of
      N: ( NKey: double);
      C: ( CKey: array [0..503] of char);
  end;

  PRecInfo = ^TRecInfo;
  TRecInfo = record
    Bookmark: Longint;
    IdxBookmark: Longint;
    BookmarkFlag: TBookmarkFlag;
  end;

  TIndex = class;

  PIndexPosInfo = ^TIndexPage;

  TIndexPage = class
  protected
    _Index : TIndex;
    _PageNo : Integer;
    _PageBuff:rNdxPage;
    _EntryNo : Integer;
    Entry : PNdxentry;
    _LowerLevel : TIndexPage;
    _UpperLevel : TIndexPage;

    procedure LocalFirst;
    procedure LocalLast;
    function  LocalPrev:boolean;
    function  LocalNext:boolean;
    function  LastEntryNo:integer;
    function  LocalInsert(const Key;Recno,LowerPage:integer):boolean;
    function  LocalDelete:boolean;

    function  GetPEntry(EntryNo:integer):PNdxEntry;
    procedure First;
    procedure Last;
    function Prev:boolean;
    function Next:boolean;
    procedure Write;
  public
    constructor Create(Parent:TIndex);
    destructor Destroy; override;

    procedure SetPageNo(page:Integer);
    procedure SetEntryNo(entryno:Integer);
    procedure WritePage(Page:integer);
    function FindNearest(const Key;Recno:integer):integer;
    function Insert(const key;Recno,LowerPage:integer):boolean;
    procedure SetEntry(const key;Recno,LowerPage:integer);
    function Delete:boolean;
    function LowerLevel : TIndexPage;
  end;



  TIndex = class(TObject)
  protected
    _IndexFile:TIndexFile;
    _IndexHdr:rIndexHdr;
    _Root:TIndexPage;
    _Spare:TIndexPage; // used during insertion and deletion
    _TagPosition:Integer;
    _Field : TField;
    _NbLevel : integer;
    function Pos:TIndexPage;
  public
    IndexRecNo:integer;
    function Prev:boolean;
    function Next:boolean;
    procedure First;
    procedure Last;
    function Find(const Key;Recno:integer;var pPos:TIndexPage):integer;
    function Insert:boolean;
    function Delete:boolean;

    procedure GotoBookMark(recinfo:TRecInfo);
    procedure Resync;
    function GetRecordCount: Integer;
    function GetRecNo: Integer;
    function GetRealRecNo: Integer;
    constructor Create(Parent:TIndexFile;Pos:integer);
    destructor Destroy; override;
  end;


  pRecordHdr = ^tRecordHdr;
  tRecordHdr = record
    DeletedFlag : char;
  end;




 ///////////////////////////////////////////////////////////////////////////
  TBS_Dbf = class(TDataSet)
  private
    FShowDeleted:boolean;
    Procedure _AppendRecord(Buffer:PChar);
    procedure _OpenFiles(Create:boolean);
    procedure _CloseFiles;
    function  _FilterRecord(Buffer: PChar): Boolean;
  protected
    FTableName: string;          // table path and file name
    FDbfFile: TBS_DbfFile;       // dbf file
    FdbtFile: TDbtFile;          // dbt file
    FIndexFilesList: TList;      // ndx and mdx files
    FIndexesList:TList;          // index
    FCurIndex: TIndex;
    FRecordBufferSize : integer;  // data + housekeeping (TRecInfo)
    FRecordInfoOffset : integer;  // offset of RecInfo in record buffer
    FCurtRecNo : integer;         // current record (0 to FDataHeader.RecCount - 1)
    FDelSgn : integer;            // 1 if current record deleted
    FIsTableOpen: Boolean;        // status
    FReadOnly : Boolean;
    FFilterBuffer:pchar;
    FHeaderSize : Integer;
    FRecordSize : Integer;

    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean):
                        TGetResult; override;
    function GetRecordSize: Word; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalPost; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    function GetRecordCount: Integer; override;
    function GetRecNo: Integer; override;
    function GetCanModify: Boolean; override;
    procedure CloseBlob(Field: TField); override;
  public
    MemIndexList:TList;
    easyfilter:string;
    property DbfFile: TBS_DbfFile read FDbfFile;
    property DbtFile:TDbtFile read  FdbtFile;
    property IndexFilesList: TList read FIndexFilesList;

    function GetFieldInfo(AFieldName:String; var dbfFieldInfo:TdbfFieldInfo): Boolean;
    function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    procedure Find(const Key);
    procedure UseIndex(IndexNo:integer);
    function OpenIndex(IndexName:string):Integer;
    procedure SetShowDeleted(v:boolean);
    constructor Create(AOwner:tComponent); override;
    destructor Destroy; override;
    procedure CreateTable;
    //Creation data file, to aTargetFile withour header
    function  CreateHeaderExtactedDataFile(aTargetFileName:String):boolean;
    function  GetLoaderDataStr(const aFieldsNames:string; aEncloseChar,aReplaceEncloseChar,aFieldsDelimiter:Char): string;
    function DelSgn: integer;
  published
    property ReadOnly : Boolean read fReadOnly write fReadonly default True;

    property Active;    // redeclared data set properties
    property AutoCalcFields;

    property Constraints;
    property FieldDefs;
    property Filter;
    property Filtered;


    property Name;
    property ObjectView;
    property TableName: string  read FTableName write FTableName;


    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
    property ShowDeleted:boolean read FShowDeleted write SetShowDeleted;
  end;




implementation

Const
  dfhVersionNumber = 14;
  BofCrack = -1; // before the first record (crack)

type

  PBufArray = PCHAR;

  rAfterHdrIII = record // Empty
  end;

  rAfterHdrV = record
    Dummy   : array[32..67] of byte;
  end;

  rFieldHdrIII = record
    FieldName   : array[0..10] of char;
    FieldType   : char; // 11
    Dummy        : array[12..15] of byte;
    FieldSize   : byte; // 16
    FieldPrecision  : byte; //17
    dummy2      : array[18..31] of byte;
  end;

  rFieldHdrV = record
    FieldName   : array[0..10] of char;
    Dummy0        : array[11..31] of byte;
    FieldType   : char; // 32
    FieldSize   : byte; // 33
    FieldPrecision  : byte; //34
    dummy2      : array[35..47] of byte;
  end;

  PMdxTag = ^rMdxTag;

  rMdxTagHdr = record
    RootPage        : longint;// 0..3
    FilePages        : longint;// 4..7
    KeyFormat        : byte;   // 8
    KeyType          : char;   // 9
    dummy            : word;   // 10..11
    IndexKeyLength   : word;   // 12..13
    MaxNbKeys       : word;   // 14..15
    SecondKeyType   : word;   // 16..17
    IndexKeyItemLen  : word;   // 18..19
    dummy2           : array [20..22] of byte;
    UniqueFlag      : byte;   // 23
  end;


  rAfterHdrV3 = record
    Dummy   : array[12..31] of byte;
  end;

  rAfterHdrV4 = record
    Dummy   : array[12..67] of byte;
  end;

  rFieldHdrV3 = record
    FieldName   : array[0..10] of char;
    FieldType   : char; // 11
    Dummy        : array[12..15] of byte;
    FieldSize   : byte; // 16
    FieldPrecision  : byte; //17
    dummy2      : array[18..31] of byte;
  end;

  rFieldHdrV4 = record
    FieldName   : array[0..10] of char;
    Dummy0        : array[11..31] of byte;
    FieldType   : char; // 32
    FieldSize   : byte; // 33
    FieldPrecision  : byte; //34
    dummy2      : array[35..47] of byte;
  end;

  TMyBlobFile = class(TMemoryStream)
  public
    Mode: TBlobStreamMode;
    Dbf: TBS_Dbf;
    Field:TField;
    MemoRecno:Integer;
    ReadSize:Integer;
    constructor Create(ModeVal:TBlobStreamMode;DbfVal: TBS_Dbf; FieldVal:TField);
    destructor destroy;  override;
  end;

procedure MyMove(Source, Dest:pointer; Count: Integer);
var
  c:char;
  i:integer;
begin
  i:=0;
  while i<Count do begin
    c:=PChar(Source)[i];
    if c=#0 then break;
    PChar(Dest)[i]:=c;
    Inc(i);
  end;
  while i<Count do begin
    PChar(Dest)[i]:=' ';
    Inc(i);
  end;
end;

// ****************************************************************************
// International separator
// thanks to Bruno Depero from Italy
function DbfStrToFloat(s:string):Extended;
var
  oDECIMALSEPARATOR:char;
  oTHOUSANDSEPARATOR:Char;
  f:Extended;
begin
  oDECIMALSEPARATOR:=DECIMALSEPARATOR;
  oTHOUSANDSEPARATOR:=THOUSANDSEPARATOR;
  try
    DECIMALSEPARATOR:='.';
    THOUSANDSEPARATOR:=#0;
    if TextToFloat(PChar(s),f,fvExtended) then Result:=f else Result:=0;
  finally
    DECIMALSEPARATOR:=oDECIMALSEPARATOR;
    THOUSANDSEPARATOR:=oTHOUSANDSEPARATOR;
  end;
end;

function FloatToDbfStr(f:Extended;size,prec:integer):string;
var
  oDECIMALSEPARATOR:char;
  oTHOUSANDSEPARATOR:Char;
  s:string;
begin
  oDECIMALSEPARATOR:=DECIMALSEPARATOR;
  oTHOUSANDSEPARATOR:=THOUSANDSEPARATOR;
  try
    DECIMALSEPARATOR:='.';
    THOUSANDSEPARATOR:=#0;
    s:=FloatToStrF(f,ffFixed,Size,prec);
    Result:=s;
  finally
    DECIMALSEPARATOR:=oDECIMALSEPARATOR;
    THOUSANDSEPARATOR:=oTHOUSANDSEPARATOR;
  end;
end;




// ****************************************************************************
// Paged File
// Implement any Header + Fixed size records File.
// I hope it will be used for all files...

function TPagedFile.CalcRecordCount:Integer;
begin
  if FRecordSize = 0 then Result:=0
  else Result:=(FStreamFile.Size - FHeaderSize) div FRecordSize;
end;

constructor TPagedFile.Create(Parent:TDataSet; const FileName: string; Mode: Word);
begin
  FDbf:=Parent;
  try
    FStreamFile:=TFileStream.Create(FileName,Mode);
  except
    FStreamFile:=nil;
  end;
end;

destructor TPagedFile.Destroy;
begin
  FStreamFile.Free;
  inherited;
end;

procedure TPagedFile.Seek(page:Integer);
var
  p:Integer;
begin
  p:=FHeaderSize + (FRecordSize * page );
  FStreamFile.Position := p;
end;

Procedure TPagedFile.ReadRecord(IntRecNum:Integer; Buffer:Pointer);
begin
  Seek(IntRecNum);
  FStreamFile.Read(Buffer^,FRecordSize);
end;

procedure TPagedFile.WriteRecord(IntRecNum:Integer; Buffer:Pointer);
begin
  Seek(IntRecNum);
  FStreamFile.Write(Buffer^, FRecordSize);
end;

(*
procedure TPagedFile.ReadHeader(var Buffer);
begin
  FStreamFile.Seek(0,soFromBeginning);
  FStreamFile.ReadBuffer(Buffer,_HeaderSize);
end;
*)

(*
procedure TPagedFile.WriteHeader(const Buffer);
begin
  FStreamFile.Seek(0,soFromBeginning);
  FStreamFile.WriteBuffer (Buffer, FHeaderSize);
end;
*)







constructor TMyBlobFile.Create(ModeVal:TBlobStreamMode;DbfVal: TBS_Dbf;FieldVal:TField);
begin
  Mode:=ModeVal;
  Dbf:=DbfVal;
	Field:=FieldVal;
end;

destructor TMyBlobFile.destroy;
begin
  if (Mode=bmWrite) then begin
    Size:=Position; // Strange but it leave tailing trash bytes if I do not write that.
    Dbf.FdbtFile.WriteMemo(MemoRecno,ReadSize,Self);
		Dbf.FdbfFile.SetFieldData(Field.FieldNo-1,ftInteger,@MemoRecno,Dbf.ActiveBuffer);
  end;
  inherited;
end;


destructor TdbfFieldInfo.destroy;
begin
  inherited;
end;


constructor TDbtFile.Create(Parent:TDataSet; const FileName: string; Mode: Word; Ver:xBaseVersion);
begin
  inherited Create(Parent,FileName,Mode);
  FVersion:=Ver;
  if fmCreate = Mode then begin
    FillChar(_MemoHdr,sizeof(_MemoHdr),0);
  end else begin
    FStreamFile.Position:=0;
    FStreamFile.read(_MemoHdr,SizeOf(_MemoHdr));
  end;
  FHeaderSize:=0;
  FRecordSize:=_MemoHdr.BlockLen;

  if (FRecordSize=0) or ((FRecordSize mod 128)<>0) then begin
    _MemoHdr.BlockLen := $200;
    FRecordSize := $200;
  end;
end;

procedure TDbtFile.ReadMemo(recno:Integer;Dst:TStream);
var
  Buff:array[0..4095] of char;
  i,lsize:integer;
  finish:boolean;
  lastc:char;
begin
  if recno=0 then Exit;
  FStreamFile.Position:= FRecordSize * recno;
  if FVersion >= xBaseIV then begin // dBase4 memofiles
    FStreamFile.read(Buff[0],8);
    if (Buff[0]=#$ff) and  (Buff[1]=#$ff) and
      (Buff[2]=#$08) and (Buff[3]=#$00) then begin
          // dbase IV memo
      lsize:=(PInteger(@Buff[4])^)-8;
    end else begin
      lsize:=0;
    end;
    repeat
      if lsize>SizeOf(Buff) then begin
        FStreamFile.read(Buff,SizeOf(Buff));
        Dst.Write(buff,SizeOf(Buff));
        Dec(lsize,SizeOf(Buff));
      end else if lsize>0 then begin
        FStreamFile.read(Buff,lsize);
        Dst.Write(buff,lsize);
        lsize:=0;
      end;
    until lsize=0;
  end else begin
    finish:=False;
    FStreamFile.read(Buff,SizeOf(Buff));
    lastc:=#0;
    repeat
      for i:=0 to SizeOf(Buff)-2 do begin
        if ((Buff[i]=#$1A) and
          ((Buff[i+1]=#$1A) or ((i=0) and (lastc=#$1A))))
          or (Buff[i]=#$0)
          then begin
          if i>0 then Dst.Write(buff,i);
          finish:=True;
          break;
        end;
      end;
      if finish then Break;
      Dst.Write(buff,512);
      lastc:=Buff[511];
      FStreamFile.read(Buff,SizeOf(Buff));
    until finish;
  end;
  Dst.Seek(0,0);
end;

procedure TDbtFile.WriteMemo(var MemoRecno:Integer;ReadSize:Integer;Src:TStream);
var
  ByteBefore:Integer;
  ByteAfter:Integer;
  Buff:array[0..511] of char;
  i:Integer;
  c:Byte;
  Append:Boolean;
begin
  if FVersion >= xBaseIV then begin // dBase4 memofiles
    ByteBefore:=8;
    ByteAfter:=0;
  end else begin // stupid files
    ByteBefore:=0;
    ByteAfter:=2;
  end;
  if Src.Size = 0 then begin
    MemoRecno:=0;
  end else begin
    if ((ByteBefore+Src.Size+ByteAfter+_MemoHdr.BlockLen-1) div _MemoHdr.BlockLen)
      <= ((ReadSize+_MemoHdr.BlockLen-1) div _MemoHdr.BlockLen)
      then begin
      Append:=false;
      //MemoRecno:=MemoRecno;
    end else begin
      Append:=True;
      MemoRecno:=_MemoHdr.NextBlock;
    end;
    FStreamFile.Seek(_MemoHdr.BlockLen * MemoRecno,0);
    i:=Src.Position;
    Src.Seek(0,0);
    if ByteBefore=8 then begin
      i:=$0008ffff;
      FStreamFile.Write(i,4);
      i:=Src.Size+ByteBefore+ByteAfter;
      FStreamFile.Write(i,4);
    end;
    repeat
      i:=Src.Read(buff,512);
      if i=0 then break;
      Inc(_MemoHdr.NextBlock);
      FStreamFile.Write(Buff,i);
    until i<512;
    if ByteAfter=2 then begin
      c:=$1A;
      FStreamFile.Write(c,1);
      FStreamFile.Write(c,1);
    end;
    if Append then begin
      FStreamFile.Seek(0,0);
      FStreamFile.Write(_MemoHdr,SizeOf(_MemoHdr))
    end;
  end;
end;




constructor TIndexPage.Create(Parent:TIndex);
begin
  _LowerLevel:=nil;
  _UpperLevel:=nil;
  _Index:=Parent;
  _PageNo:=-1;
  _EntryNo:=-1;
end;

destructor TIndexPage.Destroy;
begin
  if _LowerLevel<>nil then _LowerLevel.Free;
end;

function  TIndexPage.GetPEntry(EntryNo:integer):PNdxEntry;
begin
  Result:=PNdxentry(@_PageBuff.Entries[_Index._IndexHdr.keyreclen*entryno]);
end;

function  TIndexPage.LocalInsert(const Key;Recno,LowerPage:integer):boolean;
var
  src,dst:pointer;
  siz:integer;
begin
  if _PageBuff.NbEntries < _Index._IndexHdr.nbkey then begin
    src:=Entry;
    dst:=GetPEntry(_EntryNo+1);
    siz:=(_PageBuff.NbEntries - _EntryNo)
      * _Index._IndexHdr.keyreclen + 8;
    Move(Src^, Dst^, Siz);
    inc(_PageBuff.NbEntries);
    SetEntry(Key,Recno,LowerPage);
    Write;
    Result:=true;
  end else begin
    Result:=false;
  end;
end;


function  TIndexPage.LocalDelete:boolean;
var
  src,dst:pointer;
  siz:integer;
begin
  if _PageBuff.NbEntries > 0 then begin
    src:=GetPEntry(_EntryNo+1);
    dst:=Entry;
    siz:=(_PageBuff.NbEntries - _EntryNo - 1)
      * _Index._IndexHdr.keyreclen + 8;
    Move(Src^, Dst^, Siz);
    dec(_PageBuff.NbEntries);
    if (_EntryNo > LastEntryNo) then begin
      SetEntryNo(_EntryNo-1); // We just removed the last on this page.
      SetEntry(Entry.CKey,Entry.Recno,Entry._LowerPage);
    end;
    Write;
    Result:=true;
  end else begin
    Result:=false;
  end;
end;

function  TIndexPage.LastEntryNo:integer;
begin
  if (_LowerLevel=nil) then begin
    result := _PageBuff.NbEntries - 1;
  end else begin
    result := _PageBuff.NbEntries;
  end;
end;

procedure TIndexPage.LocalFirst;
begin
  SetEntryNo(0);
end;

procedure TIndexPage.LocalLast;
begin
  SetEntryNo(LastEntryNo);
end;

function TIndexPage.LocalPrev:boolean;
begin
  if _EntryNo>0 then begin
    SetEntryNo(_EntryNo-1);
    Result:=true;
  end else begin
    Result:=false;
  end;
end;

function TIndexPage.LocalNext:boolean;
begin
  if (_EntryNo<LastEntryNo) then begin
    SetEntryNo(_EntryNo+1);
    Result:=true;
  end else begin
    Result:=false;
  end;
end;

procedure TIndexPage.First;
begin
  LocalFirst;
  if (_LowerLevel<>nil) then LowerLevel.First;
end;

procedure TIndexPage.Last;
begin
  LocalLast;
  if (_LowerLevel<>nil) then LowerLevel.Last;
end;

function TIndexPage.Prev:boolean;
begin
  if (_LowerLevel<>nil) and LowerLevel.Prev then begin
    result:=true;
    exit;
  end;
  Result:=LocalPrev;
  if Result and (Entry._LowerPage>0) then LowerLevel.Last;
end;

function TIndexPage.Next:boolean;
begin
  if (_LowerLevel<>nil) and LowerLevel.next then begin
    result:=true;
    exit;
  end;
  Result:=LocalNext;
  if Result and (Entry._LowerPage>0) then LowerLevel.First;
end;


function TIndexPage.FindNearest(const Key;Recno:integer):integer;
var
  cmpres:integer;
  v1,v2:double;
begin
  Result:=-1;
  if @Key=nil then begin
    Exit;
  end;
  SetEntryNo(0);
  while _EntryNo<=_PageBuff.NbEntries do begin
    if _EntryNo=_PageBuff.NbEntries then break;
    if _Index._IndexHdr.keytype='C' then begin
      cmpres:=StrIComp(PChar(Key),Entry.CKey);
    end else begin
      // Numeric field... to do
      v1:=double(Key);
      v2:=Entry.NKey;
      if v1>v2 then cmpres:=1
      else if v1<v2 then cmpres:=-1
      else cmpres:=0;
    end;
    if cmpres=0 then begin
      if (Entry.RecNo=Recno) then begin
        result:=0;
        Exit;
      end else if (Entry.Recno>Recno) then begin
        result:=-1;
        Exit;
      end;
    end else if cmpres<0 then begin
      result:=-1;
      exit;
    end;
    SetEntryNo(_EntryNo+1);
  end;
  result:=1;
  Exit;
end;

procedure TIndexPage.SetEntry(const key;Recno,LowerPage:integer);
begin
  if (_EntryNo>=0) and (_EntryNo<=LastEntryNo) then begin
    if (_EntryNo=LastEntryNo) and (_UpperLevel<>nil)  then begin
      _UpperLevel.SetEntry(key,0,Self._PageNo);
    end;
    if _Index._IndexHdr.keytype='C' then begin
      mymove(@key,@Entry.CKey,_Index._IndexHdr.keylen);
    end else begin
      Entry.NKey:=double(key);
    end;
    Entry.RecNo:=RecNo;
    Entry._LowerPage:=LowerPage;
    Write;
  end;
end;
function TIndexPage.LowerLevel : TIndexPage;
begin
  if (_LowerLevel<>nil) and (_LowerLevel._PageNo<>Entry._LowerPage) then begin
    _LowerLevel.SetPageNo(Entry._LowerPage);
  end;
  result:=_LowerLevel;
end;

function TIndexPage.Insert(const key;Recno,LowerPage:integer):boolean;
var
  src,dst:PNdxentry;
  siz:integer;
  split,old_entry:integer;
begin
  if not LocalInsert(key,recno,lowerpage) then begin
    // The entry is FULL so we will split this page
    // 1 - Check parent exist
    if _UpperLevel=nil then begin
      _Index._Spare._PageNo:= _Index._IndexHdr.nbPage;
      inc(_Index._IndexHdr.nbPage);
      _Index._Spare._PageBuff.NbEntries:=1;
      _Index._Spare._UpperLevel:=nil;
      _Index._Spare._LowerLevel:=_Index._Root;
      _Index._Spare.SetEntryNo(0);
      _Index._Spare.SetEntry(Entry.CKey,0,_PageNo);
      _Index._Root._UpperLevel:=_Index._Spare;
      //_Index._Root._LowerLevel:=unchanged;
      _Index._Root:=_Index._Spare;
    end;
    old_entry:=_EntryNo;

    // 2 - Create new page with first part
    split:=_EntryNo;
    if split < _Index._IndexHdr.nbkey div 2 then begin
      split:=_Index._IndexHdr.nbkey div 2;
    end;
    _Index._Spare._PageNo:=_Index._IndexHdr.nbPage;
    inc(_Index._IndexHdr.nbPage);
    _Index._Spare._PageBuff.NbEntries:=split;
    siz:=split*_Index._IndexHdr.keyreclen+8;
    src:=@_PageBuff.Entries;
    dst:=@_Index._Spare._PageBuff.Entries;
    Move(src^,dst^,siz);
    _Index._Spare.Write;

    // 3 - Keep only end-part in this page
    siz:=(_PageBuff.NbEntries-Split);
    _PageBuff.NbEntries:=siz;

    siz:=siz*_Index._IndexHdr.keyreclen+8;
    SetEntryNo(split);
    src:=Entry;
    SetEntryNo(0);
    dst:=Entry;
    Move(src^,dst^,siz);

    if old_entry>=split then begin
      SetEntryNo(old_entry - split);
      LocalInsert(Key,Recno,LowerPage);
      _Index._Spare.Write;
    end else begin
      _Index._Spare.SetEntryNo(old_entry);
      _Index._Spare.LocalInsert(Key,Recno,LowerPage);
      Write;
    end;

    // 3 - Update upper level
    Last;
    _UpperLevel.SetEntry(Entry.Ckey,0,_PageNo);
    _Index._Spare.Last;
    _UpperLevel.Insert(_Index._Spare.Entry.CKey,0,_Index._Spare._PageNo);

    // We just need to go on inserted record now
    if old_entry>=split then begin
      _UpperLevel.LocalNext;
    end;
  end;
  Result:=true;
end;

function TIndexPage.Delete:boolean;
begin
  Result:=LocalDelete;
end;

procedure TIndexPage.SetPageNo(page:Integer);
begin
  if (_PageNo<>page) and (page>0) then begin
    _Index._IndexFile.ReadRecord(Page,@_PageBuff);
    _PageNo:=page;
    _EntryNo:=-1;
  end;
end;

procedure TIndexPage.Write;
begin
  _Index._IndexFile.WriteRecord(_PageNo,@_PageBuff);
end;

procedure TIndexPage.SetEntryNo(entryno:Integer);
begin
  if (_EntryNo<>entryno) then begin
    _EntryNo:=entryno;
    Entry:=PNdxentry(@_PageBuff.Entries[_Index._IndexHdr.keyreclen*entryno]);
  end;
end;

procedure TIndexPage.WritePage(Page:integer);
begin
  _Index._IndexFile.WriteRecord(Page,@_PageBuff);
end;


constructor TIndexFile.Create(Parent:TDataset; const FileName: string; Mode: Word);
var
  ext:string;
  i:Integer;
  lIndex:TIndex;
//  lMdxTag:rMdxTag;
begin
  inherited Create(Parent,FileName,Mode);
  FStreamFileName:=FileName;
  FHeaderSize:=0;
  FRecordSize:=512;

  ext:=ExtractFileExt(FileName);
  if (ext='.mdx') then begin
    FVersion:=xBaseIV;
    FStreamFile.read(_MdxHdr,SizeOf(_MdxHdr));
    for i:= 0 to _MdxHdr.TagUsed-1 do begin
//      FStreamFile.Position :=544 + i * _MdxHdr.TagSize;
//      FStreamFile.read(lMdxTag,SizeOf(rMdxTag));
//      lIndex:=TIndex.Create(Self,lMdxTag.pageno);
//      FIndexesList.Add(lIndex);
//      if i=0 then lIndex.ReadPage(lIndex._IndexHdr.startpage);
    end;
  end else begin
    FVersion:=xBaseIII;
    lIndex:=TIndex.Create(Self,0);
    TBS_Dbf(FDbf).FIndexesList.Add(lIndex);
    TBS_Dbf(FDbf).FIndexFilesList.Add(Self);
    //Add(lIndex);
  end;
end;

destructor TIndexFile.Destroy;
begin
  inherited;
end;

constructor TIndex.Create(Parent:TIndexFile;Pos:integer);
var
  lPos:TIndexPage;
  lChild:TIndexPage;
begin
  _IndexFile:=Parent;
  //_IndexOrder:=TList.Create;
  _IndexFile.Seek(Pos);
  _IndexFile.FStreamFile.Read(_IndexHdr,SizeOf(_IndexHdr));

  _Root:=TIndexPage.Create(Self);
  _Root.SetPageNo(_IndexHdr.startpage);
  lPos:=_Root;
  _nblevel:=1;
  repeat
    lPos.LocalFirst;
    if lPos.Entry._LowerPage=0 then break;
    inc(_nblevel);
    lChild:=TIndexPage.Create(Self);
    lChild._UpperLevel:=lPos;
    lPos._LowerLevel:=lChild;
    lChild.SetPageNo(lPos.Entry._LowerPage);
    lPos:=lChild;
  until false;

  _Spare:=TIndexPage.Create(Self);
  _Field:=_IndexFile.FDbf.FindField(_IndexHdr.KeyDesc);
  First;
end;

destructor TIndex.Destroy;
begin
  _Root.Free;
  _Spare.Free;
end;


function TIndex.Find(const Key;Recno:integer;var pPos:TIndexPage):integer;
var
  res:integer;
begin
  pPos:=_Root;
  repeat
    res:=pPos.FindNearest(Key,Recno);
    if res<>0 then begin
      if pPos.Entry._LowerPage<>0 then begin
        pPos:=pPos.LowerLevel;
        res:=2;
      end;
    end;
  until res<>2;
//  _Pos:=pPos.Entry;
  Result:=res;
end;

function TIndex.Insert:boolean;
var
  lPos:TIndexPage;
  ckey:array[0..100] of char;
  s:string;
  nkey:double;
  lCurrentRecord:integer;
begin
  lCurrentRecord := TBS_Dbf(_IndexFile.FDbf).FCurtRecNo;
  if _IndexHdr.keytype='C' then begin
    TBS_Dbf(_IndexFile.FDbf).GetFieldData(_Field,@ckey);

    s:=cKey;
    Find(s,lCurrentRecord+1,lPos);
    lPos.Insert(ckey,lCurrentRecord+1,0);
  end else begin
    // Numeric field...
    TBS_Dbf(_IndexFile.FDbf).GetFieldData(_Field,@nkey);
    Find(nKey,lCurrentRecord+1,lPos);
    lPos.Insert(nkey,lCurrentRecord+1,0);
  end;
  Result:=true;
end;

function TIndex.Delete:boolean;
var
  lPos:TIndexPage;
begin
  lpos:=_root;
  while lpos._LowerLevel<>nil do begin
    lPos:=lPos.LowerLevel;
  end;
  lPos.Delete;
  Result:=true;
end;


function TIndex.Pos:TIndexPage;
var
  p:TIndexPage;
begin
  p:=_Root;
  while p.Entry._LowerPage>0 do begin
    p:=p.LowerLevel;
  end;
  result:=p;
end;

procedure TIndex.First;
begin
  _Root.First;
end;

procedure TIndex.Last;
begin
  _Root.Last;
end;

function TIndex.Prev:boolean;
begin
  result:=_Root.Prev;
end;

function TIndex.Next:boolean;
begin
  result:=_Root.Next;
end;

function TIndex.GetRecordCount: Integer;
var
  lPos:TIndexPage;
  nbrecord:integer;
begin
  // I just read first level and Guess an approximate record count...
  nbrecord:=_Root._PageBuff.NbEntries;
  lPos:=_Root.LowerLevel;
  while lpos<>nil do begin
    nbrecord:=nbrecord*(_IndexHdr.nbkey+1);
    lPos:=lPos.LowerLevel;
  end;
  result:=nbrecord;
end;

function TIndex.GetRecNo:Integer;
var
  pos:integer;
  p:TIndexPage;
begin
  p:=_Root;
  pos:=p._EntryNo;
  while p.Entry._LowerPage>0 do begin
    p:=p.LowerLevel;
    pos:=pos*(_IndexHdr.nbkey+1) + p._EntryNo;
  end;

  result:=pos;
(*
  pos:=1;
  While prev do begin
    inc(pos);
  end;
  SetRecNo(pos);
  Result:=pos;
*)
end;

(*
procedure TIndex.SetRecNo(Value: Integer);
var
  pos:integer;
  p:TIndexPage;
  i:integer;
  ldiv:integer;
begin
  p:=_Root;
  ldiv:=1;
  while p.Entry._LowerPage>0 do begin
    ldiv:=ldiv*(_IndexHdr.nbkey+1);
    p:=p._LowerLevel;
  end;
  pos:=value div ldiv;
  p:=_Root;
  while p.Entry._LowerPage>0 do begin
    p._EntryNo:=pos;
    value:=value - pos * (_IndexHdr.nbkey+1);
    ldiv:=ldiv div (_IndexHdr.nbkey+1);
    pos:=value div ldiv;
    p:=p._LowerLevel;
  end;
{
  pos:=1;
  First;
  While pos<value do begin
    if Next = false then break;
    inc(pos);
  end;
}
end;
*)
function TIndex.GetRealRecNo:integer;
var
  pos : TIndexPage;
begin
  pos:=_Root;
  while pos._LowerLevel<>nil do begin
    pos:=pos.LowerLevel;
  end;
  Result:=pos.Entry.RecNo-1;
end;

procedure TIndex.Resync;
var
  lPos:TIndexPage;
  ckey:array[0..100] of char;
  s:string;
  nkey:double;
begin
  if _IndexHdr.keytype='C' then begin
    TBS_Dbf(_IndexFile.FDbf).GetFieldData(_Field,@ckey);
    s:=cKey;
    Find(s,TBS_Dbf(_IndexFile.FDbf).FCurtRecNo+1,lPos);
  end else begin
    // Numeric field...
    TBS_Dbf(_IndexFile.FDbf).GetFieldData(_Field,@nkey);
    Find(nKey,TBS_Dbf(_IndexFile.FDbf).FCurtRecNo+1,lPos);
  end;
end;

procedure TIndex.GotoBookMark(recinfo:TRecInfo);
begin
  First;
  repeat
    if self.Pos.Entry.RecNo=recinfo.Bookmark + 1 then begin
      exit;
    end;
  until Next=false;
end;




function TBS_Dbf.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
var
  Memoi:array[1..32] of char;
  lBlob:TMyBlobFile;
begin
  lBlob:=TMyBlobFile.Create(Mode,Self,Field);
  if FDbfFile.GetFieldData(Field.FieldNo-1, ftString, ActiveBuffer,@Memoi[1]) then begin
    lBlob.MemoRecno:=StrToIntDef(Memoi,0);
    FdbtFile.ReadMemo(lBlob.MemoRecno,lBlob);
    lBlob.ReadSize:=lBlob.Size;
  end else lBlob.MemoRecno:=0;
  Result:=lBlob;
end;

procedure TBS_Dbf.Find(const Key);
var
  Recno:integer;
  lPos:TIndexPage;
begin
  if FCurIndex<>nil then begin
    recno:=FCurIndex.Find(Key,0,lPos);
    FCurtRecNo := recno - 1;
    Resync([]);
  end;
end;

procedure TBS_Dbf.UseIndex(IndexNo:integer);
var
	lIndex:TIndex;
begin
  if (IndexNo>=0) and (IndexNo<FIndexesList.Count) then begin
    lIndex:=TIndex(FIndexesList[IndexNo]);
    FCurIndex:=lIndex;
    Refresh;
  end;
end;

function TBS_Dbf.OpenIndex(IndexName:string):integer;
var
  idx:integer;
  idf:TIndexFile;
begin
  IndexName:=LowerCase(IndexName);
  for idx:=0 to FIndexFilesList.Count-1 do begin
    idf:= TIndexFile(FIndexFilesList[idx]);
    if idf.FStreamFileName=IndexName then begin
      result:=idx;
			exit;
    end;
  end;
  TIndexFile.Create(Self,IndexName,fmOpenReadWrite + fmShareDenyNone);

  result:=FIndexFilesList.Count-1;
end;
procedure TBS_Dbf.SetShowDeleted(v:boolean);
begin
  if v<>FShowDeleted then begin
    FShowDeleted:=v;
    Refresh;
  end;
end;


Procedure TBS_Dbf._AppendRecord(Buffer:PChar);
begin
  InternalLast;
  (Buffer)^ := ' ';
  FDbfFile.AppendRecord(Buffer)
end;


// I: open the table/file
procedure TBS_Dbf.InternalOpen;
begin
  _OpenFiles(false);
  // initialize the field definitions
  InternalInitFieldDefs;

  // if there are no persistent field objects,
  // create the fields dynamically
  if DefaultFields then
    CreateFields;
  // connect the TField objects with the actual fields
  BindFields (True);

  FCurtRecNo := BofCrack;

  FRecordInfoOffset := FDbfFile.RecordSize;
  FRecordBufferSize := FDbfFile.RecordSize + sizeof (TRecInfo);
  //_dbfFile.SetHdr(FDataHeader.FullHdrSize,FDataHeader.RecordSize);

  // set the bookmark size
  BookmarkSize := sizeOf (TRecInfo);

  // everything OK: table is now open

  FIsTableOpen := True;

// ShowMessage ('InternalOpen: RecCount: ' + IntToStr (FDataHeader.RecCount));
end;


// I: define the fields
procedure TBS_Dbf.InternalInitFieldDefs;
begin
  FieldDefs.Clear;
  with FieldDefs do
  begin
    if FDbfFile=nil then begin
      _OpenFiles(false);
      FDbfFile.CreateFieldDefs(FieldDefs);
      _CloseFiles;
    end else begin
      FDbfFile.CreateFieldDefs(FieldDefs);
    end;
  end;
end;


procedure TBS_Dbf._OpenFiles(Create:boolean);
var
  fileopenmode : integer;
  lIndexName,lMemoName : string;
  ff:string;
begin
  if not FileExists (FTableName) then begin
    ff:=ExtractFilePath(Application.Exename)+ExtractFileName(FTableName);
    if FileExists(ff) then FTableName:=ff;
  end;
  lIndexName:=ChangeFileExt(fTableName,'.mdx');
  lMemoName:=ChangeFileExt(fTableName,'.dbt');

  // check if the file exists
  if Create then begin
    FDbfFile := TBS_DbfFile.Create(Self,fTableName, fmCreate);
    //_indexFile := TIndexFile.Create(Self,lIndexName, fmCreate);
    FdbtFile := TDbtFile.Create(Self,lMemoName, fmCreate,FdbfFile.Version);
  end else if not FileExists (FTableName) then begin
    raise eBinaryDataSetError.Create ('Open: Table file not found : ' + FTableName);
  end else begin
    // create a File for the file
    if fReadOnly or (csDesigning in ComponentState) then
      fileopenmode := fmOpenRead or fmShareDenyNone
    else
      fileopenmode := fmOpenReadWrite or fmShareDenyNone;
      FDbfFile := TBS_DbfFile.Create(Self,fTableName, fileopenmode);

//    if FileExists (lIndexName) then begin
//      _indexFile := TIndexFile.Create(Self,lIndexName, fileopenmode);
//    end;

    if FileExists (lMemoName) then begin
      FdbtFile := TDbtFile.Create(Self, lMemoName, fileopenmode, FdbfFile.Version);
    end;
  end;
end;


procedure TBS_Dbf._CloseFiles;
var
  i:Integer;
begin
  if FDbfFile=nil then
    exit;

  FDbfFile.Free;
  FDbfFile:=nil;

  FdbtFile.Free;
  FdbtFile:=nil;

  for i:=0 to FIndexFilesList.Count-1 do begin
    TIndexFile(FIndexFilesList[i]).Free;
  end;
  FIndexFilesList.Clear;

  for i:=0 to FIndexesList.Count-1 do begin
    TIndex(FIndexesList[i]).Free;
  end;
  FIndexesList.Clear;
  FCurIndex:=nil;
  // free the internal list field offsets
end;


// TBS_Dbf.InternalClose
procedure TBS_Dbf.InternalClose;
begin
  if fReadOnly or (csDesigning in ComponentState) then
  begin
    // nothing...
  end  else begin
    FDbfFile.WriteHeader;
  end;

  // disconnet field objects
  BindFields (False);
  // destroy field object (if not persistent)
  if DefaultFields then
    DestroyFields;

  // close the file
  FIsTableOpen := False;
  _CloseFiles;
end;

// I: is table open
function TBS_Dbf.IsCursorOpen: Boolean;
begin
  Result := FIsTableOpen;
end;


// TBS_Dbf.Create
constructor TBS_Dbf.Create(AOwner:tComponent);
begin
  inherited create(aOwner);
  MemIndexList:=TList.Create;
  FIndexesList:=TList.Create;
  FIndexFilesList:=TList.Create;
  FCurIndex:=nil;
  ReadOnly:=True;
end;

destructor TBS_Dbf.Destroy;
begin
  inherited Destroy;
  _CloseFiles; // just in case, it should be already done in 'inherited destroy'
  MemIndexList.free;
  FIndexesList.free;
  FIndexFilesList.free;
end;

// II: set the requested bookmark as current record
procedure TBS_Dbf.InternalGotoBookmark(Bookmark: Pointer);
var
  RecInfo: TRecInfo;
begin
  RecInfo := TRecInfo(Bookmark^);
  if (RecInfo.Bookmark >= 0) and (RecInfo.Bookmark < FDbfFile.RecordCount) then begin
    FCurtRecNo := RecInfo.Bookmark;
    if FCurIndex<>nil then begin
      FCurIndex.GotoBookMark(RecInfo);
    end;
  end else
    raise eBinaryDataSetError.Create ('Bookmark ' +
      IntToStr (RecInfo.Bookmark) + ' not found');
end;

// II: same as above (but passes a buffer)
procedure TBS_Dbf.InternalSetToRecord (Buffer: PChar);
var
  ReqBookmark: Integer;
begin
//  ShowMessage ('InternalSetToRecord');
  ReqBookmark:=PRecInfo(Buffer + FRecordInfoOffset).Bookmark;
  InternalGotoBookmark (@ReqBookmark);
end;

// II: retrieve bookmarks flags from buffer
function TBS_Dbf.GetBookmarkFlag (
  Buffer: PChar): TBookmarkFlag;
begin
//  ShowMessage ('GetBookmarkFlag');
  Result := PRecInfo(Buffer + FRecordInfoOffset).BookmarkFlag;
end;

// II: change the bookmark flags in the buffer
procedure TBS_Dbf.SetBookmarkFlag (Buffer: PChar;
  Value: TBookmarkFlag);
begin
//  ShowMessage ('SetBookmarkFlag');
  PRecInfo(Buffer + FRecordInfoOffset).BookmarkFlag := Value;
end;

// II: read the bookmark data from record buffer
procedure TBS_Dbf.GetBookmarkData (
  Buffer: PChar; Data: Pointer);
begin
//  ShowMessage ('GetBookmarkData: ' + IntToStr (PRecInfo(Buffer + FRecordInfoOffset).Bookmark));
  Move(PRecInfo(Buffer + FRecordInfoOffset)^,Buffer^,sizeof(PRecInfo));
end;

// II: set the bookmark data in the buffer
procedure TBS_Dbf.SetBookmarkData (
  Buffer: PChar; Data: Pointer);
begin
//  ShowMessage ('SetBookmarkData: ' + IntToStr (PInteger(Data)^));
  Move(Data^, PRecInfo(Buffer + FRecordInfoOffset)^, sizeof(PRecInfo));
end;

// II: Go to a special position before the first record
procedure TBS_Dbf.InternalFirst;
begin
  if FCurIndex<>nil then FCurIndex.First;
  FCurtRecNo := BofCrack;
end;

// II: Go to a special position after the last record
procedure TBS_Dbf.InternalLast;
begin
  if FCurIndex<>nil then FCurIndex.Last;
  FCurtRecNo := FDbfFile.RecordCount; // 1 too much
end;

// II (optional): Record count
function TBS_Dbf.GetRecordCount: Longint;
begin
  if FCurIndex<>nil then begin
    Result:=FCurIndex.GetRecordCount;
  end else begin
    CheckActive;
    Result := FDbfFile.RecordCount; // 1 too much
  end;
end;

// II (optional): Get the number of the current record
function TBS_Dbf.GetRecNo: Integer;
begin
  UpdateCursorPos;
  if FCurIndex<>nil then begin
    Result:=FCurIndex.GetRecNo;
  end else begin
    if FCurtRecNo < 0 then
      Result := 1
    else
      Result := FCurtRecNo + 1;
  end;
end;

//
function TBS_Dbf.GetCanModify: Boolean;
begin
  result:=not FReadOnly;
end;

// TBS_Dbf.SetRecNo
// II (optional): Move to the given record number
(*
procedure TBS_Dbf.SetRecNo(Value: Integer);
begin
  CheckBrowseMode;
  if FCurIndex<>nil then begin
      FCurIndex.SetRecNo(Value);
  end else begin
    if (Value >= 1) and (Value <= FDbfFile.RecordCount) then
    begin
      FCurtRecNo := Value - 1;
      Resync([]);
    end;
  end;
end;
*)

/// III: Determine the size of each record buffer in memory
function TBS_Dbf.GetRecordSize: Word;
begin
  Result := FDbfFile.RecordSize; // data only
end;

/// III: Allocate a buffer for the record
function TBS_Dbf.AllocRecordBuffer: PChar;
begin
  Result := StrAlloc(FRecordBufferSize);
end;

// III: Initialize the record (set to zero)
procedure TBS_Dbf.InternalInitRecord(Buffer: PChar);
begin
  FillChar(Buffer^, FRecordBufferSize, 0);
  Buffer^ := ' ';
end;

// III: Free the buffer
procedure TBS_Dbf.FreeRecordBuffer (var Buffer: PChar);
begin
  StrDispose(Buffer);
end;

// III: Retrieve data for current, previous, or next record
// (eventually moving to it) and return the status
function TBS_Dbf.GetRecord(Buffer: PChar;
  GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var
  Acceptable : Boolean;
begin
  result := grOk;
  if FDbfFile.RecordCount < 1 then
    Result := grEOF
  else
  repeat
    if FCurIndex<>nil then begin
      case GetMode of
        gmCurrent :
          begin
            FCurtRecNo:=FCurIndex.GetRealRecNo;
          end;
        gmNext :
          begin
            if FCurIndex.Next=false then Result := grEOF
            else FCurtRecNo:=FCurIndex.GetRealRecNo;
          end;
        gmPrior :
          begin
            if FCurIndex.Prev=false then Result := grBOF
            else FCurtRecNo:=FCurIndex.GetRealRecNo;
          end;
      end;
      // fill record data area of buffer
    end else begin
      case GetMode of
        gmCurrent :
          begin
            // ShowMessage ('GetRecord Current');
            if (FCurtRecNo >= FDbfFile.RecordCount) or
                (FCurtRecNo < 0) then
              Result := grError;
          end;
        gmNext :
          begin
            if (FCurtRecNo < (FdbfFile.RecordCount)-1) then
              Inc (FCurtRecNo)
            else
              Result := grEOF;
          end;
        gmPrior :
          begin
            if (FCurtRecNo > 0) then
              Dec(FCurtRecNo)
            else
              Result := grBOF;
          end;
      end;
    end;
    if Result = grOK then begin
      FDbfFile.ReadRecord(FCurtRecNo,Buffer);
      if Buffer[0]='*' then
        FDelSgn:=1
      else
        FDelSgn:=0;
      ClearCalcFields(Buffer);
      GetCalcFields(Buffer);
      with PRecInfo(Buffer + FRecordInfoOffset)^ do
      begin
        BookmarkFlag := bfCurrent;
        Bookmark:=FCurtRecNo;
        if FCurIndex<>nil then begin
          IdxBookmark:=FCurIndex.GetRecNo;
        end else IdxBookmark:=-1;
      end;
    end else if (Result = grError) and DoCheck then
        raise eBinaryDataSetError.Create ('GetRecord: Invalid record');
    Acceptable := (FShowDeleted or (pRecordHdr(Buffer)^.DeletedFlag = ' '))
      and _FilterRecord(Buffer);
    if (GetMode=gmCurrent) and Not Acceptable then Result := grError;
  until (Result <> grOK) or Acceptable;
end;

///////////////////////////////////////////////////////
function TBS_Dbf._FilterRecord(Buffer: PChar): Boolean;
var
  SaveState: TDatasetState;
  s:string;
begin
  Result:=True;
  if Length(easyfilter)<>0 then begin
    SetString(s,buffer,RecordSize);
    s:=LowerCase(s);
    if Pos(easyfilter,s)=0 then begin
      Result:=False;
      Exit;
    end;
  end;
  if not Assigned(OnFilterRecord) then Exit;
  if not Filtered then Exit;
  FFilterBuffer:=buffer;
  SaveState:=SetTempState(dsFilter);
  OnFilterRecord(self,Result);
  RestoreState(SaveState);
end;

// III: Write the current data to the file
procedure TBS_Dbf.InternalPost;
begin
  CheckActive;
  if State = dsEdit then
  begin
    // replace data with new data
    if FCurIndex<>nil then begin
      FCurIndex.Delete;
    end;
    FDbfFile.WriteRecord(FCurtRecNo,ActiveBuffer);
  end else begin
    // append
    _AppendRecord(ActiveBuffer);
    //FCurtRecNo:=
  end;
  if FCurIndex<>nil then begin
    FCurIndex.Insert;
  end;
end;

// III: Add the current data to the file
procedure TBS_Dbf.InternalAddRecord(
  Buffer: Pointer; Append: Boolean);
begin
  _AppendRecord(ActiveBuffer);// always append
end;

// III: Delete the current record
procedure TBS_Dbf.InternalDelete;
begin
  CheckActive;
  pRecordHdr(ActiveBuffer)^.DeletedFlag := '*'; //FDataHeader.LastDeleted;
  Self.FDelSgn:=1;
  FDbfFile.WriteRecord(FCurtRecNo,ActiveBuffer);
  Resync([]);
end;

// III: Move data from record buffer to field
function TBS_Dbf.GetFieldData (
  Field: TField; Buffer: Pointer): Boolean;
var
  Ptr: Pointer;
begin
  Result := False;
  if State=dsFilter then begin
    Ptr:=FFilterBuffer;
  end else begin
    if IsEmpty or (Field.FieldNo <= 0) then exit;
    Ptr := ActiveBuffer;
  end;
  if Assigned(Buffer) then begin
    Result:=FdbfFile.GetFieldData(Field.FieldNo - 1,Field.DataType,Ptr,Buffer);
  end else begin
    if Field.DataType=ftBoolean then
      begin
       // for some reason buffer is null sometime for boolean...
       // for example when there is checkboxes
       // It mean form me that i should write in the pointer itself ??
       Result:=FdbfFile.GetFieldData(Field.FieldNo - 1,Field.DataType,Ptr,@Buffer);
      end
    else
      Result:=false;
  end;
end;

// III: Move data from field to record buffer
procedure TBS_Dbf.SetFieldData(Field: TField; Buffer: Pointer);
begin
  if (Field.FieldNo >= 0) and Assigned (Buffer) then begin
    FDbfFile.SetFieldData(Field.FieldNo - 1,Field.DataType,Buffer,ActiveBuffer);
    DataEvent (deFieldChange, Longint(Field));
  end;
end;


// Create a new table/file
procedure TBS_Dbf.CreateTable;
var
  Ix : Integer;
begin
  CheckInactive;
  //  InternalInitFieldDefs;
  if FieldDefs.Count = 0 then
  begin
    for Ix := 0 to FieldCount - 1 do
    begin
      with Fields[Ix] do
      begin
        if FieldKind = fkData then
          FieldDefs.Add(FieldName,DataType,Size,Required);
      end;
    end;
  end;
  _OpenFiles(true);
  try
    FDbfFile.CreateTable(FieldDefs);
  finally
    _CloseFiles;// close the file
  end;
end;

// default exception handling
procedure TBS_Dbf.InternalHandleException;
begin
  // standard exception handling
  Application.HandleException(Self);
end;

procedure TBS_Dbf.CloseBlob(Field: TField);
begin
  // I do not see anything to do.
end;


//Creation data file, to aTargetFile withour header
function  TBS_Dbf.CreateHeaderExtactedDataFile(aTargetFileName:String):boolean;
begin
  Result:=FDbfFile.CreateHeaderExtactedDataFile(aTargetFileName);
end;




///////////////// TBS_DbfFile //////////////////////
constructor TBS_DbfFile.Create(Parent:TDataSet; const FileName: string; Mode: Word);
var
  lRecordCount:Integer;
begin
  inherited;
  FFieldInfoList:=TList.Create;

  if fmCreate = Mode then
    begin
      FillChar(FDataHeader,sizeof(FDataHeader),0);
      FHeaderSize:=0;
      FRecordSize:=0;
      FDataHeader.VerDBF:=$03; // Default version xBaseIV without memo
      FDataHeader.Language:='X';
    end
  else
    begin
      FStreamFile.Seek(0,soFromBeginning);
      FStreamFile.ReadBuffer (FDataHeader, SizeOf(FDataHeader));
      case FDataHeader.VerDBF of
        $03,$83:          FVersion:=xBaseIII;
        $04,$8B,$8E,$7B:  FVersion:=xBaseIV;
        $05 :             FVersion:=xbaseV;
      else
        FVersion:=xBaseIV; // My favorite...
      end;
      FHeaderSize:=FDataHeader.FullHdrSize;  // 
      FRecordSize:=FDataHeader.RecordSize;   // 
      lRecordCount:=CalcRecordCount;      // 
      if FDataHeader.RecordCount <> lRecordCount then
        begin
          {ShowMessage('Invalid Record Count,'+^M
                     +'RecordCount in Hdr : '+IntToStr(FDataHeader.RecordCount)+^M
                     +'expected : '+IntToStr(lRecordCount));}

          FDataHeader.RecordCount := lRecordCount;
        end;
    end;
end;


destructor TBS_DbfFile.destroy;
begin
  CleardbfFieldInfos;
  FFieldInfoList.free;
  inherited;
end;

procedure TBS_DbfFile.CleardbfFieldInfos;
var i:Integer;
begin
  for i:=0 to FFieldInfoList.Count-1 do begin
    TdbfFieldInfo(FFieldInfoList.Items[i]).Free;
  end;
  FFieldInfoList.Clear;
end;

procedure TBS_DbfFile.AppendRecord(buffer:Pointer);
begin
  Inc(FDataHeader.RecordCount);
  WriteRecord(FDataHeader.RecordCount-1,buffer);
end;

procedure TBS_DbfFile.WriteHeader;
var
  SystemTime: TSystemTime;
  lAfterHdrIII:rAfterHdrIII;
  lAfterHdrV:rAfterHdrV;
  lterminator:Byte;
begin
  Assert(FStreamFile<>nil,'_dbfFile=Nil');
  GetLocalTime(SystemTime);

  FDataHeader.Year := SystemTime.wYear - 1900;
  FDataHeader.Month := SystemTime.wMonth;
  FDataHeader.Day := SystemTime.wDay;
  FStreamFile.Seek(0,soFromBeginning);
  FStreamFile.WriteBuffer (FDataHeader, SizeOf(FDataHeader));

  if FVersion >= xBaseV then begin
    FillChar(lAfterHdrV,SizeOf(lAfterHdrV),0);
    FStreamFile.WriteBuffer (lAfterHdrV, SizeOf(lAfterHdrV));
  end else begin
    FillChar(lAfterHdrIII,SizeOf(lAfterHdrIII),0);
    FStreamFile.WriteBuffer (lAfterHdrIII, SizeOf(lAfterHdrIII));
  end;
  Seek(FDataHeader.RecordCount); // last byte usually...
  lterminator := $1A;
  FStreamFile.Write(lterminator,SizeOf(lterminator));
end;

procedure TBS_DbfFile.CreateFieldDefs(FieldDefs:TFieldDefs);
var
  lColumnCount,lHeaderSize,lFieldSize:Integer;
  Il : Integer;
  lFieldOffset : Integer;
  fn:string;
  ft:TFieldType;
  fs,nfs,fd:Integer;
  dbfFieldInfo:TdbfFieldInfo;
  lFieldHdrIII:rFieldHdrIII;
  lFieldHdrV:rFieldHdrV;
  charft:char;

  function ToFieldType(dbasetype:char;fs,fd:Integer):TFieldType;
  begin
    case dbasetype of
    'C' :
      begin
        Result:=ftString;
      end;
    'L' :
      begin
        Result:=ftBoolean;
      end;
    'F' :
      begin
        Result:=ftFloat;
      end;
    'N' :
      begin
        if fd=0 then begin
          if fs <= 4 then begin
            Result:=ftSmallInt;
          end else if fs <= 9 then begin
            Result:=ftInteger;
          end else begin
{$ifdef NO_LARGE_INT}
            Result:=ftInteger;
{$else}
            Result:=ftLargeInt;
{$endif}
          end;
        end else begin
          Result:=ftFloat;
        end;
      end;
    'D' :
      begin
        Result:=ftDate;
      end;
    'M' :
      begin
        Result:=ftMemo;
      end;
    else
      begin
        Result:=ftString;
      end;
    end; //case
  end;
begin
  CleardbfFieldInfos;

  if FVersion>=xBaseV then begin
    lHeaderSize:=SizeOf(rAfterHdrV) + SizeOf(rDbfHdr);
    lFieldSize:=SizeOf(rFieldHdrV);
  end else begin
    lHeaderSize:=SizeOf(rAfterHdrIII) + SizeOf(rDbfHdr);
    lFieldSize:=SizeOf(rFieldHdrIII);
  end;
  lColumnCount:= (FDataHeader.FullHdrSize - lHeaderSize) div lFieldSize;

  if (lColumnCount <= 0) or (lColumnCount > 255) then
      Raise eBinaryDataSetError.Create('Invalid Field count : ' + IntToStr(lColumnCount) + ' (must be between 1 and 255)');

  lFieldOffset := 1;

  FStreamFile.Position := lHeaderSize;
  for Il:=0 to lColumnCount-1 do begin
    if FVersion>=xBaseV then
      begin
        FStreamFile.ReadBuffer(lFieldHdrV,SizeOf(lFieldHdrV));
        fn:=PCHAR(@lFieldHdrV.FieldName[0]);
        fs:=lFieldHdrV.FieldSize;
        fd:=lFieldHdrV.FieldPrecision;
        nfs:=fs;
        ft:=ToFieldType(lFieldHdrV.FieldType,nfs,fd);
        charft:=lFieldHdrV.FieldType;
      end
    else
      begin
        FStreamFile.ReadBuffer(lFieldHdrIII,SizeOf(lFieldHdrIII));
        fn:=PCHAR(@lFieldHdrIII.FieldName[0]);
        fs:=lFieldHdrIII.FieldSize;
        fd:=lFieldHdrIII.FieldPrecision;
        nfs:=fs;
        ft:=ToFieldType(lFieldHdrIII.FieldType,nfs,fd);
        charft:=lFieldHdrIII.FieldType;
      end;

    // first create the bde field
    if ft in [ftString,ftBCD] then
      fieldDefs.Add(fn,ft,fs,false)
    else
      fieldDefs.Add(fn,ft,0,false);
    // then create the for our own fieldinfo
    dbfFieldInfo:=TdbfFieldInfo.Create;
    dbfFieldInfo.Offset:=lFieldOffset;
    dbfFieldInfo.Size:=fs;
    dbfFieldInfo.Prec:=fd;
    dbfFieldInfo.FieldName:=fn;
    dbfFieldInfo.charFieldType:=charft;
    FFieldInfoList.Add(dbfFieldInfo);
    Inc(lFieldOffset,fs);
  end;
  if (lFieldOffset <> FDataHeader.RecordSize) then
  begin
    raise eBinaryDataSetError.Create ('Invalid Record Size,'+^M+
                                     'Record Size in Hdr : '+IntToStr(FDataHeader.RecordSize)+^M+
                                     'Expected : '+IntToStr(lFieldOffset));
    FDataHeader.RecordSize := lFieldOffset;
  end;
end;

procedure TBS_DbfFile.CreateTable(FieldDefs:TFieldDefs);
var
  ix:Integer;
  lFieldHdrIII:rFieldHdrIII;
  lType:Char;
  lSize,lPrec:Integer;
  Offs:Integer;
  lterminator:Byte;
begin
  // first reset file.
  FStreamFile.Size:= 0;
  FStreamFile.Position:=SizeOf(rDbfHdr) + SizeOf(rAfterHdrIII);
  Offs:=1; // deleted mark count 1.
  for Ix:=0 to FieldDefs.Count-1 do
  begin
    with FieldDefs.Items[Ix] do
    begin
      FillChar(lFieldHdrIII,SizeOf(lFieldHdrIII),#0);
      lPrec:=0;
      case DataType of
        ftString:
          begin
            ltype:='C';
            lSize := Size;
          end;
        ftBoolean:
          begin
            ltype:='L';
            lSize := 1;
          end;
        ftSmallInt:
          begin
            ltype:='N';
            lSize := 6;
          end;
        ftInteger:
          begin
            ltype:='N';
            lSize := 11;
          end;
{$ifndef NO_LARGE_INT}
        ftLargeInt:
          begin
            ltype:='N';
            lSize := 20;
            lPrec := 0;
          end;
{$endif}
        ftFloat:
          begin
            ltype:='N';
            lSize := 20;
            lPrec := 4;
          end;
        ftDate:
          begin
            ltype:='D';
            lSize := 8;
          end;
        ftMemo:
          begin
            ltype:='M';
            lSize := 10;
          end;
        else
          begin
            raise EBinaryDataSetError.Create (
             'InitFieldsDefs: Unsupported field type');
          end;
      end; // case

      lFieldHdrIII.FieldType:=ltype; //DataType;
      StrPCopy(lFieldHdrIII.FieldName,FieldDefs.Items[Ix].Name);
      lFieldHdrIII.FieldSize:=lSize;
      lFieldHdrIII.FieldPrecision:=lPrec;

      FStreamFile.Write(lFieldHdrIII,SizeOf(lFieldHdrIII));
      Inc(Offs,lSize);
    end;
  end;
  // end of header
  lterminator := $0d;
  FStreamFile.Write(lterminator,SizeOf(lterminator));

  // update header
  FDataHeader.RecordSize := Offs;
  FDataHeader.FullHdrSize := FStreamFile.Position;
  FRecordSize := FDataHeader.RecordSize;
  FHeaderSize := FDataHeader.FullHdrSize;
  // write the updated header
  FStreamFile.Position:=0;
  WriteHeader;
end;

function TBS_DbfFile.GetFieldData(Column:Integer;DataType:TFieldType; Src,Dst: Pointer): Boolean;
var
  FieldOffset: Integer;
  FieldSize: Integer;
  s:string;
  d:TDateTime;
  ld,lm,ly:Integer;
  dbfFieldInfo:TdbfFieldInfo;
begin
  dbfFieldInfo:=TdbfFieldInfo(FFieldInfoList[Column]);
  FieldOffset := dbfFieldInfo.Offset;
  FieldSize := dbfFieldInfo.Size;
  SetString(s, PChar(Src) + FieldOffset, FieldSize );
  if DataType = ftBoolean then
    begin
      if (s[1]='T') then
        Word(Dst^) := 1
      else
        Word(Dst^) := 0;
    end
  else
   if DataType in [ftInteger,ftSmallInt
    {$ifndef NO_LARGE_INT},ftLargeInt{$endif}] then
   begin
     ld:=StrToIntDef(s,0);
     case DataType of
       ftSmallInt : SmallInt(Dst^) := ld;
{$ifndef NO_LARGE_INT}
       ftLargeint : LargeInt(Dst^) := ld;
{$endif}
     else // ftInteger :
       Integer(Dst^) := ld;
     end; // case
   end
  else
   if DataType = ftFloat then
     begin
       Double(Dst^) := DBFStrToFloat(s);
     end
   else if DataType = ftDate then
     begin
       if Trim(s)='' then
         begin
           ld:=30;
           lm:=12;
           ly:=1899;
         end
       else
         begin
           ld:=StrToIntDef(Copy(s,7,2),1);
           lm:=StrToIntDef(Copy(s,5,2),1);
           ly:=StrToIntDef(Copy(s,1,4),0);
         end;
       if ld=0 then ld:=1;
       if lm=0 then lm:=1;
       //if (ly < 1899) or (ly > 4712) then ly:= 4712;
       try
         d:=EncodeDate(ly,lm,ld);
         Integer(Dst^) := DateTimeToTimeStamp(d).Date;
       except
         Integer(Dst^) := 0;
       end;
     end
    else if DataType = ftString then
      begin
        s:=Trim(s);
        StrPCopy(Dst,s);
        OemToChar(Dst,Dst);
      end;
  Result := True;
end;

procedure TBS_DbfFile.SetFieldData(Column:integer;DataType:TFieldType; Src,Dst: Pointer);
var
  FieldSize,FieldPrec: Integer;
  s:string;
  li:Longint;
  fl:Double;
  da:TDateTime;
  ly,lm,ld:Word;
  ts:TTimeStamp;
  dbfFieldInfo:TdbfFieldInfo;
begin
  dbfFieldInfo:=TdbfFieldInfo(FFieldInfoList[Column]);
  FieldSize := dbfFieldInfo.Size;
  FieldPrec := dbfFieldInfo.Prec;
  Dst:=PChar(Dst)+dbfFieldInfo.Offset;
  if DataType = ftBoolean then begin
    if Word(Src^) = 1 then s:='T'
    else s:='F';
  end  else if DataType in [ftInteger,ftSmallInt
   {$ifndef NO_LARGE_INT},ftLargeInt{$endif}] then begin
    case DataType of
    ftSmallInt : li := SmallInt(Src^);
{$ifndef NO_LARGE_INT}
    ftLargeInt: li := LargeInt(Src^);
{$endif}
    else //ftInteger
        li := Integer(Src^);
    end;
    s:=IntToStr(li);
    if Length(s)<FieldSize then s:=StringOfChar(' ',FieldSize-Length(s)) + s;
  end else if DataType = ftFloat then begin
    fl := Double(Src^);
    s:=FloatToDbfStr(fl,FieldSize,FieldPrec);
    if Length(s)<FieldSize then s:=StringOfChar(' ',FieldSize-Length(s)) + s;
  end  else if DataType = ftDate then begin
    ts.Time:=0;
    ts.Date:=Integer(Src^);
    da:=TimeStampToDateTime(ts);
    DecodeDate(da,ly,lm,ld);
    FmtStr(s,'%4d%2d%2d',[ly,lm,ld]);
  end  else if DataType = ftString then begin
    CharToOem(Src,Src);
    s:=PChar(Src); // finish with first 0
  end;
  if Length(s)<FieldSize then begin
    s:=s+StringOfChar(' ',FieldSize-Length(s));
  end else if Length(s)>FieldSize then begin
    raise eFieldToLongError.Create('Field too long :' + IntToStr(Length(s))
     + ' (must be between 1 and ' + IntToStr(FieldSize) + ').');
  end;
  Move(PChar(s)^, Dst^, FieldSize);
end;

// TBS_Dbf.RecordCount
function TBS_DbfFile.RecordCount:Integer;
begin
  Result:=FDataHeader.RecordCount;
end;


function TBS_DbfFile.GetFieldInfo(AFieldName:String;
                                  var dbfFieldInfo:TdbfFieldInfo): Boolean;
var
  i:integer;

begin
  Result := False;
  for i:=0 to FFieldInfoList.Count-1 do
    begin
      dbfFieldInfo:=TdbfFieldInfo(FFieldInfoList[i]);
      if dbfFieldInfo.FieldName=UpperCase(AFieldName) then
        begin
          Result := True;
          break;
        end;
    end;
end;



//Creation data file, to aTargetFile withour header
function TBS_DbfFile.CreateHeaderExtactedDataFile(aTargetFileName:String):boolean;
var
  memStream:TmemoryStream;
begin
  Result:=True;
  memStream:=TmemoryStream.Create;
  try
    try
      FStreamFile.Seek(FHeaderSize,soFromBeginning);
      memStream.CopyFrom(FStreamFile,FStreamFile.Size-FHeaderSize);
      memStream.Position:=0;
      memStream.SaveToFile(aTargetFileName);
    except
      Result:=False;
    end;
  finally
    memStream.Free;
  end;
end;

function  TBS_DbfFile.GetOraLoaderStr(const AFieldName:string; const AOraFieldName:String; aUsePosition:Boolean): string;
var
  dbfFieldInfo:TdbfFieldInfo;
  theTmpStr:String;
  theFieldAdress:integer;
  theFieldsize:integer;
begin
  try

    //dbfFieldInfo:=TdbfFieldInfo.Create;
    theTmpStr:='';
    if GetFieldInfo(AFieldName,dbfFieldInfo) then
     begin
      theFieldAdress:=dbfFieldInfo.Offset+1;//Delete marker
      theFieldSize:=theFieldAdress+dbfFieldInfo.Size-1;
      theTmpStr:='  '+AOraFieldName;
      if aUsePosition then
        theTmpStr:=theTmpStr+' POSITION('+IntToStr(theFieldAdress)+':'+IntToStr(theFieldSize)+') ';

      if dbfFieldInfo.charFieldType='C' then
         theTmpStr:=theTmpStr+' CHAR NULLIF '+AOraFieldName+' = BLANKS'
      else
       if dbfFieldInfo.charFieldType='N' then
          begin
           if dbfFieldInfo.Prec=0 then
               theTmpStr:=theTmpStr+' INTEGER EXTERNAL '
           else
             theTmpStr:=theTmpStr+' DECIMAL EXTERNAL ';
           theTmpStr:=theTmpStr+' "NVL(:'+AOraFieldName+',0)"';
          end
        else
          if dbfFieldInfo.charFieldType='D' then
            theTmpStr:=theTmpStr+' DATE "YYYYMMDD" NULLIF '+AOraFieldName+'= BLANKS'
          else
            if dbfFieldInfo.charFieldType='L' then
              theTmpStr:=theTmpStr
                +' CHAR "DECODE(:'+AOraFieldName+','+CHR(39)+'T'+CHR(39)+',1,0)"'
            else
              theTmpStr:=theTmpStr+' NULL';
       Result:=theTmpStr;
     end
    else
      Result:='  '+AOraFieldName+' "NULL"'; //!!!   
     // --theTmpStr:='Field not found /'+AOraFieldName+'/';//

  finally
   // dbfFieldInfo.destroy;
  end;
end;

function TBS_DbfFile.GetOraLoaderStr(const AFieldName:string;
                                     const AOraFieldName:String): string;
begin
  Result:=GetOraLoaderStr(aFieldName,aOraFieldName,True);
end;


function  TBS_Dbf.GetFieldInfo(AFieldName:String; var dbfFieldInfo:TdbfFieldInfo): Boolean;
begin
  Result:=Self.FDbfFile.GetFieldInfo(AFieldName,dbfFieldInfo);
end;

//aFieldsNames - comma separated field names
function TBS_Dbf.GetLoaderDataStr(const aFieldsNames:string;
 aEncloseChar,aReplaceEncloseChar,aFieldsDelimiter:Char): string;
var
  dbfFieldInfo:TdbfFieldInfo;
  fieldsNames:String;
  fieldName:String;
  tmpStr:String;
  delemiterChar:Char;
  delemiterPos:integer;
  oldShortDateFormat:String;
  oldDateSeparator:Char;
  tmpDateStr:String;

function ReplaceStr(const S, Srch, Replace: string): string;
  var
    i: integer;
    Source: string;
  begin
    Source := S;
    Result := '';
    repeat
      i := Pos(Srch, Source);
      if i > 0 then
        begin
          Result := Result + Copy(Source, 1, i - 1) + Replace;
          Source := Copy(Source, i + Length(Srch), Maxint);
        end
          else Result := Result + Source;
    until i <= 0;
end;

begin
  delemiterChar:=',';
  fieldsNames:=aFieldsNames+',';
  delemiterPos:=pos(delemiterChar,fieldsNames);
  tmpStr:='';
  while delemiterPos > 0 do
    begin
      fieldName:=Copy(fieldsNames,1,delemiterPos-1);
      fieldsNames:=Copy(fieldsNames,delemiterPos+1,Length(fieldsNames));
      delemiterPos:=pos(delemiterChar,fieldsNames);
      if GetFieldInfo(fieldName,dbfFieldInfo) then
        begin
          if dbfFieldInfo.charFieldType='C' then
             tmpStr:=tmpStr+aEncloseChar
              +ReplaceStr(Trim(Self.FieldByName(fieldName).AsString),aEncloseChar,aReplaceEncloseChar)
              +aEncloseChar+aFieldsDelimiter
          else
            if dbfFieldInfo.charFieldType='N' then
              tmpStr:=tmpStr+Trim(Self.FieldByName(fieldName).AsString)+aFieldsDelimiter
            else
              if dbfFieldInfo.charFieldType='D' then
                begin
                  oldShortDateFormat:=ShortDateFormat;
                  oldDateSeparator:=DateSeparator;
                  try
                    ShortDateFormat:='YYYYMMDD';
                    DateSeparator:='/';
                    tmpDateStr:=ReplaceStr(DateToStr(Self.FieldByName(fieldName).asDateTime),'/','');
                    if Not (Self.FieldByName(fieldName).Value=0) then
                      tmpDateStr:=ReplaceStr(DateToStr(Self.FieldByName(fieldName).asDateTime),'/','')
                    else
                      tmpDateStr:='        ';
                    tmpStr:=tmpStr+tmpDateStr+aFieldsDelimiter;
                  finally
                    ShortDateFormat:=oldShortDateFormat;
                    DateSeparator:=oldDateSeparator;
                  end;
                end
              else
                if dbfFieldInfo.charFieldType='L' then
                  begin
                    if Self.FieldByName(fieldName).Value=True then
                      tmpStr:=tmpStr+'1'+aFieldsDelimiter
                    else
                      tmpStr:=tmpStr+'0'+aFieldsDelimiter
                  end
            else
              tmpStr:=tmpStr+Trim(Self.FieldByName(fieldName).AsString)+aFieldsDelimiter;
        end;
    end;
  tmpStr:=Copy(tmpStr,1,Length(tmpStr)-1);
  Result:=tmpStr;
end;

function TBS_Dbf.DelSgn: integer;
begin
  UpdateCursorPos;
  Result:=FDelSgn;
end;

end.

