// Copyright  2000 by Berdachuck S.I.   Delphi 5
// Node Classes
unit uBS_NodeClasses;
interface

uses
  Forms, SysUtils, Classes, WinTypes, WinProcs, CommCtrl,Controls,
  DB,
  Oracle, OracleData,
  IB, IBCustomDataSet,IBServices, IBSQL,IBDatabase,
  frmuObjectWindow,
  dmuDM;


type

  TBS_TreeNode = class(TComponent)
  private
    FNodeID: HTreeItem;
    FNodeName: string;
    FNodeType: word;
    FShowSystem: boolean;
    FObjectList: TStringList;
    FRefreshed : Boolean;
  protected
  public
    property NodeID: HTreeItem read FNodeID;
    property NodeName: string read FNodeName write FNodeName;
    property NodeType: word read FNodeType;
    property ShowSystem: boolean read FShowSystem write FShowSystem;
    property ObjectList: TStringList read FObjectList write FObjectList;
    property Refreshed: boolean read FRefreshed write FRefreshed;
    constructor Create(AOwner: TComponent; const NodeID: HTreeItem; const NodeName: string; const NodeType: word); reintroduce;
    destructor Destroy(); override;
  published
  end;




/////////////////////////////////////////////////////////
 TBS_TaskNode = class(TBS_TreeNode)
    NodeDataSet : TDataset;
  private
    FDataSetRecord:TBookMark;
  protected
   //
  public
    constructor Create(AOwner: TComponent;
               const NodeID: HTreeItem;
               ANodeDataSet : TDataset);
    destructor Destroy(); override;
    procedure GotoDataRecord;// Try Get Marked with BookMark Record
    function GetDataFieldValue(const AFieldName:string):Variant;
    function GetDataFieldAsString(const AFieldName:string):String;
  published//
  end;


  ////////////////////////////////////////
  TBS_ConnectionNode = class(TBS_TreeNode)
  private
    FDescription: string;
    FLastAccessed: TDateTime;
    function GetConnectionStatus:boolean;
  protected
  public
    property Connected: boolean read GetConnectionStatus;
    property Description: string read FDescription write FDescription;
    property LastAccessed: TDateTime read FLastAccessed write FLastAccessed;

    constructor Create( AOwner: TComponent;
                  const NodeID: HTreeItem;
                  const NodeName: String;
                  const NodeType: word);
    destructor Destroy(); override;
  published
  end;


  ////////////////////////////////////////
  TBS_OraSessionNode = class(TBS_TreeNode)
  private
    FDescription: string;
    FLastAccessed: TDateTime;
    FOracleSession: TOracleSession;
    //FOutputWindow: TfrmTextViewer;
    function GetOraConnectionStatus:boolean;
    function GetOraDatabase:String;
    function GetOraPassword:String;
    function GetOraUserName:String;

    procedure SetOraDatabase(const AOraDatabase:String);
    procedure SetOraPassword(const AOraPassword:String);
    procedure SetOraUserName(const AOraUserName:String);
  protected
  public
    property Active: boolean read GetOraConnectionStatus;
    property Database: string read GetOraDatabase write SetOraDatabase;
    property Password: string read GetOraPassword write SetOraPassword;
    property UserName: string read GetOraUserName write SetOraUserName;
    property Description: string read FDescription write FDescription;
    property LastAccessed: TDateTime read FLastAccessed write FLastAccessed;
    property OracleSession: TOracleSession read FOracleSession write FOracleSession;

    //property OutputWindow: TfrmTextViewer read FOutputWindow;

    constructor Create(AOwner: TComponent;
                       const NodeID: HTreeItem;
                       const NodeName,
                             Database,
                             Password,
                             UserName,
                             Description: string;
                       const LastAccessed: TDateTime;
                       const NodeType: word);

    destructor Destroy(); override;
    procedure Login;
    procedure LogOut;
    {
    procedure ShowText (const Data: TStringList; const Title: String;
        const readonly: boolean=true);
    procedure OpenTextViewer (const Service: TIBControlAndQueryService;
        const Title: String; const readonly: boolean=true);
    }
  published
  end;



  TBS_IbServerNode = class(TBS_TreeNode)
  private
    FServerName: string;
    FUserName: string;
    FPassword: string;
    FVersion : Integer;
    FDatabasesID: HTreeItem;
    FBackupFilesID: HTreeItem;
    //FServer: TIBServerProperties;
    FDescription: string;
    FLastAccessed: TDateTime;
    //FOutputWindow: TfrmTextViewer;
  protected
  public
    property Servername: string read FServername write FServername;
    property UserName: string read FUserName write FUserName;
    property Password: string read FPassword write FPassword;
    property DatabasesID: HTreeItem read FDatabasesID write FDatabasesID;
    property Version : Integer read FVersion write FVersion;
    property Description: string read FDescription write FDescription;
    property LastAccessed: TDateTime read FLastAccessed write FLastAccessed;
    property BackupFilesID: HTreeItem read FBackupFilesID write FBackupFilesID;
    //property Server: TIBServerProperties read FServer write FServer;
    //property OutputWindow: TfrmTextViewer read FOutputWindow;

    constructor Create(AOwner: TComponent;
                       const NodeID: HTreeItem;
                       const NodeName,ServerName,UserName,Password, Description: string;
                       const Protocol: TProtocol;
                       const LastAccessed: TDateTime;
                       const NodeType: word);
    destructor Destroy(); override;
    procedure ShowText (const Data: TStringList; const Title: String;
        const readonly: boolean=true);
    procedure OpenTextViewer (const Service: TIBControlAndQueryService;
        const Title: String; const readonly: boolean=true);
  published
  end;




  TBS_IBDatabaseNode = class(TBS_TreeNode)
  private
    FDatabaseFiles: TStringList;
    FUserName: string;
    FPassword: string;
    FRole: string;
    FCaseSensitiveRole: boolean;
    FDefaultBackupAlias: string;
    FSQLBuffer: TStringList;
    FSQLBufferPos: integer;
    FDatabase: TIBDatabase;
    FTransaction: TIBTransaction;
    FQryTransaction: TIBTransaction;
    FDataSet: TIBDataSet;
    FQryDataSet: TIBDataSet;
    FDataSource: TDataSource;
    FQryDataSource: TDataSource;
    FQuery: TIBSQL;

    FTablesID: HTreeItem;
    FDomainsID: HTreeItem;
    FViewsID: HTreeItem;
    FProceduresID: HTreeItem;
    FFunctionsID: HTreeItem;
    FGeneratorsID: HTreeItem;
    FExceptionsID: HTreeItem;
    FFiltersID: HTreeItem;
    FRolesID: HTreeItem;
    FTriggersID: HTreeItem;

    FObjectViewer: TFrmObjectView;

  protected
  public
    property DatabaseFiles: TStringList read FDatabaseFiles write FDatabaseFiles;
    property UserName: string read FUserName write FUserName;
    property Password: string read FPassword write FPassword;
    property Role: string read FRole write FRole;
    property CaseSensitiveRole: boolean read FCaseSensitiveRole write FCaseSensitiveRole;
    property DefaultBackupAlias: string read FDefaultBackupAlias write FDefaultBackupAlias;
    property SQLBuffer: TStringList read FSQLBuffer write FSQLBuffer;
    property SQLBufferPos: integer read FSQLBufferPos write FSQLBufferPos;
    property Database: TIBDatabase read FDatabase write FDatabase;
    property Transaction: TIBTransaction read FTransaction write FTransaction;
    property QryTransaction: TIBTransaction read FQryTransaction write FQryTransaction;
    property DataSet: TIBDataSet read FDataSet write FDataSet;
    property QryDataSet: TIBDataSet read FQryDataSet write FQryDataSet;
    property DataSource: TDataSource read FDatasource write FDatasource;
    property QryDataSource: TDataSource read FQryDatasource write FQryDatasource;
    property Query: TIBSQL read FQuery write FQuery;

    property TablesID: HTreeItem read FTablesID write FTablesID;
    property DomainsID: HTreeItem read FDomainsID write FDomainsID;
    property ViewsID: HTreeItem read FViewsID write FViewsID;
    property ProceduresID: HTreeItem read FProceduresID write FProceduresID;
    property FunctionsID: HTreeItem read FFunctionsID write FFunctionsID;
    property GeneratorsID: HTreeItem read FGeneratorsID write FGeneratorsID;
    property ExceptionsID: HTreeItem read FExceptionsID write FExceptionsID;
    property FiltersID: HTreeItem read FFiltersID write FFiltersID;
    property RolesID: HTreeItem read FRolesID write FRolesID;
    property TriggersID: HTreeItem read FTriggersID write FTriggersID;

    property ObjectViewer: TFrmObjectView read FObjectViewer;

    constructor Create(AOwner: TComponent; const NodeID: HTreeItem;
      const NodeName: string; const NodeType: word;
      const DatabaseFiles: TStringList; var NewDatabase: TIBDatabase);
    destructor Destroy(); override;
    procedure CreateObjectViewer;
  published
  end;


implementation





constructor TBS_IbServerNode.Create(AOwner: TComponent; const NodeID: HTreeItem; const NodeName,Servername,UserName,Password, Description: string; const Protocol: TProtocol; const LastAccessed: TDateTime; const NodeType: word);
begin
{
  inherited Create(AOwner,NodeID,NodeName,NodeType);
  FServername := Servername;
  FUserName := UserName;
  FPassword := Password;
  FVersion := 6;
  FServer := TIBServerProperties.Create(nil);
  FServer.ServerName := ServerName;
  FServer.Protocol := Protocol;
  FServer.LoginPrompt := false;
  FDescription := Description;
  FLastAccessed := LastAccessed;
  FOutputWindow := nil;
  }
end;

destructor TBS_IbServerNode.Destroy();
begin
{
  try
    if Self.Server.Active then
      FServer.Detach;
    FServer.Free;
  except on E: Exception do
    FServer.Free;
  end;
  inherited Destroy();
  }
end;

constructor TBS_IBDatabaseNode.Create(AOwner: TComponent; const NodeID: HTreeItem;
  const NodeName: string; const NodeType: word; const DatabaseFiles: TStringList;
  var NewDatabase: TIBDatabase);
var
  tmp: string;
begin
  inherited Create(AOwner,NodeID,NodeName,NodeType);

  FObjectViewer := nil;
  FDatabaseFiles := TStringList.Create;

  if not Assigned(NewDatabase) then
    FDatabase := TIBDatabase.Create(AOwner)
  else
    FDatabase := NewDatabase;

  FTransaction := TIBTransaction.Create(AOwner);
  FQryTransaction := TIBTransaction.Create(AOwner);
  FDataSet := TIBDataSet.Create (AOwner);
  FQryDataSet := TIBDataSet.Create (AOwner);
  FDataSource := TDataSource.Create (AOwner);
  FQryDataSource := TDataSource.Create (AOwner);
  FQuery := TIBSQL.Create(AOwner);
  FQuery.Database := FDatabase;
  FQuery.Transaction := FTransaction;

  with FDataSet do
  begin
    Database := FDatabase;
    Transaction := FTransaction;
    ObjectView := true;
    SparseArrays := true;
  end;

  with FQryDataSet do
  begin
    Database := FDatabase;
    Transaction := FQryTransaction;
    ObjectView := true;
    SparseArrays := true;
  end;

  FTransaction.DefaultDatabase := FDatabase;
  FQryTransaction.DefaultDatabase := FDatabase;
  FQryDataSource.DataSet := FQryDataset;

  FDataSource.DataSet := FDataSet;
  FDataSource.AutoEdit := false;

  FQuery.GoToFirstRecordOnExecute := true;
  FSQLBuffer := TStringList.Create;
  FSQLBufferPos := 0;
  FDatabaseFiles.Assign(DatabaseFiles);
  FDatabase.LoginPrompt := false;
  if FDatabaseFiles.Count > 0  then
  begin
    { Force the database to have a full path }
    if ExtractFilePath(FDatabaseFiles.Strings[0]) = '' then
    begin
      tmp := ExtractFilePath(Application.ExeName);
      FDatabase.DatabaseName := tmp + FDatabaseFiles.Strings[0];
    end
    else
      FDatabase.DatabaseName := FDatabaseFiles.Strings[0];
  end;
  FDatabase.DefaultTransaction := FTransaction;  
  
end;

procedure TBS_IBDatabaseNode.CreateObjectViewer;
begin
  if not Assigned(FObjectViewer) then
    FObjectViewer := TfrmObjectView.Create(Application)
  else
    FObjectViewer.WindowState := wsNormal;
end;

destructor TBS_IBDatabaseNode.Destroy();
begin
  if Assigned (FDatabase) and Assigned(FDatabase.Handle) then
  begin
    try
      FDatabase.Connected := false;
    except on E: Exception do
      FDatabase.Free;
    end;
  end;
  FSQLBuffer.Free;
  FDatabaseFiles.Free;
  FTransaction.Free;
  inherited Destroy();
end;




procedure TBS_IbServerNode.ShowText(const Data: TStringList;
  const Title: String; const readonly:boolean=true);
begin
{
  if not Assigned (FOutputwindow) then
    FOutputWindow := TfrmTextViewer.Create(Application);
  FOutputWindow.ShowText(Data, Title, Readonly);
  }
end;

procedure TBS_IbServerNode.OpenTextViewer(const Service: TIBControlAndQueryService;
  const Title: String; const readonly:boolean=true);
begin
{
  if not Assigned (FOutputwindow) then
    FOutputWindow := TfrmTextViewer.Create(Application);
  FOutputWindow.OpenTextViewer (Service, Title, readonly);
  }
end;



////////////////////  TBS_TreeNode  ///////////////////////
constructor TBS_TreeNode.Create(
  AOwner: TComponent;
  const NodeID: HTreeItem;
  const NodeName: string;
  const NodeType: word);
begin
  inherited Create(AOwner);
  FNodeID := NodeID;
  FNodeName := NodeName;
  FNodeType := NodeType;
  FObjectList := TStringList.Create;
  FShowSystem := false;
end;

destructor TBS_TreeNode.Destroy();
begin
  FObjectList.Free;
  inherited Destroy();
end;



////////////////////  TBS_OraSessionNode  ///////////////////////
constructor TBS_OraSessionNode.Create(AOwner: TComponent;
                       const NodeID: HTreeItem;
                       const NodeName,
                             Database,
                             Password,
                             UserName,
                             Description: string;
                       const LastAccessed: TDateTime;
                       const NodeType: word);
begin
  inherited Create(AOwner,NodeID,NodeName,NodeType);
  FDescription := Description;
  FLastAccessed := LastAccessed;
  FOracleSession:= TOracleSession.Create(nil);
  FOracleSession.LogonDatabase:=Database;
  FOracleSession.LogonPassword:=Password;
  FOracleSession.LogonUsername:=UserName;

  //FOutputWindow := nil;
end;

destructor TBS_OraSessionNode.Destroy();
begin
  try
    if Self.FOracleSession.Connected then
     begin
       Self.FOracleSession.Commit;
       Self.FOracleSession.LogOff;
     end;
     Self.FOracleSession.Free;
  except on E: Exception do
     Self.FOracleSession.Free;
  end;
  inherited Destroy();
end;


function TBS_OraSessionNode.GetOraConnectionStatus:boolean;
begin
  Result:=FOracleSession.Connected;
end;

procedure TBS_OraSessionNode.Login;
begin
  FOracleSession.Connected:=True;
end;

function TBS_OraSessionNode.GetOraDatabase:String;
begin
  Result:=FOracleSession.LogonDatabase;
end;

function TBS_OraSessionNode.GetOraPassword:String;
begin
  Result:=FOracleSession.LogonPassword;
end;

function TBS_OraSessionNode.GetOraUserName:String;
begin
  Result:=FOracleSession.LogonUserName;
end;

procedure TBS_OraSessionNode.SetOraDatabase(const AOraDatabase:String);
begin
  FOracleSession.LogonDatabase:=AOraDatabase;
end;

procedure TBS_OraSessionNode.SetOraPassword(const AOraPassword:String);
begin
  FOracleSession.LogonPassword:=AOraPassword;
end;

procedure TBS_OraSessionNode.SetOraUserName(const AOraUserName:String);
begin
  FOracleSession.LogonUserName:=AOraUserName;
end;

procedure TBS_OraSessionNode.LogOut;
begin
  OracleSession.Connected:=False;
end;

function TBS_ConnectionNode.GetConnectionStatus:boolean;
begin
  Result:=DM.IsOracleConnected;
end;

constructor TBS_ConnectionNode.Create(AOwner: TComponent;
                       const NodeID: HTreeItem;
                       const NodeName: String;
                       const NodeType: word);
begin
  inherited Create(AOwner,NodeID,NodeName,NodeType);
end;

destructor TBS_ConnectionNode.Destroy();
begin
  inherited Destroy();
end;




constructor TBS_TaskNode.Create(AOwner: TComponent;
                       const NodeID: HTreeItem;
                       ANodeDataSet : TDataset);
begin
  inherited Create(AOwner,NodeID,
                   'Node'+ANodeDataSet.FieldByName('TASK_ID').AsString,
                   ANodeDataSet.FieldByName('NODETYPE_ID').Value);
  NodeDataSet:=ANodeDataSet;
  FDataSetRecord:=NodeDataSet.GetBookmark;
end;

destructor TBS_TaskNode.Destroy();
begin
  NodeDataSet.FreeBookMark(FDataSetRecord);
  inherited Destroy();
end;

procedure TBS_TaskNode.GotoDataRecord;// Try Get Mared with BookMark Record
begin
  NodeDataSet.GotoBookmark(FDataSetRecord);
end;


// Try Get Data Field Value
function TBS_TaskNode.GetDataFieldValue(const AFieldName:string):Variant;
var
 thePrevPosition:TBookMark;
begin
  thePrevPosition:=nil;
  with NodeDataSet do
  try
    thePrevPosition:=GetBookmark;
    GotoBookmark(FDataSetRecord);
    Result:=NodeDataSet.FieldByName(AFieldName).Value;
  finally
    if Assigned(thePrevPosition) then
      begin
       GotoBookmark(thePrevPosition);
       FreeBookMark(thePrevPosition);
      end;
  end;
end;

function TBS_TaskNode.GetDataFieldAsString(const AFieldName:string):String;
var
 thePrevPosition:TBookMark;
begin
  thePrevPosition:=nil;
  with NodeDataSet do
  try
    thePrevPosition:=GetBookmark;
    GotoBookmark(FDataSetRecord);
    Result:=NodeDataSet.FieldByName(AFieldName).AsString;
  finally
    if Assigned(thePrevPosition) then
      begin
       GotoBookmark(thePrevPosition);
       FreeBookMark(thePrevPosition);
      end;
  end;
end;


end.
